#!/usr/bin/env python3
##########################################################################
#  This file is part of the TAI abstract interpreter.                    #
#                                                                        #
#  Copyright (C) 2024                                                    #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  You can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file LICENSE).                      #
##########################################################################

# Simple python script to pretty print results of the analysis from CSV file
from csv import reader, writer
from re import sub
from typing import NamedTuple, Union, Optional

from run import DOMAIN, FILENAME, USE_TERMINAL_CODES, PREFIX


TIMES_CSV = PREFIX / "times.csv"
NUMERIC_CSV = PREFIX / "numeric.csv"
Number = Union[int, float]  # Sadly debian docker is still on python 3.9 :(


class DataSeries(NamedTuple):
    """Series of numbers and statistical indicators"""

    average: Number
    length: int
    standard_deviation: float
    sum: Number
    data: list[Number]
    ratio: Optional[float]


MILLION = "\033[31mM\033[37m" if USE_TERMINAL_CODES else "M"
THOUSAND = "\033[33mk\033[37m" if USE_TERMINAL_CODES else "k"


STR_OR_NUM = Union[str, Number]


class Table:
    """Used to pretty-print a table"""

    table: list[list[Union[str, Number]]]
    widths: list[int]

    column_delimiter = " "  # "|"
    line_delimiter = "-"
    cross_delimiter = "-"

    footer = 2

    def column_width(self, column: int) -> int:
        """Compute the width of a column"""
        width = 0
        for row in self.table:
            elt = row[column]
            if isinstance(elt, float):
                elt = round(elt, 2)
                if len(str(elt)) >= 4:
                    elt = round(elt)
            if (isinstance(elt, int) or isinstance(elt, float)) and elt > 1_000:
                if elt > 1_000_000:
                    elt = str(round(elt / 1_000_000)) + "M"
                else:
                    elt = str(round(elt / 1_000)) + "k"
            width = max(width, len(str(elt)))
        return width

    def pad_column(self, column: int, pad_numbers_right: bool = True) -> list[str]:
        """Create a padded version of the column"""
        width = self.widths[column]
        padded = list()
        for row in self.table:
            elt = row[column]
            if isinstance(elt, float):
                elt = round(elt, 2)
                if len(str(elt)) >= 4:
                    elt = round(elt)
            if isinstance(elt, int) or isinstance(elt, float):
                extra_width = 0  # extra width to account for ANSI codes
                if elt > 1_000_000:
                    elt = str(round(elt / 1_000_000)) + MILLION
                    extra_width = len(MILLION) - 1
                elif elt > 1_000:
                    elt = str(round(elt / 1_000)) + THOUSAND
                    extra_width = len(THOUSAND) - 1
                if pad_numbers_right:
                    padded.append(str(elt).rjust(width + extra_width))
                else:
                    padded.append(str(elt).ljust(width + extra_width))
            elif pad_numbers_right and elt == "--":
                padded.append(str(elt).rjust(width))
            else:
                padded.append(str(elt).ljust(width))
        return padded

    def __init__(self, values: list[list[STR_OR_NUM]]) -> None:
        self.table = values
        self.widths = [self.column_width(i) for i in range(len(values[0]))]

    def pp_ascii_line_sep(self, use_ansi: bool) -> None:
        sep = self.cross_delimiter.join(
            self.line_delimiter * width for width in self.widths
        )
        if use_ansi:
            sep = "\033[1;32m" + sep + "\033[0m"
        print(sep)

    def pp_ascii(self, use_ansi: bool = USE_TERMINAL_CODES) -> None:
        """Render as ascii and print to stdout"""
        padded = [self.pad_column(i) for i in range(len(self.widths))]
        # self.pp_ascii_line_sep(use_ansi)
        bg_colors = ("\033[48;5;17m", "\033[48;5;16m")  # blue, black
        if use_ansi:
            # Bold first line
            print("\033[1;32;4;48;5;16m", end="")
        size = len(padded[0])
        for row in range(size):
            if row == size - self.footer:
                bg_colors = (bg_colors[1], bg_colors[0])
                if use_ansi:
                    print(bg_colors[row % 2], end="")
                self.pp_ascii_line_sep(use_ansi)
            text = self.column_delimiter.join(col[row] for col in padded)
            if use_ansi:
                if row == 0:
                    text = "\033[1;32;4;48;5;16m" + text + "\033[0m"
                else:
                    text = "\033[37m" + bg_colors[row % 2] + text + "\033[0m"
            print(text)
            if row == 0 and not USE_TERMINAL_CODES:
                self.pp_ascii_line_sep(use_ansi)
        # self.pp_ascii_line_sep(use_ansi)


def compute_series(
    data: list[Number], reference_avg: Optional[Number]
) -> Optional[DataSeries]:
    if data == []:
        return None
    total = sum(data)
    length = len(data)
    average = total // length if total % length == 0 else total / length
    deviation = 0.0
    for x in data:
        deviation += (x - average) ** 2
    deviation = deviation**0.5
    return DataSeries(
        average=average,
        length=length,
        sum=total,
        data=data,
        standard_deviation=deviation,
        ratio=None if reference_avg is None else average / reference_avg,
    )


def group_by(results: list[list[str]], key: int) -> dict[str, list[list[str]]]:
    """Group elements by the given key,
    i.e. returns dict elt: [res1, resn] where
    exists j such that resi = results[j].remove(key) and results[j][key] == elt

    Destroys results in the process
    """
    grouped: dict[str, list[list[str]]] = dict()
    for row in results:
        elt = row[key]
        del row[key]
        if elt in grouped:
            grouped[elt].append(row)
        else:
            grouped[elt] = [row]
    return grouped


def read_csv() -> tuple[list[str], dict[str, dict[str, dict[str, list[list[str]]]]]]:
    with open(FILENAME, "r") as file:
        rows = list(reader(file))
    headers = rows[0].copy()
    domain_id = headers.index("Domain")
    sva_id = headers.index("SVA")
    del rows[0]
    by_file = group_by(rows, 0)
    by_file_sva = dict()
    for filename in by_file:
        by_file_sva[filename] = group_by(by_file[filename], sva_id - 1)
    by_file_sva_domain: dict[str, dict[str, dict[str, list[list[str]]]]] = dict()
    for filename in by_file:
        by_file_sva_domain[filename] = dict()
        for sva in by_file_sva[filename]:
            by_file_sva_domain[filename][sva] = group_by(
                by_file_sva[filename][sva], domain_id - 1
            )
    return headers, by_file_sva_domain


def shorten_domain(x: STR_OR_NUM) -> STR_OR_NUM:
    """shorten a domain name for terminal display"""
    if not isinstance(x, str):
        return x
    return sub(
        r"\[.*?\]",
        "",
        x.replace(" ", "").replace("LiftGVN", "G").replace("Lift", "L"),
    )


def shorten_filename(x: STR_OR_NUM) -> STR_OR_NUM:
    if not isinstance(x, str):
        return x
    if x.startswith("examples/csmith/"):
        return "c/" + x[len("examples/csmith/") :]
    if x.startswith("examples/snippets/"):
        return "s/" + x[len("examples/snippets/") :]
    return x


def remove_duplicates(table: list[list[STR_OR_NUM]], column: int) -> None:
    """Remove duplicate in successive row, i.e. replace
    [['foo', ...], ['foo', ...]] with [['foo', ...], ['---', ...]]
    This make printing somewhat more legible"""
    for i, row in reversed(list(enumerate(table))):
        if i != 0 and row[column] == table[i - 1][column]:
            row[column] = "." * len(str(row[column]))


# Order of the columns in the paper
sorted_columns = DOMAIN


def sort_columns(dom: str) -> tuple[int, Union[str, int]]:
    if dom in sorted_columns:
        return (0, sorted_columns.index(dom))
    return (1, dom)


def sums_and_averages(
    table: list[list[STR_OR_NUM]],
) -> tuple[list[STR_OR_NUM], list[STR_OR_NUM]]:
    sums: list[STR_OR_NUM] = []
    avgs: list[STR_OR_NUM] = []
    for col in range(0, len(table[0])):
        values: list[Number] = []
        for row in range((len(table))):
            cell = table[row][col]
            if isinstance(cell, str):
                continue
            values.append(cell)
        series = compute_series(values, None)
        if series is None:
            sums.append("")
            avgs.append("")
        else:
            sums.append(series.sum)
            avgs.append(series.average)
    return sums, avgs


if __name__ == "__main__":
    hd, csv = read_csv()

    # Times table
    all_doms: set[str] = set()

    for filename in csv:
        for sva in csv[filename]:
            all_doms = all_doms.union(set(csv[filename][sva].keys()))

    def simpl(x: str) -> str:
        return sub(r"\[.*?\]", "", x)

    all_domains = list(set(simpl(x) for x in all_doms))
    all_domains.sort(key=sort_columns)
    table_hd = ["File", "SVA"] + all_domains
    table: list[list[STR_OR_NUM]] = []
    time_id = hd.index("total time") - 3
    for filename in csv:
        for sva in csv[filename]:
            line: list[STR_OR_NUM] = [filename, sva]
            for domain in all_domains:
                for by_dom in csv[filename][sva]:
                    if simpl(by_dom) != domain:
                        continue
                    dom = csv[filename][sva][by_dom]
                    value = compute_series([int(x[time_id]) for x in dom], None)
                    if value is None:
                        line.append("--")
                    else:
                        line.append(value.average)
                    break
                else:
                    line.append("--")
            table.append(line)
    table.sort()
    sums, avgs = sums_and_averages(table)
    sums[0] = "Sum"
    avgs[0] = "Average"
    table.append(sums)
    table.append(avgs)

    with open(TIMES_CSV, "w") as file:
        csv_writer = writer(file)
        csv_writer.writerow(table_hd)
        csv_writer.writerows(table)

    for row in table:
        row[0] = shorten_filename(row[0])
    remove_duplicates(table, 0)
    table.insert(
        0,
        [shorten_domain(x) for x in table_hd],
    )

    times = Table(table)
    if USE_TERMINAL_CODES:
        print("\033[33;1mAnalysis times, averaged on all runs, in microseconds:\033[0m")
    else:
        print("Analysis times, averaged on all runs, in microseconds:")
    print(
        "Note that the paper table is in milliseconds\n\n"
        "In an attempt to get this table to fit a terminal (~140 columns):\n"
        "- filename are shortened ('examples/csmith/' -> 'c/' and 'examples/snippets/' -> 's/')\n"
        "- domain names have been shortened (stripped SVA, renamed Lift -> L and LiftGVN -> G)\n\n"
        "No value (indicated by '--') is expected in repeat calls of non-numeric domains (that dont contain N or QN).\n"
        "Extra non-values are most likely timeouts\n\n"
        f"The full unabbreviated table has also been written to {TIMES_CSV}\n"
    )
    times.pp_ascii()

    # Num compare_table

    table2: list[list[STR_OR_NUM]] = []
    for filename in csv:
        for sva in csv[filename]:
            for domain in csv[filename][sva]:
                if csv[filename][sva][domain][0][-1] == "--":
                    continue
                line2: list[STR_OR_NUM] = [filename, sva, domain]
                line2.extend(
                    list(int(x) for x in csv[filename][sva][domain][0][-1:-9:-1])[::-1]
                )
                table2.append(line2)

    for row in table2:
        row[0] = shorten_filename(row[0])
        row[2] = shorten_domain(row[2])

    table2.sort()
    sums, avgs = sums_and_averages(table2)
    sums[0] = "Sum"
    avgs[0] = "Average"
    table2.append(sums)
    table2.append(avgs)

    with open(NUMERIC_CSV, "w") as file:
        csv_writer = writer(file)
        csv_writer.writerow(
            [
                "File",
                "SVA",
                "Domain",
                "Numeric state: equal",
                "Numeric state: SSA more precise",
                "Numeric state: SSA less precise",
                "Numeric state: incomparable",
                "Numeric succs: equal",
                "Numeric succs: SSA more precise",
                "Numeric sucss: SSA less precise",
                "Numeric succs: incomparable",
            ]
        )
        csv_writer.writerows(table2)

    remove_duplicates(table2, 0)
    remove_duplicates(table2, 1)
    table2.insert(
        0,
        [
            "File",
            "SVA",
            "Domain",
            "SSA=Cl",
            "SSA>Cl",
            "SSA<Cl",
            "incomp",
            "SSA=Cl",
            "SSA>Cl",
            "SSA<Cl",
            "incomp",
        ],
    )
    table2[-1][1] = ""
    print("")
    if USE_TERMINAL_CODES:
        print("\033[33;1mComparing precision of numeric analysis:\033[0m")
    else:
        print("Comparing precision of numeric analysis:")
    print(f"The full table has also been written to {NUMERIC_CSV}\n")
    prec = Table(table2)
    if USE_TERMINAL_CODES:
        print(
            "\033[48;5;16;1;32m",
            " " * (sum(prec.widths[0:3]) + 3),
            "\033[4m",
            "All states all vars".center(sum(prec.widths[3:7]) + 3),
            "\033[0m\033[48;5;16;1;32m",
            prec.column_delimiter,
            "\033[4m",
            "Successors only".center(sum(prec.widths[7:11]) + 3),
            "\033[0m",
            sep="",
        )
    else:
        print(
            " " * (sum(prec.widths[0:3]) + 3),
            "All states all vars".center(sum(prec.widths[3:7]) + 3),
            prec.column_delimiter,
            "Successors only".center(sum(prec.widths[7:11]) + 3),
            sep="",
        )

    prec.pp_ascii()
