#!/usr/bin/env python3
##########################################################################
#  This file is part of the TAI abstract interpreter.                    #
#                                                                        #
#  Copyright (C) 2024                                                    #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  You can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file LICENSE).                      #
##########################################################################

# Simple python script to call our analysis multiple times in parrallel to
# quickly build results table
from os import path, walk
from queue import Empty, Queue
from subprocess import CalledProcessError, TimeoutExpired, check_call, check_output
from sys import stderr, stdout
from threading import Thread
from typing import NamedTuple
from pathlib import Path
from random import shuffle


# FIXME: Change this value according to the specs of your machine
# Should be:
# - smaller than your number of CPUs
# - smaller than your RAM / 10GB (each thread can take up to 10GB, although that
#   is exceptional, most cases are under 1GB and a few between 1GB and 7GB)
# For reference, on my machine (specs below), running the docker, the analysis
# with the present settings took 11 minutes real time (49 min single threaded)
# and used up to 40 GB or RAM
# CPU : 13th Gen Intel i7-13800H (20) @ 5.000GHz and RAM : 64 GiB
NB_THREADS = 5

# Change to False to print text without color/ANSI codes
USE_TERMINAL_CODES = stdout.isatty()

# Timeout for each analysis. You might need to increase this value on older machines
# With this value, I got 22 timeout out of 1536 analysis with
TIMEOUT = 100.0  # seconds

# Path to write output files to
PREFIX = Path("/home/opam/results")
PREFIX.mkdir(parents=True, exist_ok=True)
FILENAME = PREFIX / "results.csv"

# The following variables define which analysis will be run
# An analysis will be run for each case in each of the following lists, so longer
# list will lead to more analysis
STRATEGY = ["recursive"]  # subset of: ["recursive", "iterative"]
SVA = ["I", "IKS"]  # subset of ["I", "K", "S", "IS", "KS", "IKS"]
DOMAIN = [
    "QN",
    "LiftGVN(QN)",
    "Q(N*FA)",
    "Q(LiftGVN(N)*FA)",
    "FA",
    "LiftGVN(FA)",
    "LiftGVN(Q(N*FA))",
    # "LiftGVN(FA)",
    # "Lift(FA)"
    # "N * FA",
    # # "LiftGVN(N) * FA",
    "N*Q(LiftGVN(N))",
    "N*Q(LiftGVN(N*FA))",
    "E(LiftGVN(E(QN*FA)))",
    # "E(Q(N * FA))",
]
NB_BACK_ITERATIONS = [3]  # any positive integer


ITERATIONS = 2  # Number of times to run each analysis, to get an average time

# Folders/files to test files
# - if file, must end in .c
# - if folder, recursively find all contained .c files
FOLDERS = [
    # Examples from the paper table
    # "examples/csmith/c04.c",
    # "examples/csmith/c07.c",
    # "examples/csmith/c18.c",
    # "examples/csmith/c00.c",
    # "examples/csmith/c30.c",
    # "examples/csmith/c02.c",
    # "examples/csmith/c29.c",
    # "examples/csmith/c23.c",
    # "examples/csmith/c24.c",
    # All example files
    "./examples/snippets",
    "./examples/csmith",
]

# TAI executable
TAI = ["frama-c", "-no-autoload-plugins", "-load-plugin", "tai"]
# TAI = ["dune", "exec", "--", "frama-c", "-no-autoload-plugins", "-load-plugin", "tai"]

counter = 0
total = 0
errors = 0


class Parameters(NamedTuple):
    file: str
    iterative_strategy: bool
    sva: str
    domain: str
    nb_back_iter: int


pre_tasks: list[Parameters] = list()
tasks: Queue[Parameters] = Queue()


def format_command(param: Parameters) -> list[str]:
    """Transform the parameters into a valid command line call"""
    cmd = TAI + [
        "-tai",
        "-quiet",
        param.file,
        "-tai-sva",
        param.sva,
        "-tai-domain",
        param.domain,
        "-tai-nb-back-iter",
        str(param.nb_back_iter),
        "-tai-display",
        "CSV",
    ]
    if not USE_TERMINAL_CODES:
        cmd.append("-tai-no-color")
    if param.iterative_strategy:
        cmd.append("-tai-iterative-strategy")
    return cmd


def run_command(command: list[str]) -> str:
    """Runs the given command in a subprocess ITERATION times,
    The command should print a single line of comma separated values"""
    output = check_output(command, timeout=TIMEOUT)
    return output.decode().strip()


res: Queue[Parameters] = Queue()


class MyThread(Thread):
    def run(self) -> None:
        global counter, total, file, errors
        while True:
            try:
                param = tasks.get_nowait()
            except Empty:
                return

            command = format_command(param)

            try:
                results = run_command(command)
                file.write(results + "\n")
                if USE_TERMINAL_CODES:
                    print("                ", end="\r")
            except (CalledProcessError, TimeoutExpired) as err:
                if USE_TERMINAL_CODES:
                    print("                ", end="\r")
                    print(
                        "\033[31m" + str(err).replace(",", "") + "\033[0m", file=stderr
                    )
                else:
                    print(str(err).replace(",", ""), file=stderr)
                errors += 1
            counter += 1
            if USE_TERMINAL_CODES:
                print(f"Done: {counter} / {total}", end="\r")
            elif counter % 100 == 0:
                print(f"Done: {counter} / {total}")


def add_tasks(file: str) -> None:
    """adds all analysis relative to the given file to the tasks queue"""
    global total
    for dom in DOMAIN:
        for sva in SVA:
            for strategy in STRATEGY:
                iterative_strategy = strategy == "iterative"
                for nb_iters in NB_BACK_ITERATIONS:
                    for _ in range(ITERATIONS):
                        param = Parameters(
                            file=file,
                            iterative_strategy=iterative_strategy,
                            sva=sva,
                            domain=dom,
                            nb_back_iter=nb_iters,
                        )
                        pre_tasks.append(param)
                        total += 1
            if "N" not in dom.replace("LiftGVN", "LG"):
                # Domain doesn't depend on SVA
                break


if __name__ == "__main__":
    file = open(FILENAME, "w")
    nb_files = 0
    for folder in FOLDERS:
        if folder.endswith(".c"):
            add_tasks(folder)
            nb_files += 1
            continue
        for root, dirs, files in walk(folder):
            for filename in files:
                _, extension = path.splitext(filename)
                if extension == ".c":
                    add_tasks(path.join(root, filename))
                    nb_files += 1
    print(
        f"Running analysis on {nb_files} files, {total // nb_files // ITERATIONS} analysis per file,"
        f" {ITERATIONS} iterations.\nParallel computing using {NB_THREADS} threads."
    )

    # for filename in ['loop', 'euclid']:
    #     add_tasks(path.join("examples/", filename + ".c"))

    check_call(["dune", "build"])
    check_call(["dune", "install"])

    file.write(
        'File,Function,"input nodes","input assertions",Domain,SVA,strategy,'
        '"widening edges","output nodes","output assertions","output dead ends",'
        '"WTO time","convergence time","graph time","total time",'
        '"Numeric state: equal","Numeric state: SSA more precise","Numeric state: SSA less precise","Numeric state: incomparable",'
        '"Numeric succs: equal","Numeric succs: SSA more precise","Numeric sucss: SSA less precise","Numeric succs: incomparable"\n'
    )

    print(f"Results will be written to '{FILENAME}'.")
    print("")

    # Shuffling the tasks spreads the performance heavy analysis instead of leaving
    # them clumped together, hopefully, this helps keep RAM usage low
    shuffle(pre_tasks)
    for t in pre_tasks:
        tasks.put(t)

    threads = []
    for i in range(NB_THREADS):
        thread = MyThread()
        threads.append(thread)
        thread.start()

    print(f"Done: {counter} / {total}", end="\r")
    try:
        for thread in threads:
            thread.join()
    except KeyboardInterrupt:
        pass
    file.close()
    print("")

    if errors:
        if USE_TERMINAL_CODES:
            print(f"\033[31m{errors} analysis had errors.\033[0m")
        else:
            print(f"{errors} analysis had errors.")
    print("")
    print("Now run format_results.py to format the results")


# for i in threads:
#     i.kill()
