(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Std_extra

type binop = Frama_c_kernel.Cil_types.binop =
  | PlusA  (** arithmetic + *)
  | PlusPI  (** pointer + integer *)
  | MinusA  (** arithmetic - *)
  | MinusPI  (** pointer - integer *)
  | MinusPP  (** pointer - pointer *)
  | Mult  (** * *)
  | Div  (** / *)
  | Mod  (** % *)
  | Shiftlt  (** shift left *)
  | Shiftrt  (** shift right *)
  | Lt  (** <  (arithmetic comparison) *)
  | Gt  (** >  (arithmetic comparison) *)
  | Le  (** <= (arithmetic comparison) *)
  | Ge  (** >= (arithmetic comparison) *)
  | Eq  (** == (arithmetic comparison) *)
  | Ne  (** != (arithmetic comparison) *)
  | BAnd  (** bitwise and *)
  | BXor  (** exclusive-or *)
  | BOr  (** inclusive-or *)
  | LAnd  (** logical and. *)
  | LOr  (** logical or.  *)
[@@deriving ord, eq]

type unop = Frama_c_kernel.Cil_types.unop =
  | Neg
  | BNot
  | LNot
[@@deriving ord, eq]

type 'var expr_node =
  | E_Var of 'var
  | E_Const of Z.t
  | E_Binop of 'var expr * binop * 'var expr
  | E_Unop of unop * 'var expr
  | E_IfThenElse of 'var expr * 'var expr * 'var expr
and 'var expr = 'var expr_node Hashconsing.hash_consed

module Binop = struct
  type t = binop [@@deriving ord, eq] [@@deriving ord, eq]

  let order a b = Utils.order_of_compare (compare a b)

  let hash = function
    | PlusA -> 1
    | PlusPI -> 2
    | MinusA -> 3
    | MinusPP -> 4
    | MinusPI -> 5
    | Mult -> 6
    | Div -> 7
    | Mod -> 8
    | Shiftlt -> 9
    | Shiftrt -> 10
    | Lt -> 11
    | Gt -> 12
    | Le -> 13
    | Ge -> 14
    | Eq -> 15
    | Ne -> 16
    | BAnd -> 17
    | BXor -> 18
    | BOr -> 19
    | LAnd -> 20
    | LOr -> 21

  let pp fmt b =
    Format.fprintf fmt "%s"
      (match b with
      | PlusA | PlusPI -> "+"
      | MinusA | MinusPP | MinusPI -> "-"
      | Mult -> "*"
      | Div -> "/"
      | Mod -> "%"
      | Shiftlt -> "<<"
      | Shiftrt -> ">>"
      | Lt -> "<"
      | Gt -> ">"
      | Le -> Utils.Unicode.leq ()
      | Ge -> Utils.Unicode.geq ()
      | Eq -> Utils.Unicode.eq ()
      | Ne -> Utils.Unicode.neq ()
      | BAnd -> "&"
      | BXor -> "^"
      | BOr -> "|"
      | LAnd -> Utils.Unicode.logical_and ()
      | LOr -> Utils.Unicode.logical_or ())

  (* see https://en.cppreference.com/w/c/language/operator_precedence *)
  let level = function
    | Mult | Div | Mod -> 3
    | PlusA | PlusPI | MinusA | MinusPI | MinusPP -> 4
    | Shiftlt | Shiftrt -> 5
    | Lt | Gt | Le | Ge -> 6
    | Eq | Ne -> 7
    | BAnd -> 8
    | BXor -> 9
    | BOr -> 10
    | LAnd -> 11
    | LOr -> 12
end

module Unop = struct
  type t = unop [@@deriving ord, eq]

  let order a b = Utils.order_of_compare (compare a b)

  let hash = function
    | Neg -> 1
    | BNot -> 2
    | LNot -> 3

  let to_string = function
    | Neg -> "-"
    | BNot -> "~"
    | LNot -> Utils.Unicode.logical_not ()

  let pp fmt x = Format.fprintf fmt "%s" (to_string x)

  let level (_ : t) = 1
end

let z_to_int x = if Z.fits_int x then Z.to_int x else 0

module rec ExprNode : (Types.UNARY_TYPE with type 'a t = 'a expr_node) = struct
  type 'a t = 'a expr_node

  let equal equal_var l r =
    match (l, r) with
    | E_Var l, E_Var r -> equal_var l r
    | E_Const l, E_Const r -> Z.equal l r
    | E_Binop (ll, lo, lr), E_Binop (rl, ro, rr) ->
        Expr.equal equal_var ll rl && Binop.equal lo ro && Expr.equal equal_var lr rr
    | E_Unop (lo, le), E_Unop (ro, re) -> Unop.equal lo ro && Expr.equal equal_var le re
    | E_IfThenElse (li, lt, le), E_IfThenElse (ri, rt, re) ->
        Expr.equal equal_var li ri && Expr.equal equal_var lt rt && Expr.equal equal_var le re
    | _, _ -> false

  let compare compare_var l r =
    match (l, r) with
    | E_Var l, E_Var r -> compare_var l r
    | E_Var _, _ -> Utils.compare_inf
    | _, E_Var _ -> Utils.compare_sup
    | E_Const l, E_Const r -> Z.compare l r
    | E_Const _, _ -> Utils.compare_inf
    | _, E_Const _ -> Utils.compare_sup
    | E_Binop (ll, lo, lr), E_Binop (rl, ro, rr) ->
        let cmp = Expr.compare compare_var ll rl in
        if cmp <> 0 then cmp
        else
          let cmp = Binop.compare lo ro in
          if cmp <> 0 then cmp else Expr.compare compare_var lr rr
    | E_Binop _, _ -> Utils.compare_inf
    | _, E_Binop _ -> Utils.compare_sup
    | E_Unop (lo, le), E_Unop (ro, re) ->
        let cmp = Unop.compare lo ro in
        if cmp <> 0 then cmp else Expr.compare compare_var le re
    | E_Unop _, _ -> Utils.compare_inf
    | _, E_Unop _ -> Utils.compare_sup
    | E_IfThenElse (li, lt, le), E_IfThenElse (ri, rt, re) ->
        let cmp = Expr.compare compare_var li ri in
        if cmp <> 0 then cmp
        else
          let cmp = Expr.compare compare_var lt rt in
          if cmp <> 0 then cmp else Expr.compare compare_var le re

  let order order a b = Utils.unary_order_of_compare compare order a b

  let level = function
    | E_Var _ | E_Const _ -> 0
    | E_Binop (_, o, _) -> Binop.level o
    | E_Unop (u, _) -> Unop.level u
    | E_IfThenElse _ -> 13

  let rec pp lvl pp_var fmt =
    let parens = pp_parens lvl pp_var in
    function
    | E_Var v -> pp_var fmt v
    | E_Const z -> Z.pp_print fmt z
    | E_Binop (l, o, r) -> Format.fprintf fmt "@[%a@ %a@ %a@]" parens l Binop.pp o parens r
    | E_Unop (o, e) -> Format.fprintf fmt "%a%a" Unop.pp o parens e
    | E_IfThenElse (i, t, e) ->
        Format.fprintf fmt "@[if %a@ then %a@ else %a@]" parens i parens t parens e

  and pp_parens lvl pp_var fmt expr =
    let level = level expr.node in
    if level >= lvl then Format.fprintf fmt "(%a)" (pp level pp_var) expr.node
    else pp level pp_var fmt expr.node

  let pp pp_var fmt e = Format.fprintf fmt "@[@{<rgb(254,246,140)>%a@}@]" (pp 14 pp_var) e

  let hash hash_v =
    let open Std_extra.Utils in
    let hash_sum = hash_sum ~log_total:3 in
    function
    | E_Var v -> hash_sum ~nb:0 (hash_v v)
    | E_Const z -> hash_sum ~nb:1 (z_to_int z)
    | E_Binop (l, o, r) ->
        hash_sum ~nb:2
          (hash_combine (Expr.hash hash_v l) (hash_combine (Expr.hash hash_v r) (Binop.hash o)))
    | E_Unop (o, e) -> hash_sum ~nb:3 (hash_combine (Expr.hash hash_v e) (Unop.hash o))
    | E_IfThenElse (i, t, e) ->
        hash_sum ~nb:4
          (hash_combine (Expr.hash hash_v i)
             (hash_combine (Expr.hash hash_v t) (Expr.hash hash_v e)))
end

and Expr : (Types.UNARY_TYPE with type 'a t = 'a expr) = struct
  type 'a t = 'a expr
  open Hashconsing
  let compare _ a b = Int.compare a.tag b.tag
  let equal _ a b = Int.equal a.tag b.tag
  let hash _ a = a.hkey
  let pp pp fmt a = ExprNode.pp pp fmt a.node
  let order _ _ _ = assert false
end

let bind of_node f =
  let rec bind (e : 'a expr) =
    match e.node with
    | E_Var v -> f v
    | E_Const z -> of_node (E_Const z)
    | E_Binop (l, o, r) -> of_node (E_Binop (bind l, o, bind r))
    | E_Unop (o, e) -> of_node (E_Unop (o, bind e))
    | E_IfThenElse (i, t, e) -> of_node (E_IfThenElse (bind i, bind t, bind e))
  in
  bind

let map of_node f = bind of_node (fun v -> of_node (E_Var (f v)))

let rec size (e : 'a expr) =
  match e.node with
  | E_Var _ | E_Const _ -> 1
  | E_Binop (l, _, r) -> 1 + size l + size r
  | E_Unop (_, e) -> 1 + size e
  | E_IfThenElse (i, t, e) -> 1 + size i + size t + size e

(* let rec is_const_node (e : 'a expr_node) =
   match e with
   | E_Var _ -> false
   | E_Const _ -> true
   | E_Binop (l, _, r) -> is_const l && is_const r
   | E_Unop (_, e) -> is_const e
   | E_IfThenElse (i, t, e) -> is_const i && is_const t && is_const e *)

(** Check if the expression is a constant *)
let is_const_toplevel = function
  | E_Const _ -> true
  | _ -> false

let is_constant (e : 'a expr) =
  match e.node with
  | E_Const z -> Some z
  | _ -> None

(** Check if the expression is an immediate operation on constants *)
let is_const_shallow = function
  | E_Var _ -> false
  | E_Const _ -> false
  | E_Binop (l, op, r) ->
      op <> Div && op <> Mod && is_const_toplevel l.node && is_const_toplevel r.node
  | E_Unop (_, e) -> is_const_toplevel e.node
  | E_IfThenElse (i, t, e) ->
      is_const_toplevel i.node && is_const_toplevel t.node && is_const_toplevel e.node

(** Since we simplify constant expression as we build them, a shallow check is sufficient *)
let is_const (e : 'a expr) = is_const_shallow e.node

let get_const (e : 'a expr) =
  match e.node with
  | E_Const z -> z
  | _ -> failwith "Get const of non-constant expr"

let eval_const_node (e : 'a expr_node) =
  let open Z in
  match e with
  | E_Var _ -> failwith "Eval const of non-constant expr"
  | E_Const z -> z
  | E_Unop (Neg, e) -> ~-(get_const e)
  | E_Unop (LNot, e) -> if equal (get_const e) zero then one else zero
  | E_Unop (BNot, e) -> ~!(get_const e)
  | E_IfThenElse (i, t, e) -> if equal (get_const i) zero then get_const e else get_const t
  | E_Binop (l, b, r) -> (
      match b with
      | PlusA | PlusPI -> get_const l + get_const r
      | MinusA | MinusPI | MinusPP -> get_const l - get_const r
      | Mult -> get_const l * get_const r
      | Div ->
          let r = get_const r in
          if equal r zero then raise Division_by_zero else get_const l / r
      | Mod ->
          let r = get_const r in
          if equal r zero then raise Division_by_zero else get_const l mod r
      | Shiftlt ->
          shift_left (get_const l)
            (let x = z_to_int (get_const r) in
             if x < 0 then 0 else x)
      | Shiftrt ->
          shift_right (get_const l)
            (let x = z_to_int (get_const r) in
             if x < 0 then 0 else x)
      | Lt -> if compare (get_const l) (get_const r) < 0 then one else zero
      | Gt -> if compare (get_const l) (get_const r) > 0 then one else zero
      | Le -> if compare (get_const l) (get_const r) <= 0 then one else zero
      | Ge -> if compare (get_const l) (get_const r) >= 0 then one else zero
      | Eq -> if compare (get_const l) (get_const r) = 0 then one else zero
      | Ne -> if compare (get_const l) (get_const r) <> 0 then one else zero
      | BAnd -> get_const l land get_const r
      | BXor -> get_const l lxor get_const r
      | BOr -> get_const l lor get_const r
      | LAnd ->
          if equal (get_const l) zero then zero else if equal (get_const r) zero then zero else one
      | LOr ->
          if equal (get_const l) zero then if equal (get_const r) zero then zero else one else one)

let rec depth (e : 'a expr) =
  match e.node with
  | E_Var _ | E_Const _ -> 1
  | E_Binop (l, _, r) -> 1 + max (depth l) (depth r)
  | E_Unop (_, e) -> 1 + depth e
  | E_IfThenElse (i, t, e) -> 1 + max (depth i) (max (depth t) (depth e))

module type S = sig
  type var
  include Types.TYPE with type t = var expr

  module Set : PatriciaTree.Set_S with type elt = var expr
  module Map : PatriciaTree.Map_S with type key = var expr

  val of_node : var expr_node -> var expr

  val var : var -> var expr
  val of_int : int -> var expr
  val of_const : Z.t -> var expr
  val unop : unop -> var expr -> var expr
  val binop : binop -> var expr -> var expr -> var expr
  val if_ : var expr -> var expr -> var expr -> var expr

  val zero : var expr
  val one : var expr
  val two : var expr
  val minus_one : var expr

  val simplify_expr : (var expr -> Z.t option) -> var expr -> var expr
end

module Make (V : Types.TYPE) = struct
  type var = V.t
  include Instances.Instanciate (Expr) (V)

  module Expr = struct
    type nonrec t = t

    let to_int (x : t) = x.tag
  end

  module VExprNode = Instances.Instanciate (ExprNode) (V)
  module H = Hashconsing.Make (VExprNode)

  module Set = PatriciaTree.MakeSet (Expr)
  module Map = PatriciaTree.MakeMap (Expr)

  let table = H.create 200

  let of_node x =
    H.hashcons table
      (if is_const_shallow x then try E_Const (eval_const_node x) with Division_by_zero -> x
       else x)

  let var x = H.hashcons table (E_Var x)
  let of_const z = H.hashcons table (E_Const z)
  let of_int x = H.hashcons table (E_Const (Z.of_int x))

  let if_ i t e = of_node (E_IfThenElse (i, t, e))
  let binop op l r = of_node (E_Binop (l, op, r))
  let unop op e = of_node (E_Unop (op, e))

  let zero = of_const Z.zero
  let one = of_const Z.one
  let two = of_int 2
  let minus_one = of_const Z.minus_one

  (** Simplify an expression *)
  let rec simplify_expr query_is_constant (expr : t) =
    let ( = ) = equal in
    match query_is_constant expr with
    | Some z -> of_const z
    | _ -> (
        match expr.node with
        | E_Binop (expr_l, op, expr_r) ->
            let expr_l = simplify_expr query_is_constant expr_l in
            let expr_r = simplify_expr query_is_constant expr_r in
            let nexpr =
              match op with
              | PlusA | PlusPI ->
                  (* Rewrite x + 0 or 0 + x into x *)
                  if expr_l = zero then expr_r
                  else if expr_r = zero then expr_l
                  else binop op expr_l expr_r
              | MinusA | MinusPI | MinusPP ->
                  (* x - 0 is x *)
                  if expr_r = zero then expr_l else binop op expr_l expr_r
              | Shiftlt | Shiftrt ->
                  (* x << 0 is x; 0 << x is 0 *)
                  if expr_r = zero then expr_l
                  else if expr_l = zero then expr_l
                  else binop op expr_l expr_r
              | Mult ->
                  (* x * 0 or 0 * x is 0; 1 * x or x * 1 is x *)
                  if expr_r = zero || expr_l = zero then zero
                  else if expr_l = one then expr_r
                  else if expr_r = one then expr_l
                  else binop op expr_l expr_r
              | Div ->
                  (* 0 / x is 0, x / 1 is x  *)
                  (* if expr_l = zero then zero
                     else *)
                  if expr_r = one then expr_l else binop op expr_l expr_r
              | Mod ->
                  (* 0 mod x is 0, x mod 1 is 0  *)
                  if expr_l = zero then zero
                  else if expr_r = one then zero
                  else binop op expr_l expr_r
              | Lt | Gt | Ne | BXor ->
                  (* x != x is false *)
                  if expr_l = expr_r then zero else binop op expr_l expr_r
              | Le | Ge | Eq ->
                  (* x == x is true *)
                  if expr_l = expr_r then one else binop op expr_l expr_r
              | BAnd ->
                  (* x & x is x; x & 0 and 0 & x is 0 *)
                  if expr_l = expr_r then expr_l
                  else if expr_l = zero || expr_r = zero then zero
                  else binop op expr_l expr_r
              | BOr ->
                  (* x | x is x; x | 0 and 0 | x is x *)
                  if expr_l = expr_r then expr_l
                  else if expr_l = zero then expr_r
                  else if expr_r = zero then expr_l
                  else binop op expr_l expr_r
              | LAnd ->
                  (* 0 and something is 0 *)
                  if expr_l = zero || expr_r = zero then zero else binop op expr_l expr_r
              | LOr -> binop op expr_l expr_r
            in
            let default = binop op expr_l expr_r in
            if nexpr = binop op expr_l expr_r then ()
            else Std_extra.Utils.debug_printf "Rewriting %a into %a@." pp default pp nexpr;
            nexpr
        | E_Unop (u, e) -> unop u (simplify_expr query_is_constant e)
        | E_IfThenElse (i, t, e) ->
            let i = simplify_expr query_is_constant i in
            if i = zero then simplify_expr query_is_constant e
            else if query_is_constant i <> None then simplify_expr query_is_constant t
            else
              let t = simplify_expr query_is_constant t in
              let e = simplify_expr query_is_constant e in
              if e = t then t else if_ i t e
        | E_Var _ | E_Const _ -> expr)
end
