(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Syntax tree for program expressions (both IMP and SSA ones).

  The main expression type is the unary {!expr}, which is hashconsed version
  of {!expr_node}. {!expr_node} is private to avoid breaking hash-consing: it can
  be inspected, but must be built using the constructors in provided by the
  {!Make} functor. *)

open Std_extra
open Types

(** Binary operators *)
type binop = Frama_c_kernel.Cil_types.binop =
  | PlusA  (** arithmetic + *)
  | PlusPI  (** pointer + integer, unsupported *)
  | MinusA  (** arithmetic - *)
  | MinusPI  (** pointer - integer, unsupported  *)
  | MinusPP  (** pointer - pointer, unsupported  *)
  | Mult  (** * *)
  | Div  (** / *)
  | Mod  (** % *)
  | Shiftlt  (** shift left *)
  | Shiftrt  (** shift right *)
  | Lt  (** <  (arithmetic comparison) *)
  | Gt  (** >  (arithmetic comparison) *)
  | Le  (** <= (arithmetic comparison) *)
  | Ge  (** >= (arithmetic comparison) *)
  | Eq  (** == (arithmetic comparison) *)
  | Ne  (** != (arithmetic comparison) *)
  | BAnd  (** bitwise and *)
  | BXor  (** exclusive-or *)
  | BOr  (** inclusive-or *)
  | LAnd  (** logical and. *)
  | LOr  (** logical or.  *)

(** Unary operators *)
type unop = Frama_c_kernel.Cil_types.unop =
  | Neg  (** Arithmetic negation *)
  | BNot  (** Bitwise not *)
  | LNot  (** Logical negation (non-zero -> zero, zero -> 1) *)

(** Expressions are parametrized by the type of variable
    so we can reuse them for SSA.
    They are also hashconsed, hence the private type
    Use the [Make(Var)] functor's smart constructors to build expressions *)
type 'var expr_node = private
  | E_Var of 'var
  | E_Const of Z.t
  | E_Binop of 'var expr * binop * 'var expr
  | E_Unop of unop * 'var expr
  | E_IfThenElse of 'var expr * 'var expr * 'var expr

and 'var expr = 'var expr_node Hashconsing.hash_consed

module type S = sig
  type var
  include TYPE with type t = var expr

  module Set : PatriciaTree.Set_S with type elt = var expr
  module Map : PatriciaTree.Map_S with type key = var expr

  val of_node : var expr_node -> var expr

  (** Constructors: These build hashconsed expressions. They also immediatly
      evaluate constant expressions, so [binop PlusA one one] is turned into [2]. *)

  val var : var -> var expr
  val of_int : int -> var expr
  val of_const : Z.t -> var expr
  val unop : unop -> var expr -> var expr
  val binop : binop -> var expr -> var expr -> var expr
  val if_ : var expr -> var expr -> var expr -> var expr

  val zero : var expr
  val one : var expr
  val two : var expr
  val minus_one : var expr

  val simplify_expr : (var expr -> Z.t option) -> var expr -> var expr
  (** [simplify_expr query_is_constant expr] performs a number of small rewrites
      on the expression [expr]. [query_is_constant] is used to determine if some
      variables/expression are constant. *)
end

module Binop : TYPE with type t = binop
module Unop : TYPE with type t = unop

val bind : ('b expr_node -> 'b expr) -> ('a -> 'b expr) -> 'a expr -> 'b expr
(** The first argument of [bind] and [map] is [of_node] to turn an element
    into it's hashconsed version *)

val map : ('b expr_node -> 'b expr) -> ('a -> 'b) -> 'a expr -> 'b expr

val is_const : 'a expr -> bool
(** [true] if the expression contains no variables *)

val size : 'a expr -> int
(** Expression size (number of nodes) *)

val depth : 'a expr -> int
(** Expression depth *)

val is_constant : 'a expr -> Z.t option

(** Create an expression domain with the given variable type, provide smart
    constructors for it and manage the hashconsing *)
module Make (Var : TYPE) : S with type var = Var.t
