(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Std_extra

let simple_error message =
  Utils.printf "@{<bold>@{<red>ERROR@}: %s@}@." message;
  exit 5

let error pp o message =
  Utils.printf "@{<bold>@{<red>ERROR@}: %s@}@.  %a@." message pp o;
  exit 5

let error_at loc pp o msg =
  Utils.printf "@{<bold>@{<red>ERROR@} at %a: %s@}@.  %a@." Frama_c_kernel.Cil_printer.pp_location
    loc msg pp o;
  exit 5

(** {1 Program Locations}                                *)
(*********************************************************)

(** Location are just integers (from the analysis' point of view)
    In practice they have a specific format to tie them back to original program
    locations when pretty-printing:
    - func_id: bits Sys.max .. Sys.max-8 (max 256 functions),
      encodes the function name via {!func_names}
    - func_call_id: bits Sys.max-8..Sys.max-16 (max 256 calls to a function)
      uniquely identifies each call of the given function
    - bits Sys.max-16..0 encode location inside a given function *)

let func_id_offset = Sys.int_size - 8
let func_id_mask = 0xff lsl func_id_offset
let call_id_offset = Sys.int_size - 16
let call_id_mask = 0xff lsl call_id_offset
let pre_call_offset = call_id_offset - 1
let pre_call_mask = 1 lsl pre_call_offset
let pos_id_mask = -1 lxor (func_id_mask lor call_id_mask lor pre_call_mask)

let pgrm_start = 0
let pgrm_end = -1 land pos_id_mask

type loc_info = {
  func_id : int;  (** between 0 and 255 *)
  call_id : int;  (** between 0 and 255 *)
  pre_call : bool;  (** true for extra, precall locations (argument setting) *)
  pos : int;
}

let loc_of_info info =
  if info.func_id < 0 || info.func_id > 255 then
    simple_error "Too many functions: only support < 255 functions"
  else if info.call_id < 0 || info.call_id > 255 then
    simple_error "Too many function calls: only supports < 255 calls to a single function"
  else if info.pos land pos_id_mask <> info.pos then
    simple_error "Too many position inside a single function"
  else
    (info.func_id lsl func_id_offset) lor (info.call_id lsl call_id_offset)
    lor (if info.pre_call then pre_call_mask else 0)
    lor info.pos

let info_of_loc loc =
  {
    func_id = (loc land func_id_mask) lsr func_id_offset;
    call_id = (loc land call_id_mask) lsr call_id_offset;
    pre_call = loc land pre_call_mask <> 0;
    pos = loc land pos_id_mask;
  }

type func_name = {
  id : int;
  name : string;
  mutable call_nb : int;
}

let func_names = ref []

let find_func_by_name name = List.find_opt (fun x -> x.name = name) !func_names
let find_func_by_id id = List.find_opt (fun x -> x.id = id) !func_names

module Loc = struct
  include Instances.Int

  let pp fmt i =
    let info = info_of_loc i in
    let func_name =
      if info.func_id = 0 then "" (* no name for main function to keep display short *)
      else
        match find_func_by_id info.func_id with
        | Some x -> x.name ^ ":"
        | None -> Format.sprintf "%d:" info.func_id
    in
    let pos =
      if info.pos = pgrm_start then "entry"
      else if info.pos = pgrm_end then "exit"
      else string_of_int info.pos
    in
    let func_call =
      if info.call_id = 0 && info.func_id = 0 then "" else Format.sprintf "%d:" info.call_id
    in
    let pre_call = if info.pre_call then "arg_" else "" in
    Format.fprintf fmt "@{<green>%s%s%s%s@}" func_name func_call pre_call pos

  let to_int x = x
end
module LocSet = PatriciaTree.MakeSet (Loc)
module LocMap = PatriciaTree.MakeMap (Loc)

(** {1 Program Variables and expressions}                *)
(*********************************************************)

module VarBase = struct
  include Instances.String

  let pp fmt x = Format.fprintf fmt "@{<rgb(246,180,0)>%s@}" x
end
module Var = Intify.Make (VarBase)
module VarExpr = Expr.Make (Var)
module VarMap = PatriciaTree.MakeMap (Var)
module VarSet = PatriciaTree.MakeSet (Var)

let tmp_var x = Var.get_or_create (Format.sprintf "tmp$%d" x)

(** {1 SSA Variables, Expression and Bindings}           *)
(*********************************************************)

module SSAVar = struct
  include Instances.InstanciateBoth (Instances.Pair) (Var) (Instances.Int)
  let pp fmt (v, id) = Format.fprintf fmt "@{<rgb(206,210,0)>%a@@%d@}" Var.pp v id
  let to_int x = hash x
end
module SSAVarMap = PatriciaTree.MakeMap (SSAVar)
module SSAExpr = Expr.Make (SSAVar)

let ssavar_init v = (v, 0)
let ssavar_join v i = (v, i)
let ssavar_rename ((v, i) as var) ~old n = if i = old then (v, n) else var

module Bindings = struct
  type t = SSAExpr.t SSAVarMap.t [@@deriving eq, ord]

  let order a b = Utils.order_of_compare (compare a b)

  let pp fmt bindings =
    if SSAVarMap.is_empty bindings then Format.fprintf fmt "%s" (Utils.Unicode.emptyset ())
    else
      Format.fprintf fmt "@[%a@]"
        (SSAVarMap.pp_map ~pp_sep:Utils.pp_semicolon (fun fmt var expr ->
             Format.fprintf fmt "@[%a%s%a@]" SSAVar.pp var (Utils.Unicode.colon_eq ()) SSAExpr.pp
               expr))
        bindings

  let hash bindings =
    SSAVarMap.fold
      (fun var expr hash ->
        Utils.hash_combine hash @@ Utils.hash_combine (SSAVar.hash var) (SSAExpr.hash expr))
      bindings 0
end

let bindings_pp pp_state fmt (l, r) =
  Format.fprintf fmt "@[<v>State:@,  @[<v>%a@]@,Bindings:@,  @[<v>%a@]@]" pp_state l Bindings.pp r

module MkBindings (SSA_State : Types.TYPE) = struct
  include Instances.InstanciateBoth (Instances.Pair) (SSA_State) (Bindings)
  let pp fmt x = bindings_pp SSA_State.pp fmt x
end

(** {1 Relations between program points}                 *)
(*********************************************************)

type relation =
  | Guard of VarExpr.t
  | Assign of Var.t * VarExpr.t
[@@deriving eq, ord]

let expr_of_rel = function
  | Guard e -> e
  | Assign (_, e) -> e

module Relation = struct
  type t = relation [@@deriving eq, ord]

  let order a b = Utils.order_of_compare (compare a b)

  let hash = function
    | Guard x -> Utils.hash_sum ~log_total:1 ~nb:0 (VarExpr.hash x)
    | Assign (v, e) ->
        Utils.hash_sum ~log_total:1 ~nb:1 (Utils.hash_combine (Var.hash v) (VarExpr.hash e))

  let pp fmt = function
    | Guard x -> Format.fprintf fmt "@{<rgb(246,116,0)>if %a@}" VarExpr.pp x
    | Assign (v, e) ->
        Format.fprintf fmt "@{<rgb(246,116,0)>%a@ %s@ %a@}" Var.pp v (Utils.Unicode.colon_eq ())
          VarExpr.pp e
end

(** {1 Program graph}                                                         *)
(******************************************************************************)

let goto = Guard VarExpr.one

type graph = {
  name : string;
  starting_vars : Var.t list;
  preds : (relation * Loc.t) list LocMap.t;
}

(** Flip edges of the given graph *)
let co_graph graph =
  (* We use a seen set to ensure we add childless elements as well *)
  let co_graph, seen =
    LocMap.fold
      (fun child parents (co, seen) ->
        ( List.fold_left
            (fun co (label, parent) -> LocMap.List.add parent (label, child) co)
            co parents,
          LocSet.add child seen ))
      graph (LocMap.empty, LocSet.empty)
  in
  LocSet.fold
    (fun x co_graph ->
      match LocMap.find_opt x co_graph with
      | Some _ -> co_graph
      | None -> LocMap.add x [] co_graph)
    seen co_graph

(** Count assertion in the input graph. An assertion is simply a node with no
    a single guarded child (or two children whose guards aren't "c" and "!c") *)
let count_assertions co_graph =
  LocMap.fold
    (fun loc children nb ->
      if Loc.equal loc pgrm_end then nb
      else
        match children with
        | [] -> nb + 1
        | [ (Assign (_, _), _) ] -> nb
        | [ (Guard e, _) ] -> (
            match e.node with
            | Expr.E_Const z when not (Z.equal z Z.zero) -> nb
            | _ -> nb + 1)
        | [ (Guard e, _); (Guard e', _) ] ->
            if VarExpr.(equal e (unop LNot e') || equal e' (unop LNot e)) then nb
            else (* Format.eprintf "Weird branching, shouldn't occur"; *)
              nb + 1
        | _ ->
            (* List.iter
               (fun (rel, x) -> Utils.printf "child %a through %a@." Loc.pp x Relation.pp rel)
               children; *)
            (* failwith "unsupported branching") *)
            nb)
    co_graph 0

let graph_pp fmt graph =
  Format.fprintf fmt "@[<v>%a@]"
    (LocMap.pp_map ~pp_sep:Format.pp_print_cut (fun fmt k v ->
         Format.fprintf fmt "%a:@   @[<v>%a@]" Loc.pp k
           (Format.pp_print_list ~pp_sep:Format.pp_print_cut (fun fmt (label, elem) ->
                Format.fprintf fmt "@[<hv><--{@[<hv>%a@]}--@ %a@]" Relation.pp label Loc.pp elem))
           v))
    graph.preds

(** {2 Convert from frama-c/Cil to our graph representation}                  *)
(******************************************************************************)

open Frama_c_kernel
open Cil_types

(* let inline_once = ref false *)

let compile_constant (cst : constant) error =
  VarExpr.of_const
    (match cst with
    | CInt64 (i, _, _) -> i
    | CChr c -> Z.of_int (int_of_char c)
    | CStr _ | CWStr _ -> error "Unsupported constants: string"
    | CReal (_, _, _) -> error "Unsupported constants: floats"
    | CEnum _ -> error "Unsupported constants: enum")

let var_name' func_info v =
  Var.get_or_create (if func_info.id = 0 then v else func_info.name ^ "#" ^ v)
let var_name func_info v = var_name' func_info v.vname

let compile_lval func_info (v : lval) error =
  match v with
  | Var v, _ -> var_name func_info v
  | _ -> error "Unsupported lval"

let check_binop (b : binop) error =
  match b with
  | PlusPI | MinusPI | MinusPP -> error "Unsupported binary operator"
  | b -> b

let rec compile_expr func_info (expr : exp) =
  let error = error_at expr.eloc Cil_printer.pp_exp expr in
  match expr.enode with
  | Const c -> compile_constant c error
  | Lval v -> VarExpr.var (compile_lval func_info v error)
  | UnOp (u, e, _) -> VarExpr.unop u (compile_expr func_info e)
  | BinOp (b, l, r, _) ->
      VarExpr.binop (check_binop b error) (compile_expr func_info l) (compile_expr func_info r)
  | CastE (_, e) -> compile_expr func_info e
  | AddrOf _ | StartOf _ | SizeOf _ | SizeOfE _ | SizeOfStr _ | AlignOf _ | AlignOfE _ ->
      error_at expr.eloc Cil_printer.pp_exp expr "Unsupported expression construct"
let single_succ l =
  match l with
  | [ s ] -> s
  | _ -> failwith "expected single value"

let double_succ l =
  match l with
  | [ a; b ] -> (a, b)
  | _ -> failwith "expected double value"

module LocMapF = LocMap.WithForeign (LocMap.BaseMap)

(** Simplify the starting graph a bit, any node with a single "if 1" predecessor
    is merged with its predecessor (except the last node). *)
let simplify_gotos preds =
  let to_simplify =
    LocMap.fold
      (fun loc preds to_simplify ->
        match preds with
        | [ (x, l) ] when x = goto ->
            if (info_of_loc loc).pos = pgrm_end then to_simplify
            else
              (* Utils.debug_printf "Parsing: Removing %a (child of %a)@." Loc.pp loc Loc.pp l; *)
              LocMap.add loc l to_simplify
        | _ -> to_simplify)
      preds LocMap.empty
  in
  let rec rename i =
    match LocMap.find i to_simplify with
    | j -> rename j
    | exception Not_found -> i
  in
  let preds =
    LocMapF.update_multiple_from_inter_with_foreign to_simplify { f = (fun _ _ _ -> None) } preds
  in
  LocMap.map (List.map (fun (r, i) -> (r, rename i))) preds

module SMap = Map.Make (String)

let vars = ref VarSet.empty
let funcs = ref SMap.empty

let minimal info x y =
  if x.func_id = info.id then
    match y with
    | None -> Some x
    | Some y when x.pos < y.pos -> Some x
    | _ -> y
  else y

let min_point func_info locmap =
  LocMap.List.fold
    (fun k (_, k') min ->
      minimal func_info (info_of_loc k') (minimal func_info (info_of_loc k) min))
    locmap None

let add_edge pre rel post map = LocMap.List.add post (rel, pre) map

let change_id x id =
  let z = loc_of_info { (info_of_loc x) with call_id = id } in
  (* Utils.debug_printf "Parsing: Renaming %a to %a@." Loc.pp x Loc.pp z; *)
  z

let rec compile_instr func_info pre_id post_id map (instr : instr) =
  (* Most instruction are a simple edge from pre_id to post_id *)
  let edge rel = add_edge pre_id rel post_id map in
  match instr with
  | Local_init (v, AssignInit (SingleInit e), _) ->
      edge (Assign (var_name func_info v, compile_expr func_info e))
  | Set (v, e, _) ->
      edge
        (Assign
           (compile_lval func_info v (error Cil_printer.pp_instr instr), compile_expr func_info e))
  | Code_annot (_, _) | Skip _ -> edge goto
  | Call (res, f, args, _) -> (
      match f.enode with
      | Lval (Var v, _) when v.vname = "__FC_assert" ->
          edge (Guard (compile_expr func_info (List.hd args)))
      | Lval (Var v, _) ->
          let func, info, fundec = compile_func_by_name v.vname in
          (* Utils.debug_printf "Parsing: Inlining %s at site %d@." v.vname info.call_nb; *)
          (* Chain of assigment for the arguments *)
          let pos, map, _ =
            List.fold_left2
              (fun (pos, map, n) v expr ->
                let next_id =
                  loc_of_info
                    { pre_call = true; pos = n; func_id = info.id; call_id = info.call_nb }
                in
                let map =
                  add_edge pos (Assign (var_name info v, compile_expr func_info expr)) next_id map
                in
                (next_id, map, n + 1))
              (pre_id, map, 1) fundec.sformals args
          in
          (* Goto from the chain to function start *)
          let map =
            add_edge pos goto
              (loc_of_info
                 { pre_call = false; pos = pgrm_start; func_id = info.id; call_id = info.call_nb })
              map
          in
          (* Function body *)
          let map =
            LocMap.List.fold
              (fun post (rel, pre) map ->
                add_edge (change_id pre info.call_nb) rel (change_id post info.call_nb) map)
              func map
          in
          (* Return *)
          let function_end =
            loc_of_info
              { pre_call = false; pos = pgrm_end; func_id = info.id; call_id = info.call_nb }
          in
          let map =
            match res with
            | None -> add_edge function_end goto post_id map
            | Some v ->
                add_edge function_end
                  (Assign
                     ( compile_lval func_info v (error Cil_printer.pp_instr instr),
                       VarExpr.var (var_name' info "__retres") ))
                  post_id map
          in

          info.call_nb <- info.call_nb + 1;
          map
      | _ ->
          error Cil_printer.pp_instr instr
            "Unsupported function call. Only calls to __FC_assert are supported")
  | Asm (_, _, _, _) | Local_init _ -> error Cil_printer.pp_instr instr "Unsupported instruction"

and compile_statement func_info map (stmt : stmt) =
  (* Format.printf "Examining %a = %a:@.  preds = %a@.  succs = %a@." Loc.pp stmt.sid
     Cil_printer.pp_stmt stmt
     (Format.pp_print_list ~pp_sep:Utils.pp_semicolon Loc.pp)
     (List.map (fun x -> x.sid) stmt.preds)
     (Format.pp_print_list ~pp_sep:Utils.pp_semicolon Loc.pp)
     (List.map (fun x -> x.sid) stmt.succs); *)
  let id pos =
    loc_of_info { func_id = func_info.id; call_id = func_info.call_nb; pos; pre_call = false }
  in
  let stmt_id = id stmt.sid in
  match stmt.skind with
  | Instr i -> compile_instr func_info stmt_id (id (single_succ stmt.succs).sid) map i
  | Return (_, _) ->
      add_edge stmt_id goto
        (loc_of_info
           { pos = pgrm_end; func_id = func_info.id; call_id = func_info.call_nb; pre_call = false })
        map
  | Goto (_, _) | Break _ | Continue _ ->
      add_edge stmt_id goto (id (single_succ stmt.succs).sid) map
  | If (exp, if_, else_, _) ->
      let rel = compile_expr func_info exp in
      let i, e = double_succ stmt.succs in
      let map = add_edge stmt_id (Guard rel) (id i.sid) map in
      let map = add_edge stmt_id (Guard (VarExpr.unop LNot rel)) (id e.sid) map in
      let map = compile_block func_info map if_.bstmts in
      compile_block func_info map else_.bstmts
  | Loop (_, b, _, _, _) | Block b ->
      let map = add_edge stmt_id goto (id (single_succ stmt.succs).sid) map in
      compile_block func_info map b.bstmts
  | UnspecifiedSequence seq ->
      let map = add_edge stmt_id goto (id (single_succ stmt.succs).sid) map in
      compile_block func_info map (List.map (fun (stmt, _, _, _, _) -> stmt) seq)
  | Switch (_, _, _, _)
  | Throw (_, _)
  | TryCatch (_, _, _)
  | TryFinally (_, _, _)
  | TryExcept (_, _, _, _) ->
      error Cil_printer.pp_stmt stmt "Unsupported instruction"

and compile_block func_info map stmt_list =
  List.fold_left (compile_statement func_info) map stmt_list

and compile_func func_info (fn : fundec) =
  let preds = compile_block func_info LocMap.empty fn.sbody.bstmts in
  let min =
    match min_point func_info preds with
    | None -> simple_error "Empty function"
    | Some x ->
        Utils.debug_printf "Found min: %a@." Loc.pp (loc_of_info x);
        x
  in
  let preds =
    if min.pos = pgrm_start then preds
    else add_edge (loc_of_info { min with pos = pgrm_start }) goto (loc_of_info min) preds
  in
  let preds = simplify_gotos preds in
  funcs := SMap.add fn.svar.vname (preds, func_info, fn) !funcs;
  List.iter (fun v -> vars := VarSet.add (var_name func_info v) !vars) fn.sformals;
  List.iter (fun v -> vars := VarSet.add (var_name func_info v) !vars) fn.slocals;
  preds

and compile_func_by_name name =
  match SMap.find_opt name !funcs with
  | Some x -> x
  | None ->
      let fundec =
        match
          List.find_map
            (function
              | GFun (x, _) when x.svar.vname = name -> Some x
              | _ -> None)
            (Frama_c_kernel.Ast.get ()).globals
        with
        | None -> simple_error ("Can't find definition for function " ^ name)
        | Some x -> x
      in
      let fun_info =
        match find_func_by_name name with
        | Some _ ->
            simple_error
              ("Recursive definition are unsupported: function " ^ name
             ^ " calls itself (possibly indirectly)")
        | None -> { name; call_nb = 0; id = List.length !func_names }
      in
      func_names := fun_info :: !func_names;
      (* Utils.debug_printf "Parsing: Compiling %s with id %d@." name fun_info.id; *)
      (compile_func fun_info fundec, fun_info, fundec)

let get_graph () =
  let { globals; _ } = Frama_c_kernel.Ast.get () in
  match
    List.find_map
      (function
        | GFun (x, _) -> Some x.svar.vname
        | _ -> None)
      (List.rev globals)
  with
  | None -> simple_error "Expected a c file containing a function"
  | Some x ->
      let preds, _, _ = compile_func_by_name x in
      { preds; name = x; starting_vars = List.of_seq (VarSet.to_seq !vars) }

(** {2 To dotty}                                                              *)
(******************************************************************************)

let guard_arrow_style fmt () = Format.fprintf fmt ", color=\"#00AA00\", penwidth=2"
let assign_arrow_style fmt () = Format.fprintf fmt ", style=dashed, color=blue, penwidth=3"

let relation_arrow_style fmt = function
  | Assign _ -> assign_arrow_style fmt ()
  | Guard _ -> guard_arrow_style fmt ()

let graph_dotty_printer fmt (graph : graph) =
  Format.fprintf fmt
    "digraph A {@\n\
    \  label = \"Input graph\\nGuard arrows in green\\nAssign arrows in dashed blue\\n\";@\n\
    \  labelloc = \"t\";@\n";
  Format.fprintf fmt "  @[\"%a\" [shape = \"diamond\"];@]@\n" Loc.pp pgrm_start;
  LocMap.List.iter
    (fun dst (label, src) ->
      Format.fprintf fmt "  @[<h>\"%a\" -> \"%a\" [label=\"%a\"%a];@]@\n" Loc.pp src Loc.pp dst
        Relation.pp label relation_arrow_style label)
    graph.preds;
  (* Format.fprintf fmt "@]@\n}@." *)
  Format.fprintf fmt "}@."
