(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Abstract syntax for our programs: a control flow graph with edges annotated by
    {!relation}, which are either {!Guard} or {!Assign}. This also contains SSA program
    variable and expressions, as well as {!get_graph}, the function which translated
    frama-c's Cil representation to our {!graph} representation. *)

open Std_extra
open Types

(** {1 Program Locations}                                *)
(*********************************************************)

module Loc : TYPE
module LocSet : PatriciaTree.Set_S with type elt = Loc.t
module LocMap : PatriciaTree.Map_S with type key = Loc.t

(* Initial and final locations *)
val pgrm_start : Loc.t
val pgrm_end : Loc.t

(** {1 Program Variables and expressions}                *)
(*********************************************************)

module Var : TYPE
module VarMap : PatriciaTree.Map_S with type key = Var.t
module VarSet : PatriciaTree.Set_S with type elt = Var.t

val tmp_var : int -> Var.t

module VarExpr : Expr.S with type var = Var.t

(** {1 SSA Variables, Expression and Bindings}           *)
(*********************************************************)

module SSAVar : TYPE
module SSAVarMap : PatriciaTree.Map_S with type key = SSAVar.t

val ssavar_init : Var.t -> SSAVar.t
val ssavar_join : Var.t -> int -> SSAVar.t
val ssavar_rename : SSAVar.t -> old:int -> int -> SSAVar.t

module SSAExpr : Expr.S with type var = SSAVar.t

module Bindings : TYPE with type t = SSAExpr.t SSAVarMap.t
module MkBindings (T : TYPE) : TYPE with type t = T.t * Bindings.t

(** {1 Relations between program points}                 *)
(*********************************************************)

type relation =
  | Guard of Var.t Expr.expr
  | Assign of Var.t * Var.t Expr.expr

val expr_of_rel : relation -> Var.t Expr.expr

module Relation : TYPE with type t = relation

(** {1 Program graph}                                    *)
(*********************************************************)

type graph = {
  name : string;
  starting_vars : Var.t list;
  preds : (relation * Loc.t) list LocMap.t;
}

val graph_pp : Format.formatter -> graph -> unit

val co_graph : ('a * Loc.t) list LocMap.t -> ('a * Loc.t) list LocMap.t
(** Flip the edges of the given graph *)

val count_assertions : (relation * Loc.t) list LocMap.t -> int
(** Counts assertion in given graph. An assertion is simply a node with no
    a single guarded child (or two children whose guards aren't "c" and "!c") *)

val get_graph : unit -> graph
(** Get the graph from frama-c's parser *)

val graph_dotty_printer : Format.formatter -> graph -> unit
(** Print the graph in dotty format *)

(** {2 Styles for dotty arrows} *)

val guard_arrow_style : unit pp
val assign_arrow_style : unit pp
val relation_arrow_style : relation pp
