(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Single_value_abstraction.Sig

module L = Domains.Lift

let out_prefix = "/home/opam/results/"
let out_graph_file = out_prefix ^ "output.dot"
let in_graph_file = out_prefix ^ "input.dot"

type display =
  | InputCFG
  | InputGraph
  | WTO
  | WideningEdges
  | ConvergencePoint
  | OutCFG
  | OutGraph
  | Summary
  | CSV

let display =
  [
    (InputCFG, "InputCFG", "The initial program CFG parsed by Frama-C (may be very long)");
    (InputGraph, "InputGraph", "Write initial program CFG in dot format to " ^ in_graph_file);
    (WTO, "WTO", "The weak-topological order");
    (WideningEdges, "WideningEdges", "The set of widening edges used");
    (ConvergencePoint, "ConvergencePoint", "The convergence point p# (may be very long)");
    (OutCFG, "OutCFG", "The newly generated graph (if using an analysis that generates graphs)");
    (OutGraph, "OutGraph", "Write the newly generated graph in dot format to " ^ out_graph_file);
    (Summary, "Summary", "A brief summary of the analysis");
    (CSV, "CSV", "A CSV version of summary, overrides all other options");
  ]

let default_display x = x != CSV

type arguments = {
  enabled : bool;
  iterative_strategy : bool;
  program : Ast.Program.graph;
  numeric : (module SingleValueAbstraction);
  domain : Domain_builder.classical_domain;
  nb_backwards_iterations : int;
  display : display -> bool;
}

let help_msg = "Simple abstract interpretor using the frama-c parser"

let name = "tai"
let prefix = "-" ^ name ^ "-"

module Self = Frama_c_kernel.Plugin.Register (struct
  let name = name
  let shortname = name
  let help = help_msg
end)

module Enabled = Self.False (struct
  let option_name = "-" ^ name
  let help = "Enable tai"
end)

(* let programs =
     let open Ast.Program in
     List.map (fun x -> (x.name, x)) programs

   let format (p : Ast.Program.graph) = Format.sprintf "$(b,%s) (%s)" (Arg.doc_quote p.name) p.doc

   let program =
     let enum = programs in
     let doc =
       "The program to analyse. $(docv) should be one of "
       ^ String.concat "; " (List.map format Ast.Program.programs)
     in
     let info = Arg.info [ "p"; "program" ] ~doc ~docv:"PROGRAM" in
     Arg.value (Arg.opt (Arg.enum enum) (List.hd Ast.Program.programs) info) *)

module Domains = Self.String (struct
  let option_name = prefix ^ "domain"
  let arg_name = "DOMAIN"
  let default = "N *b LiftGVN(N *l FA)"
  let help =
    "The domain to use for the analysis.\n" ^ arg_name
    ^ " is an expression describing the domain. See doc for grammar."
end)

module IterativeStrategy = Self.False (struct
  let option_name = prefix ^ "iterative-strategy"
  let help =
    "Perform analysis using iterative strategy instead of recursive one (stabilize outer loops \
     first)"
end)

(* module Output = Self.String (struct
     let option_name = prefix ^ "out"
     let arg_name = "FILE.dot"
     let default = "result.dot"
     let help = "Output file where the generated graph (if any) is written"
   end) *)

module NbIters = Self.Int (struct
  let option_name = prefix ^ "nb-back-iter"
  let arg_name = "N"
  let default = 3
  let help =
    "Number of iteration when back-propagating information on numeric domains. Default="
    ^ string_of_int default
end)

let format (_, (module A : SingleValueAbstraction)) = Format.sprintf "%s: %s" A.id A.doc

module SVA = Self.String (struct
  let option_name = prefix ^ "sva"
  let arg_name = "SVA"
  let default = fst (List.hd Domain_builder.svas)
  let help =
    "The Single Value Abstraction domain to use for the analysis. " ^ arg_name
    ^ " should be one of:\n- "
    ^ String.concat "\n- " (List.map format Domain_builder.svas)
    ^ "\n\nDefault: " ^ default
end)
let () = SVA.set_possible_values (List.map fst Domain_builder.svas)

let format (_, name, doc) = Format.sprintf "%s: %s" name doc
let display_assoc = List.map (fun (d, n, _) -> (d, n)) display
module Display = Self.String_set (struct
  let option_name = prefix ^ "display"
  let arg_name = "DISPLAY"
  let help =
    "Choose what to output, accepts a comma separated list of the following:\n- "
    ^ String.concat "\n- " (List.map format display)
    ^ "\n\nDefaults to all except CSV."
end)

module UseColor = Self.True (struct
  let option_name = prefix ^ "color"
  let help = "Use ANSI codes to color output."
end)

module StepByStep = Self.False (struct
  let option_name = prefix ^ "step-by-step"
  let help = "Print and pause at each fixpoint computation step (press enter to resume)"
end)

let get_args () =
  let enabled = Enabled.get () in
  if UseColor.is_set () then Format.printf "using color %b@." (UseColor.get ());
  Std_extra.Utils.use_color := UseColor.is_set ();
  (* let output = Output.get () in *)
  let domain = Domains.get () in
  let lexbuf = Lexing.from_string domain in
  match Cli_parser.full_expr Cli_lexer.tokens lexbuf with
  | exception Cli_parser.Error ->
      Std_extra.Utils.printf "@{<bold>@{<red>PARSING ERROR@}: invalid value for -tai-domain@}@.";
      exit 8
  | domain ->
      let sva = List.assoc (SVA.get ()) Domain_builder.svas in
      let iterative_strategy = IterativeStrategy.get () in
      let program = Ast.Program.get_graph () in
      let nb_backwards_iterations = NbIters.get () in
      let display =
        if Display.is_empty () then default_display
        else fun x -> Display.mem (List.assoc x display_assoc)
      in
      Std_extra.Utils.set_debug (StepByStep.get ());
      {
        enabled;
        iterative_strategy;
        program;
        domain;
        nb_backwards_iterations;
        numeric = sva;
        display;
      }
