(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** {1 Domain AST}                                                            *)

(******************************************************************************)
(** Simple syntax tree used to represent the domain used. Useful to specify
    the analysis from the command line via the -tai-domain argument.
    See parser.mly for the parser that creates this tree. *)

type domain_id = string

(** The three types of product, matching the products in {!Domains.Product} *)
type product =
  | Basic
  | Left
  | Both

(** Syntax tree for an SSA domain (i.e. a domain that appears under the
    {!Domains.Lift} functor *)
type ssa_domain =
  | SSA_Domain of domain_id
  | SSA_Functor of domain_id * ssa_domain
  | SSA_Product of ssa_domain * ssa_domain
  | SSA_Numeric of domain_id * domain_id

(** Syntax tree for a classical/IMP (top-level) domain. *)
type classical_domain =
  | Classical_Domain of domain_id
  | Classical_Functor of domain_id * classical_domain
  | Classical_Product of classical_domain * classical_domain
  | Classical_Numeric of domain_id * domain_id
  | Lift of bool * ssa_domain

(** {1 Domains}                                                               *)

(******************************************************************************)
(** List of all domains from the {!Domains} library and single value abstractions
    from the {!Single_value_abstraction} library                              *)

(** {2 Single value abstractions}                                             *)
(******************************************************************************)

(** All single value abstractions *)
let svas =
  let open Single_value_abstraction in
  List.map
    (fun x ->
      let module T = (val x : Sig.SingleValueAbstraction) in
      (T.id, x))
    [
      (module Interval);
      (module Known_bits);
      (module Sentinel);
      (module Product.Interval_Sentinel);
      (module Product.KnownBits_Sentinel);
      (module Product.Interval_KnownBits_Sentinel);
    ]

(** {2 Functors}                                                              *)
(******************************************************************************)

(** All transformation functors, with their identifier *)
let transformations =
  let open Domains.Transformations in
  List.map
    (fun x ->
      let module T = (val x : Domains.Sig.Simple_Transformation) in
      (T.id, x))
    [
      (module SimplifyExpression);
      (module DivGuard);
      (module ShiftGuard);
      (module Ternary);
      (module Query);
    ]

(** All SSA transformation functors, with their identifier *)
let ssa_transformations =
  let open Domains.Transformations in
  List.map
    (fun x ->
      let module T = (val x : Domains.Sig.SSA_Transformation) in
      (T.id, x))
    [ (module SSA_SimplifyExpression); (module SSA_Query) ]

(** {1 Builder for the full domain}                                           *)
(******************************************************************************)

(** {2 Types for casting down to contained domains}                           *)

open Ast.Program

type ('state, 'sva) numeric = {
  numeric_state : 'state -> 'sva VarMap.t;
  expr_forwards : 'state -> VarExpr.t -> 'sva;
}
(** Used to cast to a contained IMP Numeric domain *)

type ('state, 'sva) simple_ssa_numeric = {
  ssa_numeric_state : 'state -> 'sva SSAExpr.Map.t;
  ssa_expr_forwards : 'state -> SSAExpr.t -> 'sva;
}
(** Used to cast to a contained SSA Numeric domain from an SSA Domain *)

type ('state, 'sva) ssa_numeric = {
  ssa_numeric_state : 'state -> 'sva SSAExpr.Map.t;
  ssa_expr_forwards : 'state -> SSAExpr.t -> 'sva;
  ssa_bind : 'state -> VarExpr.t -> SSAExpr.t;
  var_store : 'state -> SSAExpr.t VarMap.t;
}
(** Used to cast to a contained SSA Numeric domain from an IMP Domain *)

(** Used to cast to a contained free algebra to generate the graph *)
module type GRAPH_WITH_CAST = sig
  include Domains.Sig.Graph
  type full_state
  val cast : full_state -> state
end

(** {2 Builder module}                                                        *)

(** The builder is parameterized by the default single value abstraction,
    used when unspecified in numeric domains. This default is the only one
    usable for numeric comparisons. *)
module Builder (Zs : Single_value_abstraction.Sig.SingleValueAbstraction) = struct
  (** A fully build domain: it contains
      - an instance of {!Domains.Sig.Classical_Domain} to use for the analysis
      - some tools to analyse the result further:
        - an optional graph module to build the graph if this domain contains a
          {!Domains.Free_algebra} domain
        - optional [to_numeric] and [to_ssa_numeric] values which cast to the
          contained {!Domains.Numeric} domains. *)
  module type BUILT_DOMAIN = sig
    include Domains.Sig.Classical_Domain with type relation = relation

    val graph : (module GRAPH_WITH_CAST with type full_state = State.t) option

    val to_numeric : (State.t, Zs.t) numeric option

    val to_ssa_numeric : (State.t, Zs.t) ssa_numeric option
  end

  (** SSA Variant of the {!BUILT_DOMAIN} *)
  module type SSA_BUILT_DOMAIN = sig
    include Domains.Sig.SSA_Domain

    val graph : (module GRAPH_WITH_CAST with type full_state = SSA_State.t) option

    val to_ssa_numeric : (SSA_State.t, Zs.t) simple_ssa_numeric option
  end

  let error message =
    Std_extra.Utils.printf "@{<bold>@{<red>ERROR@}: invalid value for -tai-domain:@.  %s@}@."
      message;
    exit 9

  let all_ids x =
    Format.asprintf "'%a'"
      (Format.pp_print_list
         ~pp_sep:(fun fmt () -> Format.fprintf fmt "', '")
         (fun fmt (id, _) -> Format.fprintf fmt "%s" id))
      x

  (* let prod_id = function
     | Basic -> "*"
     | Left -> " *l "
     | Both -> " *b " *)

  module SSA_N = struct
    include Domains.Numeric.SSA (Zs)
    let id = id ^ "[" ^ Zs.id ^ "]"

    let graph = None
    let to_ssa_numeric =
      Some { ssa_numeric_state = (fun x -> x); ssa_expr_forwards = expr_forwards }
  end

  module SSA_QN = struct
    include Domains.Numeric.SSAQuery (Zs)
    let id = id ^ "[" ^ Zs.id ^ "]"

    let graph = None
    let to_ssa_numeric =
      Some { ssa_numeric_state = (fun x -> x); ssa_expr_forwards = expr_forwards }
  end

  let rec build_ssa domain : (module SSA_BUILT_DOMAIN) =
    match domain with
    | SSA_Domain "FA" ->
        (module struct
          include Domains.Free_algebra.SSA

          let to_ssa_numeric = None
          let graph =
            Some
              (module struct
                include Graph
                type full_state = SSA_State.t
                let cast x = x
              end : GRAPH_WITH_CAST
                with type full_state = SSA_State.t)
        end)
    | SSA_Domain "N" -> (module SSA_N)
    | SSA_Numeric ("N", x) when x = Zs.id -> (module SSA_N)
    | SSA_Numeric ("N", x) -> (
        match List.assoc_opt x svas with
        | None ->
            error
              ("Invalid single value abstraction '" ^ x ^ "'.\n  Valid values are " ^ all_ids svas)
        | Some (module SVA) ->
            (module struct
              include Domains.Numeric.SSA (SVA)
              let id = id ^ "[" ^ SVA.id ^ "]"

              let graph = None
              let to_ssa_numeric = None
            end))
    | SSA_Domain "QN" -> (module SSA_QN)
    | SSA_Numeric ("QN", x) when x = Zs.id -> (module SSA_QN)
    | SSA_Numeric ("QN", x) -> (
        match List.assoc_opt x svas with
        | None ->
            error
              ("Invalid single value abstraction '" ^ x ^ "'.\n  Valid values are " ^ all_ids svas)
        | Some (module SVA) ->
            (module struct
              include Domains.Numeric.SSAQuery (SVA)
              let id = id ^ "[" ^ SVA.id ^ "]"

              let graph = None
              let to_ssa_numeric = None
            end))
    | SSA_Domain id ->
        error ("Invalid SSA domain '" ^ id ^ "'.\n  Valid values are 'FA', 'N' or 'QN'")
    | SSA_Numeric (id, _) ->
        error ("Invalid SSA numeric domain '" ^ id ^ "'.\n  Valid values are 'N' or 'QN'")
    | SSA_Product (left, right) ->
        let (module Left : SSA_BUILT_DOMAIN) = build_ssa left in
        let (module Right) = build_ssa right in
        (module struct
          include Domains.Product.SSA (Left) (Right)
          let id =
            let basic_id = Left.id ^ "*" ^ Right.id in
            match left with
            | SSA_Product _ -> "(" ^ basic_id ^ ")"
            | _ -> basic_id
          let graph =
            match (Left.graph, Right.graph) with
            | Some (module G), _ ->
                Some
                  (module struct
                    include G
                    type full_state = SSA_State.t
                    let cast x = G.cast (fst x)
                  end : GRAPH_WITH_CAST
                    with type full_state = SSA_State.t)
            | _, Some (module G) ->
                Some
                  (module struct
                    include G
                    type full_state = SSA_State.t
                    let cast x = G.cast (snd x)
                  end)
            | None, None -> None

          let to_ssa_numeric =
            match (Left.to_ssa_numeric, Right.to_ssa_numeric) with
            | Some num, _ ->
                Some
                  {
                    ssa_numeric_state = (fun (x, _) -> num.ssa_numeric_state x);
                    ssa_expr_forwards = (fun (x, _) -> num.ssa_expr_forwards x);
                  }
            | _, Some num ->
                Some
                  {
                    ssa_numeric_state = (fun (_, x) -> num.ssa_numeric_state x);
                    ssa_expr_forwards = (fun (_, x) -> num.ssa_expr_forwards x);
                  }
            | None, None -> None
        end)
    | SSA_Functor (func, domain) -> (
        let (module Base) = build_ssa domain in
        match List.assoc_opt func ssa_transformations with
        | None ->
            error
              ("Invalid SSA functor '" ^ func ^ "'.\n  Valid values are "
             ^ all_ids ssa_transformations)
        | Some (module Functor) ->
            (module struct
              include Functor.Domain (Base)

              let id = Functor.id ^ "(" ^ Base.id ^ ")"

              let graph = Base.graph
              let to_ssa_numeric = Base.to_ssa_numeric
            end))

  module N = struct
    include Domains.Numeric.Classical (Zs)
    let id = id ^ "[" ^ Zs.id ^ "]"

    let graph = None
    let to_numeric = Some { numeric_state = (fun x -> x); expr_forwards }
    let to_ssa_numeric = None
  end

  module QN = struct
    include Domains.Numeric.ClassicalQuery (Zs)
    let id = id ^ "[" ^ Zs.id ^ "]"

    let graph = None
    let to_numeric = Some { numeric_state = (fun x -> x); expr_forwards }
    let to_ssa_numeric = None
  end

  let rec build_classical domain : (module BUILT_DOMAIN) =
    match domain with
    | Classical_Domain "FA" ->
        (module struct
          include Domains.Free_algebra.Classical
          let to_numeric = None
          let to_ssa_numeric = None
          let graph =
            Some
              (module struct
                include Graph
                type full_state = State.t
                let cast x = x
              end : GRAPH_WITH_CAST
                with type full_state = State.t)
        end)
    | Classical_Domain "N" -> (module N)
    | Classical_Numeric ("N", x) when x = Zs.id -> (module N)
    | Classical_Numeric ("N", x) -> (
        match List.assoc_opt x svas with
        | None ->
            error
              ("Invalid single value abstraction '" ^ x ^ "'.\n  Valid values are " ^ all_ids svas)
        | Some (module SVA) ->
            (module struct
              include Domains.Numeric.Classical (SVA)
              let id = id ^ "[" ^ SVA.id ^ "]"

              let graph = None
              let to_numeric = None
              let to_ssa_numeric = None
            end))
    | Classical_Domain "QN" -> (module QN)
    | Classical_Numeric ("QN", x) when x = Zs.id -> (module QN)
    | Classical_Numeric ("QN", x) -> (
        match List.assoc_opt x svas with
        | None ->
            error
              ("Invalid single value abstraction '" ^ x ^ "'.\n  Valid values are " ^ all_ids svas)
        | Some (module SVA) ->
            (module struct
              include Domains.Numeric.ClassicalQuery (SVA)
              let id = id ^ "[" ^ SVA.id ^ "]"

              let graph = None
              let to_numeric = None
              let to_ssa_numeric = None
            end))
    | Classical_Domain id ->
        error ("Invalid classical domain '" ^ id ^ "'.\n  Valid values are 'FA', 'N' or 'QN'")
    | Classical_Numeric (id, _) ->
        error ("Invalid classical numeric domain '" ^ id ^ "'.\n  Valid values are 'N' or 'QN'")
    | Classical_Product (left, right) ->
        let (module Left : BUILT_DOMAIN) = build_classical left in
        let (module Right) = build_classical right in
        (module struct
          include Domains.Product.Classical (Left) (Right)
          let id =
            let basic_id = Left.id ^ "*" ^ Right.id in
            match left with
            | Classical_Product _ -> "(" ^ basic_id ^ ")"
            | _ -> basic_id

          let graph =
            match (Left.graph, Right.graph) with
            | Some (module G), _ ->
                Some
                  (module struct
                    include G
                    type full_state = State.t
                    let cast x = G.cast (fst x)
                  end : GRAPH_WITH_CAST
                    with type full_state = State.t)
            | _, Some (module G) ->
                Some
                  (module struct
                    include G
                    type full_state = State.t
                    let cast x = G.cast (snd x)
                  end : GRAPH_WITH_CAST
                    with type full_state = State.t)
            | None, None -> None

          let to_numeric =
            match (Left.to_numeric, Right.to_numeric) with
            | Some num, _ ->
                Some
                  {
                    numeric_state = (fun (x, _) -> num.numeric_state x);
                    expr_forwards = (fun (x, _) -> num.expr_forwards x);
                  }
            | _, Some num ->
                Some
                  {
                    numeric_state = (fun (_, x) -> num.numeric_state x);
                    expr_forwards = (fun (_, x) -> num.expr_forwards x);
                  }
            | None, None -> None

          let to_ssa_numeric =
            match (Left.to_ssa_numeric, Right.to_ssa_numeric) with
            | Some num, _ ->
                Some
                  {
                    ssa_numeric_state = (fun (x, _) -> num.ssa_numeric_state x);
                    ssa_expr_forwards = (fun (x, _) -> num.ssa_expr_forwards x);
                    ssa_bind = (fun (x, _) -> num.ssa_bind x);
                    var_store = (fun (x, _) -> num.var_store x);
                  }
            | _, Some num ->
                Some
                  {
                    ssa_numeric_state = (fun (_, x) -> num.ssa_numeric_state x);
                    ssa_expr_forwards = (fun (_, x) -> num.ssa_expr_forwards x);
                    ssa_bind = (fun (_, x) -> num.ssa_bind x);
                    var_store = (fun (_, x) -> num.var_store x);
                  }
            | None, None -> None
        end)
    | Classical_Functor (func, domain) -> (
        let (module Base) = build_classical domain in
        match List.assoc_opt func transformations with
        | None ->
            error
              ("Invalid classical functor '" ^ func ^ "'.\n  Valid values are "
             ^ all_ids transformations)
        | Some (module Functor) ->
            (module struct
              include Functor.Domain (Base)

              let id = Functor.id ^ "(" ^ Base.id ^ ")"

              let graph = Base.graph
              let to_numeric = Base.to_numeric
              let to_ssa_numeric = Base.to_ssa_numeric
            end))
    | Lift (gvn, domain) ->
        let (module SSA) = build_ssa domain in
        (module struct
          include
            Domains.Lift.SSA_Lift
              (SSA)
              (struct
                let do_gvn = gvn
              end)

          let id = id ^ "(" ^ SSA.id ^ ")"

          let graph =
            match SSA.graph with
            | Some (module G) ->
                Some
                  (module struct
                    include G
                    type full_state = State.t
                    let cast (x : State.t) = cast x.ssa_state
                  end : GRAPH_WITH_CAST
                    with type full_state = State.t)
            | None -> None
          let to_numeric = None
          let to_ssa_numeric =
            match SSA.to_ssa_numeric with
            | None -> None
            | Some num ->
                Some
                  {
                    ssa_bind = bind;
                    var_store = (fun x -> x.var_store);
                    ssa_expr_forwards = (fun (x : State.t) -> num.ssa_expr_forwards x.ssa_state);
                    ssa_numeric_state = (fun (x : State.t) -> num.ssa_numeric_state x.ssa_state);
                  }
        end)
end
