(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Builds the domain used for the analysis, as specified by the expression
    given to [-tai-domain]. *)

(** {1 Domain AST}                                                            *)

(******************************************************************************)
(** Simple syntax tree used to represent the domain used. Useful to specify
    the analysis from the command line via the -tai-domain argument.
    See parser.mly for the parser that creates this tree. *)

type domain_id = string

(** The three types of product, matching the products in {!Domains.Product} *)
type product =
  | Basic
  | Left
  | Both

(** Syntax tree for an SSA domain (i.e. a domain that appears under the
    {!Domains.Lift} functor *)
type ssa_domain =
  | SSA_Domain of domain_id
  | SSA_Functor of domain_id * ssa_domain
  | SSA_Product of ssa_domain * ssa_domain
  | SSA_Numeric of domain_id * domain_id

(** Syntax tree for a classical/IMP (top-level) domain. *)
type classical_domain =
  | Classical_Domain of domain_id
  | Classical_Functor of domain_id * classical_domain
  | Classical_Product of classical_domain * classical_domain
  | Classical_Numeric of domain_id * domain_id
  | Lift of bool * ssa_domain

val svas : (string * (module Single_value_abstraction.Sig.SingleValueAbstraction)) list
(**  All single value abstractions from the {!Single_value_abstraction} library *)

(** {1 Builder for the full domain}                                           *)
(******************************************************************************)

(** {2 Types for casting down to contained domains}                           *)

type ('state, 'sva) numeric = {
  numeric_state : 'state -> 'sva Ast.Program.VarMap.t;
  expr_forwards : 'state -> Ast.Program.VarExpr.t -> 'sva;
}
(** Used to cast to a contained IMP Numeric domain *)

type ('state, 'sva) ssa_numeric = {
  ssa_numeric_state : 'state -> 'sva Ast.Program.SSAExpr.Map.t;
  ssa_expr_forwards : 'state -> Ast.Program.SSAExpr.t -> 'sva;
  ssa_bind : 'state -> Ast.Program.VarExpr.t -> Ast.Program.SSAExpr.t;
  var_store : 'state -> Ast.Program.SSAExpr.t Ast.Program.VarMap.t;
}
(** Used to cast to a contained SSA Numeric domain from an IMP Domain *)

(** Used to cast to a contained free algebra to generate the graph *)
module type GRAPH_WITH_CAST = sig
  include Domains.Sig.Graph
  type full_state
  val cast : full_state -> state
end

(** {2 Builder module}                                                        *)

(** The builder is parameterized by the default single value abstraction,
    used when unspecified in numeric domains. This default is the only one
    usable for numeric comparisons. *)
module Builder (Zs : Single_value_abstraction.Sig.SingleValueAbstraction) : sig
  (** A fully build domain: it contains
      - an instance of {!Domains.Sig.Classical_Domain} to use for the analysis
      - some tools to analyse the result further:
        - an optional graph module to build the graph if this domain contains a
          {!Domains.Free_algebra} domain
        - optional [to_numeric] and [to_ssa_numeric] values which cast to the
          contained {!Domains.Numeric} domains. *)
  module type BUILT_DOMAIN = sig
    include Domains.Sig.Classical_Domain with type relation = Ast.Program.relation

    val graph : (module GRAPH_WITH_CAST with type full_state = State.t) option
    val to_numeric : (State.t, Zs.t) numeric option
    val to_ssa_numeric : (State.t, Zs.t) ssa_numeric option
  end

  val build_classical : classical_domain -> (module BUILT_DOMAIN)
end
