(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Main function of the analyser *)

open Ast.Program
open Std_extra
open Cli

let () =
  Utils.use_unicode := Frama_c_kernel.Kernel.Unicode.get ();
  Frama_c_kernel.Kernel.Unicode.add_set_hook (fun _ x -> Utils.use_unicode := x)

(* By default OCaml uses floats to record dates, which isn't the most precise.
   This simple C API does not have this defect. *)
external bench_record_time : unit -> unit = "record_time"
external bench_return_time : unit -> int = "caml_return_time"

let print_dot_file file printer arg =
  (* Before printing to file:
     - Disable unicode, dotty doesn't like it
     - Set width to very wide, as dotty doesn't like line breaks *)
  let open Frama_c_kernel.Kernel in
  let u = Unicode.get () in
  Unicode.set false;
  let ch = open_out file in
  let fmt = Format.formatter_of_out_channel ch in
  Format.pp_set_geometry fmt ~max_indent:100000 ~margin:200000;
  Format.fprintf fmt "%a" printer arg;
  close_out ch;
  Unicode.set u

type gen_graph = {
  time : int;
  size : unit -> int;
  print : unit -> unit;
  to_dot : string -> unit;
  assertions_and_dead_ends : unit -> int * int;
}

let main args =
  if not args.enabled then ()
  else
    (* ========== Setup analysis ============================================ *)
    let (module SVA) = args.numeric in
    let module Builder = Domain_builder.Builder (SVA) in
    let (module Domain) = Builder.build_classical args.domain in
    let module Fixpoint = Wto.Fixpoint (Domain) in
    let () = Domains.Numeric.set_nb_backwards_iter args.nb_backwards_iterations in
    let co_graph = co_graph args.program.preds in
    let preds x = Option.value ~default:[] @@ LocMap.find_opt x args.program.preds in
    let succs x = Option.value ~default:[] @@ LocMap.find_opt x co_graph in

    (* ========== Compute Weak Topological Order (WTO) ====================== *)
    bench_record_time ();
    let wto = Wto.partition ~init:pgrm_start ~succs in
    let widening_edges = Wto.widening_edges_preds ~preds wto in
    (* Compute widening edges BEFORE switching to iterative *)
    let wto =
      if args.iterative_strategy then
        (* Utils.printf "@{<bold>Switching to iterative strategy:@}@.%a@.@." Wto.pp wto; *)
        Wto.iterative_strategy wto
      else wto
    in
    let time_wto = bench_return_time () in

    (* ========== Iterate until fixed point ================================= *)
    bench_record_time ();
    let pconv =
      Fixpoint.fixpoint ~preds ~widening_edges ~loc0:pgrm_start
        ~entrypoint:(Domain.entrypoint args.program.starting_vars)
        wto
    in
    let time_conv = bench_return_time () in

    (* =========== Compute new graph ======================================== *)
    (* (Only if the domain used contains one of the free-algebra domains)     *)
    let autom =
      match Domain.graph with
      | None -> None
      | Some (module Graph) ->
          bench_record_time ();
          let autom = Graph.make ~cast:Graph.cast ~join_preds:(Fixpoint.join_preds ~preds) pconv in
          let time_graph = bench_return_time () in
          (* We can't have the graph type escape this match, so we return a record
              with all operations neeeded *)
          Some
            {
              time = time_graph;
              size = (fun () -> Graph.size autom);
              print =
                (fun () ->
                  Utils.printf "@{<bold>Generated graph: %d nodes@}@.%a@.@." (Graph.size autom)
                    Graph.pp autom);
              to_dot =
                (fun file ->
                  print_dot_file file Graph.dotty_printer autom;
                  Utils.printf "@{<bold>Generated graph saved to \"%s\"@}@.@." file);
              assertions_and_dead_ends =
                (fun () -> Graph.assertions_and_dead_ends ~cast:Graph.cast autom pconv);
            }
    in

    (* ========== Compare numerical domains ================================= *)
    (* Only if both classical and SSA numeric domain are present              *)
    let module NumCompare = Num_compare.NumericComparison (SVA) in
    let basic_comp, child_comp =
      match (Domain.to_numeric, Domain.to_ssa_numeric) with
      | Some num, Some ssa_num ->
          ( Some (NumCompare.all_state_all_variables Domain.State.pp num ssa_num pconv),
            Some (NumCompare.children_only num ssa_num pconv ~co_graph) )
      | _ -> (None, None)
    in

    (* ========== Pretty printing the results =============================== *)
    let strategy = if args.iterative_strategy then "iterative" else "recursive" in
    let in_size = LocMap.cardinal args.program.preds in
    let nb_widening_edges = List.length widening_edges in
    let file = Frama_c_kernel.Filepath.relativize (Frama_c_kernel.Kernel.Files.As_string.get ()) in
    let ms = Utils.Unicode.microseconds () in

    let assertions = count_assertions co_graph in
    if args.display Cli.CSV then
      (* Compact easy-to-parse format. Output a single comma separated line.size

         Columns correspond to:
         File,function,in_nodes,in_assertions,Domain,SVA,Transforms,strategy,
         widening_edges,out_nodes,out_assertions,out_dead_ends,
         WTO_time,Conv_time,Graph_time,total_time *)
      match autom with
      | None ->
          Utils.printf "%s,%s,%d,%d,\"%s\",%s,%s,%d,--,--,--,%d,%d,--,%d,%a,%a@." file
            args.program.name in_size assertions Domain.id SVA.id strategy nb_widening_edges
            time_wto time_conv (time_wto + time_conv) Num_compare.pp_csv basic_comp
            Num_compare.pp_csv child_comp
      | Some x ->
          let new_assertions, dead_ends = x.assertions_and_dead_ends () in
          (* File,function,in_nodes,in_assertions,SVA,Domain,Transforms,strategy,
             widening_edges,out_nodes,out_assertions,out_dead_ends,
             WTO_time,Conv_time,Graph_time,total_time,eq_num,le_num,ge_num*)
          Utils.printf "%s,%s,%d,%d,\"%s\",%s,%s,%d,%d,%d,%d,%d,%d,%d,%d,%a,%a@." file
            args.program.name in_size assertions Domain.id SVA.id strategy nb_widening_edges
            (x.size ()) new_assertions dead_ends time_wto time_conv x.time
            (time_wto + time_conv + x.time)
            Num_compare.pp_csv basic_comp Num_compare.pp_csv child_comp
    else (
      if args.display Cli.InputCFG then
        Utils.printf "@.@{<bold>Analyzing program: %s (%d nodes)@}@.%a@.@." args.program.name
          in_size Ast.Program.graph_pp args.program;
      if args.display Cli.InputGraph then (
        print_dot_file Cli.in_graph_file Ast.Program.graph_dotty_printer args.program;
        Utils.printf "@.@{<bold>Input graph saved to \"%s\"@}@.@." Cli.in_graph_file);
      if args.display Cli.WTO then
        Utils.printf "@{<bold>Weak topological ordering (%s mode):@}@.%a@.@." strategy Wto.pp wto;
      if args.display Cli.WideningEdges then (
        Utils.printf "@{<bold>Set of widening edges:@}@.";
        List.iter
          (fun (p, l, c) ->
            Utils.printf "@[<hv>%a@ --{%a}-->@ %a@]@." Loc.pp p Relation.pp l Loc.pp c)
          widening_edges;
        Utils.printf "@.");
      if args.display Cli.ConvergencePoint then
        let map_pp fmt m =
          Format.fprintf fmt "  @[<v>%a@]"
            (LocMap.pp_map
               ~pp_sep:(fun fmt () -> Format.pp_print_break fmt 1 0)
               (fun fmt k v -> Format.fprintf fmt "@[<v>%a: @[<v>%a@]@]" Loc.pp k Domain.State.pp v))
            m
        in
        Utils.printf "@{<bold>Convergence point reached:@}@.%a@.@." map_pp pconv);
    (if args.display Cli.OutCFG then
       match autom with
       | None -> ()
       | Some x -> x.print ());
    (if args.display Cli.OutGraph then
       match autom with
       | None -> ()
       | Some x -> x.to_dot Cli.out_graph_file);
    if args.display Cli.Summary then
      match autom with
      | None ->
          Utils.printf
          "@{<bold>Analysis parameters:@}@.\
          \  Input program: %s : %s (%d nodes; %d assertions)@.\
          \  Domain: %s@.\
          \  Strategy: %s@.@{<bold>Analysis results:@}@.\
          \  Widening edges: %d@.\
          \  No output graph@.\
          \  Numeric state comparison:@.\
          \    All State all variables: %a@.\
          \    Children expressions: %a@.\
          \  WTO time: %d %s@.\
          \  Convergence time: %d %s@.\
          \  Total time: %d %s@." file args.program.name in_size assertions
          Domain.id
          strategy nb_widening_edges
          Num_compare.pp_summary basic_comp Num_compare.pp_summary child_comp
          time_wto ms time_conv ms
          (time_wto + time_conv) ms
          [@ocamlformat "disable"]
      | Some x ->
          let new_assertions, dead_ends = x.assertions_and_dead_ends () in
          Utils.printf
          "@{<bold>Analysis parameters:@}@.\
          \  Input file: %s@.\
          \  Input program: %s (%d nodes; %d assertions)@.\
          \  Domain: %s@.\
          \  Strategy: %s@.@{<bold>Analysis results:@}@.\
          \  Widening edges: %d@.\
          \  Output graph: %d nodes; %d assertions; %d dead ends@.\
          \  Numeric state comparison:@.\
          \    All State all variables: %a@.\
          \    Children expressions: %a@.\
          \  WTO time: %d %s@.\
          \  Convergence time: %d %s@.\
          \  Graph generation time: %d %s@.\
          \  Total time: %d %s@." file args.program.name in_size assertions
          Domain.id
          strategy nb_widening_edges (x.size ()) new_assertions dead_ends
          Num_compare.pp_summary basic_comp
          Num_compare.pp_summary child_comp
          time_wto ms time_conv
          ms x.time ms (time_wto + time_conv + x.time) ms
          [@ocamlformat "disable"]

let run () = main (Cli.get_args ())
let () = Frama_c_kernel.Db.Main.extend run
