(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Std_extra
open Ast.Program

type totals = {
  same_precision : int;
  ssa_more_precise : int;
  classical_more_precise : int;
  incomparable : int;
}

let empty_totals =
  { same_precision = 0; ssa_more_precise = 0; classical_more_precise = 0; incomparable = 0 }

let pp_csv fmt = function
  | None -> Format.fprintf fmt "--,--,--,--"
  | Some totals ->
      Format.fprintf fmt "%d,%d,%d,%d" totals.same_precision totals.ssa_more_precise
        totals.classical_more_precise totals.incomparable

let pp_summary fmt = function
  | None -> Format.fprintf fmt "None"
  | Some totals ->
      Format.fprintf fmt "equal=%d, SSA more precise=%d, SSA less precise=%d, incomparable=%d"
        totals.same_precision totals.ssa_more_precise totals.classical_more_precise
        totals.incomparable

(** Module used to compare the IMP Numeric domain ({!Domains.Numeric.Classical})
    and the SSA one ({!Domains.Numeric.SSA}) *)
module NumericComparison (N : Single_value_abstraction.Sig.SingleValueAbstraction) = struct
  open Domain_builder

  let compare_abstraction ~num_value ~ssa_value totals =
    match (num_value, ssa_value) with
    | None, None -> { totals with same_precision = totals.same_precision + 1 }
    | Some _, None -> { totals with ssa_more_precise = totals.ssa_more_precise + 1 }
    | None, Some ssa_value ->
        Utils.printf "Diff: None and %a@." N.pp ssa_value;
        { totals with classical_more_precise = totals.classical_more_precise + 1 }
    | Some num_value, Some ssa_value ->
        if N.equal num_value ssa_value then
          { totals with same_precision = totals.same_precision + 1 }
        else if N.subseteq ssa_value num_value then
          { totals with ssa_more_precise = totals.ssa_more_precise + 1 }
        else if N.subseteq num_value ssa_value then
          {
            (* Utils.printf "Diff: %a and %a@." N.pp num_value N.pp ssa_value; *)
            totals with
            classical_more_precise = totals.classical_more_precise + 1;
          }
        else
          {
            (* Utils.printf "Incomp: %a and %a@." N.pp num_value N.pp ssa_value; *)
            totals with
            incomparable = totals.incomparable + 1;
          }

  let one_state_all_variables _pos num ssa_num state =
    (* Utils.printf "== %a@." Simple.State.pp (l_state, r_state); *)
    VarMap.fold
      (fun var num_value totals ->
        let expr = VarMap.find var (ssa_num.var_store state) in
        let ssa_value =
          try Some (ssa_num.ssa_expr_forwards state expr) with Division_by_zero -> None
        in
        (* Utils.printf "    %a:%a = %a or %a@." Var.pp v SSAExpr.pp expr N.pp n N.pp res'; *)
        let t' = compare_abstraction ~num_value:(Some num_value) ~ssa_value totals in
        (* if t'.incomparable <> totals.incomparable then
             Utils.printf "Uni Incomparable %a: %a@." Loc.pp pos Var.pp var;
           if t'.classical_more_precise <> totals.classical_more_precise then
             Utils.printf "Uni Classical %a: %a@." Loc.pp pos Var.pp var; *)
        t')
      (num.numeric_state state) empty_totals

  let all_state_all_variables _state_pp num ssa_num pconv =
    LocMap.fold
      (fun _pos state sums ->
        (* Utils.printf "== %a@." Simple.State.pp (l_state, r_state); *)
        VarMap.fold
          (fun var num_value totals ->
            let expr = VarMap.find var (ssa_num.var_store state) in
            let ssa_value =
              try Some (ssa_num.ssa_expr_forwards state expr) with Division_by_zero -> None
            in
            (* Utils.printf "    %a:%a = %a or %a@." Var.pp v SSAExpr.pp expr N.pp n N.pp res'; *)
            let t' = compare_abstraction ~num_value:(Some num_value) ~ssa_value totals in
            (* if t'.incomparable <> totals.incomparable then
                 Utils.printf "Incomparable %a: %a@." Loc.pp _pos Var.pp var;
               if t'.classical_more_precise <> totals.classical_more_precise then
                 Utils.printf "Classical %a: %a@." Loc.pp _pos Var.pp var; *)
            t')
          (num.numeric_state state) sums)
      pconv empty_totals

  let children_only num ssa_num pconv ~co_graph =
    LocMap.List.fold
      (fun parent (relation, _) totals ->
        match LocMap.find_opt parent pconv with
        | None -> totals
        | Some parent_state ->
            let expr = expr_of_rel relation in
            let num_value =
              try Some (num.expr_forwards parent_state expr) with Division_by_zero -> None
            in
            let ssa_value =
              try Some (ssa_num.ssa_expr_forwards parent_state (ssa_num.ssa_bind parent_state expr))
              with Division_by_zero -> None
            in
            let t' = compare_abstraction ~num_value ~ssa_value totals in
            if t'.incomparable <> totals.incomparable then
              Utils.printf "Incomparable %a@." Loc.pp parent;
            if t'.classical_more_precise <> totals.classical_more_precise then
              Utils.printf "Classical %a with %a %a@." Loc.pp parent VarExpr.pp expr SSAExpr.pp
                (ssa_num.ssa_bind parent_state expr);
            t')
      co_graph empty_totals
end
