(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Compares classical and SSA numeric abstractions. *)

open Std_extra.Types

type totals = {
  same_precision : int;
  ssa_more_precise : int;
  classical_more_precise : int;
  incomparable : int;
}
(** Result type for comparisons between SSA Numeric and classical numeric domains
    Fields represent the number of times where:
    - num and ssa_num have the same precision
    - ssa_num is more precise
    - num is more precise
    - ssa_num and num are incomparable *)

val pp_csv : totals option pp
(** Pretty print {!totals}, using CSV format *)

val pp_summary : totals option pp
(** Pretty print {!totals}, longer format *)

module NumericComparison (N : Single_value_abstraction.Sig.SingleValueAbstraction) : sig
  val one_state_all_variables :
    Ast.Program.Loc.t ->
    ('state, N.t) Domain_builder.numeric ->
    ('state, N.t) Domain_builder.ssa_numeric ->
    'state ->
    totals
  (** Compares precision in the given state, for all variables
      i.e. at each program point, evaluate each variable in the classical and
      SSA numeric domain, and compare their precisions. *)

  val all_state_all_variables :
    'state pp ->
    ('state, N.t) Domain_builder.numeric ->
    ('state, N.t) Domain_builder.ssa_numeric ->
    'state Ast.Program.LocMap.t ->
    totals
  (** Compares precision in all states, for all variables
      i.e. at each program point, evaluate each variable in the classical and
      SSA numeric domain, and compare their precisions. *)

  val children_only :
    ('state, N.t) Domain_builder.numeric ->
    ('state, N.t) Domain_builder.ssa_numeric ->
    'state Ast.Program.LocMap.t ->
    co_graph:(Ast.Program.relation * _) list Ast.Program.LocMap.t ->
    totals
  (** Compare on children only: i.e., for all states, compare the expression
      that appear on outgoing edges only. This metric has less multiplicity than
      {!all_state_all_varibles}. *)
end
