(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(* Bourdoncle's WTO algorithm builds on Tarjan's SCC algorithm. In Tarjan:

   - We visit every node once, starting from root, by following the
     successors; this creates a spanning tree of the graph. SCCs are
     subtrees of this spanning tree, whose root is the head of the SCC
     (although in non-natural SCCs, it is possible to enter into a SCC
     without going through the head).

   - This spanning tree is obtained using DFS. What DFS guarantees is
     that there is no path from a child c of a node n to other
     children of n, provided that there is no path from c to an
     ancestor of n. Thus when we visit other children of n, we know
     that they are no path to them from the descendants of c.

   - We assign consecutive numbers to each node in the order in which
     they have been visited. As the iteration is depth-first search,
     this gives a depth-first numbering (DFN).

   - Each time we visit a node n, we push it on a stack. After the
     visit, n is popped, unless a path exists from n to an element
     earlier on the stack. So the stack contains elements currently
     visited or that belongs to a non-trivial scc. Moreover, they
     are in topological order.

   About the proof of Tarjan:
   http://ls2-www.cs.uni-dortmund.de/~wegener/papers/connected.pdf
*)

open Ast.Program
open Std_extra

(** Each component of the graph is either an individual node of the graph
      (without) self loop, or a strongly connected component where a node is
      designed as the head of the component and the remaining nodes are given
      by a list of components topologically ordered. *)
type component =
  | Component of Loc.t * partition
      (** A strongly connected component, described by its head node and the
          remaining sub-components topologically ordered *)
  | Node of Loc.t  (** A single node without self loop *)

and partition = component list
(** A list of strongly connected components, sorted topologically *)

type t = partition

(** A fold on the WTO's heads *)
let fold_heads f acc l =
  let rec partition acc l = List.fold_left component acc l
  and component acc = function
    | Node _ -> acc
    | Component (h, l) -> partition (f acc h l) l
  in
  partition acc l

let flatten wto =
  let rec f acc = function
    | [] -> acc
    | Node v :: l -> f (v :: acc) l
    | Component (v, w) :: l -> f (f (v :: acc) w) l
  in
  List.rev (f [] wto)

(** Change the WTO so that the fixpoint on it gives the iterative strategy *)
let iterative_strategy l : t =
  List.map
    (function
      | Node _ as n -> n
      | Component (hd, w) -> Component (hd, flatten w |> List.map (fun n -> Node n)))
    l

let rec equal_component x y =
  match (x, y) with
  | Node x, Node y -> Loc.equal x y
  | Component (x, cx), Component (y, cy) -> Loc.equal x y && equal_partition cx cy
  | _ -> false

and equal_partition x y = try List.for_all2 equal_component x y with Invalid_argument _ -> false

let rec pp_partition fmt =
  Format.fprintf fmt "%a" (Format.pp_print_list ~pp_sep:Format.pp_print_space pp_component)

and pp_component fmt = function
  | Node n -> Loc.pp fmt n
  | Component (head, part) ->
      Format.fprintf fmt "@[(@{<ul>%a@}@ %a)@]" Loc.pp head pp_partition part

let pp fmt partition = Format.fprintf fmt "@[%a@]" pp_partition partition
let equal l r = equal_partition l r

module DFN = Hashtbl.Make (Loc)

module Bourdoncle_Original = struct
  type state = {
    dfn : int DFN.t;
        (* Mapping from nodes to its dfn, depth-first
           numbering.  Note that we replaced the DFN=0
           test by presence in the Hashtable. *)
    mutable num : int; (* Number of visited nodes. *)
    succs : Loc.t -> Loc.t list; (* Successors transition. *)
    stack : Loc.t Stack.t;
  }

  (* Visit [vertex], and all the vertices reachable from [vertex]
     which have not been explored yet (this is a depth-first search).
     Also gives [partition], which is the partition built so far

     Returns a pair (lowlink,partition) where
     - [lowlink] is the deepest node in the stack, that is the
       successor of an explored node (i.e. a node which is below than
       [vertex] in the spanning tree built by the DFS);
     - [partition] is returned completed.

     Except for the call to component, and changing the DFN of vertex
     to max_int, this function exactly performs Tarjan's SCC
     algorithm. *)
  let rec visit state vertex partition =
    Stack.push vertex state.stack;
    let n = state.num + 1 in
    state.num <- n;
    DFN.replace state.dfn vertex n;
    let succs = state.succs vertex in
    let head, loop, partition =
      List.fold_left
        (fun (head, loop, partition) succ ->
          let min, partition =
            (* If already visited. *)
            if
              DFN.mem state.dfn succ
              (* Is another branch, which can have been visited before or after [vertex]. *)
            then (succ, partition)
            else visit state succ partition
          in
          (* OPTIM: On doit pouvoir sauvegarder le dfn de head et min pour gagner des lookups. *)
          assert (DFN.mem state.dfn min);
          assert (DFN.mem state.dfn head);
          (* If an element below in the spanning tree links to deepest
             in the stack than [vertex]. *)
          if DFN.find state.dfn min <= DFN.find state.dfn head then (min, true, partition)
          else (head, loop, partition))
        (vertex, false, partition) succs
    in
    (* If no element below in the spanning tree link to deepest than [vertex]. *)
    if Loc.equal head vertex then (
      (* Makes edges to [vertex] invisible in the second visit. *)
      DFN.replace state.dfn vertex max_int;
      let element = Stack.pop state.stack in
      (* If an element below in the spanning tree links to [vertex],
           which is the head of the SCC. *)
      if loop then (
        let rec loop element =
          if not @@ Loc.equal element vertex then (
            (* Remove from DFN for a second visit. *)
            DFN.remove state.dfn element;
            loop (Stack.pop state.stack))
          else ()
        in
        loop element;
        (head, component state vertex :: partition))
      else (head, Node vertex :: partition))
    else
      (* [vertex] is part of a strongly connected component, and is not
         the root.  Do not update partition; the vertex will be added in
         the next call to [component]. *)
      (head, partition)

  (* We found a SCC; revisit it with the [edge] to its head made
     invisible. *)
  and component state vertex =
    let succs = state.succs vertex in
    let partition =
      List.fold_left
        (fun partition succ ->
          (* Select only the branches of the spanning tree which are in a SCC with [vertex];
             and which have been removed from the DFN by [visit]. *)
          if DFN.mem state.dfn succ then partition
          else
            let _, part = visit state succ partition in
            part)
        [] succs
    in
    Component (vertex, partition)

  let partition ~init ~succs =
    let state = { dfn = DFN.create 17; num = 0; succs; stack = Stack.create () } in
    snd @@ visit state init []

  let _ = partition
end

module With_Preference = struct
  (* This implements Francois Bobot's improvements (preference on
     the choice of a head, maybe others, but possibly the WTO
     construction is slower. *)

  type level = int

  (** Status of a visited vertex during the algorithm. *)
  type status =
    | Invisible
        (** The vertex have already been added into the partition and
                      is hidden until the end of the search. *)
    | Parent of level
        (** The vertex have been visited and given a [level]. For
                            the algorithm, this implies that there is a path
                            between this vertex and the current vertex. *)

  (** Result of one [visit]  *)
  type loop =
    | NoLoop  (** The vertex is not in a loop *)
    | Loop of Loc.t * level
        (** The vertex is inside at least one loop, and
                                level is the smallest level of all these loops *)

  let min_loop x y =
    match (x, y) with
    | NoLoop, z | z, NoLoop -> z
    | Loop (_, xi), Loop (_, yi) -> if xi <= yi then x else y

  type state = {
    dfn : status DFN.t;
        (* Mapping from nodes to its dfn, depth-first
           numbering.  Note that we replaced the DFN=0
           test by presence in the Hashtable. *)
    mutable num : level; (* Number of visited nodes. *)
    succs : Loc.t -> Loc.t list; (* Successors transition. *)
    stack : Loc.t Stack.t;
  }

  (** Visit [vertex], and all the vertices reachable from [vertex]
        which have not been explored yet (this is a depth-first search).
        Also gives [partition], which is the partition built so far

        Returns a pair (loop,partition) where
        - [loop] tells whether we are in a loop or not and gives the vertex of
          this loop with the lowest level. This vertex is also the deepest in the
          stack and the neareast vertex from the root that is below [vertex] in
          the spanning tree built by the DFS);
        - [partition] is returned completed. *)
  let rec visit ~pref state vertex partition =
    match DFN.find state.dfn vertex with
    (* The vertex is already in the partition *)
    | Invisible -> (NoLoop, partition (* skip it *))
    (* The vertex have been visited but is not yet in the partition *)
    | Parent i -> (Loop (vertex, i), partition (* we are in a loop *))
    (* The vertex have not been visited yet *)
    | exception Not_found -> (
        (* Put the current vertex into the stack *)
        Stack.push vertex state.stack;
        (* Number it and mark it as visited *)
        let n = state.num + 1 in
        state.num <- n;
        DFN.replace state.dfn vertex (Parent n);
        (* Visit all its successors *)
        let succs = state.succs vertex in
        let loop, partition =
          List.fold_left
            (fun (loop, partition) succ ->
              let loop', partition = visit ~pref state succ partition in
              let loop = min_loop loop loop' in
              (loop, partition))
            (NoLoop, partition) succs
        in
        match loop with
        (* We are not in a loop. Add the vertex to the partition *)
        | NoLoop ->
            let _ = Stack.pop state.stack in
            DFN.replace state.dfn vertex Invisible;
            (NoLoop, Node vertex :: partition)
        (* We are in a loop and the current vertex is the head of this loop *)
        | Loop (head, _) when Loc.equal head vertex ->
            (* Unmark all vertices in the loop, and, if pref is given, try to
               return a better head *)
            let rec reset_SCC best_head =
              (* pop until vertex *)
              let element = Stack.pop state.stack in
              DFN.remove state.dfn element;
              if not (Loc.equal element vertex) then
                let best_head =
                  match pref with
                  (* the strict is important because we are conservative *)
                  | Some cmp when cmp best_head element < 0 -> element
                  | _ -> best_head
                in
                reset_SCC best_head
              else best_head
            in
            let best_head = reset_SCC vertex in
            let vertex, succs =
              if Loc.equal best_head vertex then (vertex, succs)
              else (best_head, state.succs best_head)
            in
            (* Makes [vertex] invisible in the subsequents visits. *)
            DFN.replace state.dfn vertex Invisible;
            (* Restart the component analysis *)
            let component =
              List.fold_left
                (fun component succ ->
                  let loop, component = visit ~pref state succ component in
                  (* Since we reset the component we should have no loop *)
                  assert (loop = NoLoop);
                  component)
                [] succs
            in
            (NoLoop, Component (vertex, component) :: partition)
        | _ ->
            (* [vertex] is part of a strongly connected component but is not the
               head. Do not update partition; the vertex will
               be added during the second visit of the SCC. *)
            (loop, partition))

  type pref = Loc.t -> Loc.t -> int

  let pref : pref option = None

  let partition ~init ~succs =
    let state = { dfn = DFN.create 17; num = 0; succs; stack = Stack.create () } in
    let loop, component = visit ~pref state init [] in
    assert (loop = NoLoop);
    component
end

let partition ~init ~succs =
  (* Bourdoncle_Original.partition ~init ~succs:(fun x -> List.map snd (succs x)) *)
  With_Preference.partition ~init ~succs:(fun x -> List.map snd (succs x))

let widening_edges_preds ~preds partition =
  fold_heads
    (fun edges head comp ->
      let comp = head :: flatten comp in
      let incomming = preds head in
      List.fold_left
        (fun edges (label, parent) ->
          if List.exists (fun x -> Loc.equal x parent) comp then (parent, label, head) :: edges
          else edges)
        edges incomming)
    [] partition

(** Set of edges from [parent] to any vertex in [set] *)
let edges_to_set ~succs set parent acc =
  let edges = succs parent in
  List.fold_left
    (fun acc (label, child) ->
      if List.exists (Loc.equal child) set then (parent, label, child) :: acc else acc)
    acc edges

let widening_edges_succs ~succs (wto : t) =
  let rec partition heads acc l = List.fold_left (component heads) acc l
  and component heads acc = function
    | Node n -> edges_to_set ~succs heads n acc
    | Component (h, l) ->
        let heads = h :: heads in
        partition heads (edges_to_set ~succs heads h acc) l
  in
  partition [] [] wto

(** Fixpoint for a function Loc.t -> L.t, stored as a map *)
module Fixpoint (L : Domains.Sig.Classical_Domain) = struct
  let widening_map widening_edges =
    List.fold_left
      (fun map (src, label, dest) -> LocMap.List.add dest (label, src) map)
      LocMap.empty widening_edges

  let[@tail_mod_cons] rec filter_map map = function
    | [] -> []
    | (rel, pred) :: list -> (
        match LocMap.find_opt pred map with
        | None -> filter_map map list
        | Some p -> (rel, pred, p) :: filter_map map list)

  (** Returns the join of the predecessors of [pos], along with a boolean [should_widen]
        indicating whether or not a back-edge was used to construct this join *)
  let join_preds widen ~preds widening_edges pos map =
    let preds = preds pos in
    match filter_map map preds with
    | [] -> None
    | [ (rel, _, pre_state) ] ->
        Utils.debug_printf "Single predecessor@.";
        L.apply rel pre_state
    | preds ->
        Utils.debug_printf "Multiple predecessors: ";
        let set, should_widen =
          List.fold_left
            (fun (set, should_widen) (rel, pred, pre_state) ->
              match L.apply rel pre_state with
              | None -> (set, should_widen)
              | Some apply ->
                  Utils.debug_printf "%a " Loc.pp pred;
                  (L.StateSet.add apply set, should_widen || List.mem (rel, pred) widening_edges))
            (L.StateSet.empty, false) preds
        in
        if widen && should_widen then (
          match LocMap.find_opt pos map with
          | Some old ->
              Utils.debug_printf "Widening@.";
              L.join_and_widen pos old set
          | None ->
              Utils.debug_printf "No widening@.";
              L.join set)
        else L.join set

  let is_included _pos l r =
    match (l, r) with
    | None, _ -> true
    | _, None -> false
    | Some l, Some r -> (*L.is_included pos l r*) L.State.equal l r

  let add_opt k o m =
    match o with
    | None -> LocMap.remove k m
    | Some v -> LocMap.add k v m

  type params = {
    preds : Loc.t -> (L.relation * Loc.t) list;
    widening_edges : (L.relation * Loc.t) list LocMap.t;
    loc0 : Loc.t;
    entrypoint : L.State.t;
  }

  (** Replace the state at [pos] by the join of its predecessors, except the bottom predecessors,
      [params.widening_edges] is the edges we look for when widening

      Also returns a boolean indicating, if [pos] is a [widenings] point,
      wether or not it has stabilized. *)
  let transfer params pos map =
    if Loc.equal params.loc0 pos then (Some params.entrypoint, true)
    else
      let next =
        join_preds true ~preds:params.preds (LocMap.List.find pos params.widening_edges) pos map
      in
      let old = LocMap.find_opt pos map in
      Utils.debug_printf "Transfer function at location %a:@.  @[%a@]@." Loc.pp pos
        (Format.pp_print_option ~none:(fun fmt () -> Format.fprintf fmt "None") L.State.pp)
        next;
      Utils.debug_pause ();
      (next, is_included pos next old)

  let rec fixpoint_component params (map, stable) = function
    | Node n ->
        let new_state, stable' = transfer params n map in
        (add_opt n new_state map, stable && stable')
    | Component (head, part) ->
        let new_head, _ = transfer params head map in
        (* Note: if we don't pass map here and start from the old map for every loop,
           it works but is much slower. Check why and see the algorithm in the paper. *)
        let rec loop map head_prestate =
          let map = add_opt head head_prestate map in
          let map, stable = fixpoint_partition params map part in
          let new_head_prestate, stable' = transfer params head map in
          let stable = stable && stable' in
          if stable then map else loop map new_head_prestate
        in
        (loop map new_head, stable)

  and fixpoint_partition params map list =
    List.fold_left (fixpoint_component params) (map, true) list

  let fixpoint ~preds ~widening_edges ~loc0 ~entrypoint wto =
    let map = LocMap.singleton loc0 entrypoint in
    let widening_edges = widening_map widening_edges in
    let params = { preds; widening_edges; loc0; entrypoint } in
    fst (fixpoint_partition params map wto)

  let pp fmt m =
    Format.fprintf fmt "  @[<v 2>%a@]@."
      (LocMap.pp_map
         ~pp_sep:(fun fmt () -> Format.fprintf fmt "@,")
         (fun fmt k v -> Format.fprintf fmt "%a -> %a" Loc.pp k L.State.pp v))
      m

  let join_preds ~preds pos map = join_preds false ~preds [] pos map
end
