(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** This modules implements functions to compute a graph's
    weak topological ordering (WTO).
    And use these to find a valid set of widening edges.
    It also contains the {!Fixpoint} module, which performs the main analysis
    (iterate until fixpoint is reached) *)

open Ast.Program

type t
(** Weak topological ordering *)

val equal : t -> t -> bool
val pp : Format.formatter -> t -> unit

val partition : init:Loc.t -> succs:(Loc.t -> ('label * Loc.t) list) -> t
(** Compute the WTO on the given LTS, [~succs] maps each node
    to a list of labelled outgoing edges *)

val widening_edges_preds :
  preds:(Loc.t -> ('label * Loc.t) list) -> t -> (Loc.t * 'label * Loc.t) list
(** Compute widening edges from preds function *)

val widening_edges_succs :
  succs:(Loc.t -> ('label * Loc.t) list) -> t -> (Loc.t * 'label * Loc.t) list
(** Compute widening edges from succs function *)

val iterative_strategy : t -> t
(** Change the WTO so that the fixpoint on it gives the iterative strategy
    Compute widening edges BEFORE calling this *)

(** Calculate the abstract program state (a function [Loc.t -> L.State.t] represented as a map)
    @param [R] is the relation type in the source, it can be [L.Relation] but might be different
            if the rewrite want to change it significantly
    @param [L] is the abstract state
    @param [NMap] is a map module with key [Loc.t], [Map.Make(Loc)] is fine here
            but more efficient maps might exist depending on Loc *)
module Fixpoint (L : Domains.Sig.Classical_Domain) : sig
  val join_preds :
    preds:(Loc.t -> (L.relation * Loc.t) list) -> Loc.t -> L.State.t LocMap.t -> L.State.t option

  val fixpoint :
    preds:(Loc.t -> (L.relation * Loc.t) list) ->
    widening_edges:(Loc.t * L.relation * Loc.t) list ->
    loc0:Loc.t ->
    entrypoint:L.State.t ->
    t ->
    L.State.t LocMap.t (* Not defined on elements who are bottom *)

  val pp : Format.formatter -> L.State.t LocMap.t -> unit
end
