(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Std_extra
open Types
open Ast.Program

(** We can easily turn this into a functor that doesn't know anything about
    the relation type being used. However, doing so prevents us from counting
    assertions in the resulting graph *)
module Classical (*(Relation : TYPE)*) = struct
  let id = "FA"
  let name fmt () = Format.fprintf fmt "%s" "FreeAlgebra.Classical"
  let state_name = "Algebraic location"
  let doc = "Classical domain implemented as free algebra of the domain signature"

  type relation = Relation.t

  module rec StateNode : sig
    type t =
      | Entrypoint
      | Loc of Loc.t
      | Apply of Relation.t * State.t
      | Join of StateSet.t

    include TYPE with type t := t
  end = struct
    type t =
      | Entrypoint
      | Loc of Loc.t
      | Apply of Relation.t * State.t
      | Join of StateSet.t

    let equal a b =
      match (a, b) with
      | Entrypoint, Entrypoint -> true
      | Loc a, Loc b -> Loc.equal a b
      | Apply (ar, al), Apply (br, bl) -> al.tag == bl.tag && Relation.equal ar br
      | Join a, Join b -> StateSet.equal a b
      | _ -> false

    let compare a b =
      match (a, b) with
      | Entrypoint, Entrypoint -> 0
      | Entrypoint, _ -> -1
      | _, Entrypoint -> 1
      | Loc a, Loc b -> Loc.compare a b
      | Loc _, _ -> -1
      | _, Loc _ -> 1
      | Apply (ar, al), Apply (br, bl) ->
          let cmp = Int.compare al.tag bl.tag in
          if cmp <> 0 then cmp else Relation.compare ar br
      | Apply _, _ -> -1
      | _, Apply _ -> 1
      | Join a, Join b -> StateSet.compare a b

    let order a b = Utils.order_of_compare (compare a b)

    let rec pp fmt = function
      | Entrypoint -> Format.fprintf fmt "@{<blue>Entrypoint@}"
      | Loc l -> Format.fprintf fmt "@{<blue>Loc@}(@[%a@])" Loc.pp l
      | Apply (c, l) -> pp_Apply fmt [ c ] l (* joins nested Applys *)
      | Join l ->
          if StateSet.is_empty l then Format.fprintf fmt "@{<blue>%s@}" (Utils.Unicode.bottom ())
          else
            Format.fprintf fmt "@{<blue>Join@}{%a}" (StateSet.pp_set ~pp_sep:Utils.pp_semicolon) l

    (* Print "Apply(build_graph.bar, x)" instead of "Apply(bar, Apply(build_graph, x))" *)
    and pp_Apply fmt seen x =
      match x.node with
      | Apply (c, l) -> pp_Apply fmt (c :: seen) l
      | x ->
          Format.fprintf fmt "@{<blue>Apply@}(@[%a,@ %a@])"
            (Format.pp_print_list ~pp_sep:Utils.pp_dot Relation.pp)
            seen pp x

    let hash =
      let hash = Utils.hash_sum ~log_total:2 in
      function
      | Entrypoint -> hash ~nb:0 0
      | Loc i -> hash ~nb:1 (Loc.hash i)
      | Apply (label, l) ->
          let hsh = Utils.hash_combine (Relation.hash label) l.hkey in
          hash ~nb:2 hsh
      | Join l ->
          let hsh = StateSet.fold (fun x y -> Utils.hash_combine x.hkey y) l 0 in
          hash ~nb:3 hsh
  end

  and StateSet : (Types.SET with type elt = State.t) = Utils.Set (State)

  and State : (TYPE_TO_INT with type t = StateNode.t Hashconsing.hash_consed) =
    Instances.InstanciateHcons (StateNode)

  module H = Hashconsing.Make (StateNode)

  let tbl = H.create 100

  (* Smart constructors *)
  let of_node n : State.t = H.hashcons tbl n
  let loc l = of_node (Loc l)
  let apply label (state : State.t) = Some (of_node (Apply (label, state)))

  (** Don't construct [Join] with 0 or 1 elements *)
  let join set =
    match StateSet.is_singleton set with
    | None -> if StateSet.is_empty set then None else Some (of_node (Join set))
    | Some _ as s -> s

  (* let join_bin a b = join (StateSet.add b (StateSet.singleton a))*)
  (* let bottom = of_node (Join StateSet.empty) *)
  let entrypoint = of_node Entrypoint
  let entrypoint _ = entrypoint

  let is_included loc l r =
    State.equal l r (*|| State.equal l bottom*)
    ||
    match (l.node, r.node) with
    | Join l', Join r -> StateSet.subset l' r || StateSet.mem l r
    | _, Join r -> StateSet.mem l r
    | _, Loc r when Loc.equal loc r -> true
    | _, _ -> false

  (* let is_included loc l r =
     let rep = is_included loc l r in
     Utils.printf "Comparing %a to %a : %b@." State.pp l State.pp r rep;
     rep *)

  let join_and_widen pos _ _ = Some (loc pos)

  let query_is_constant = None
  let query_truth_value = None

  (* let widen edges pos state f_state =
     let res = widen edges pos state f_state in
     Utils.printf "Widening %a : %a AND %a = %a@." Loc.pp pos State.pp state State.pp f_state
       State.pp res;
     res *)

  (* =================== Graph generation ===================== *)
  module Graph = struct
    module StateMap = PatriciaTree.MakeMap (State)
    module EdgeSet = Set.Make (struct
      type t = State.t * Relation.t * State.t [@@deriving ord]
    end)
    type t = (relation * State.t) list StateMap.t
    type state = State.t

    let size = StateMap.cardinal

    (** Returns the set of (unrenamed) edges, and the map of renames *)
    let rec build_graph ~cast ~join_preds psharp (next : State.t list) seen edgeset renames =
      match next with
      | [] -> (edgeset, renames)
      | n :: next -> (
          if StateSet.mem n seen then build_graph ~cast ~join_preds psharp next seen edgeset renames
          else
            let seen = StateSet.add n seen in
            match n.node with
            | Entrypoint -> build_graph ~cast ~join_preds psharp next seen edgeset renames
            | Apply (r, l) ->
                build_graph ~cast ~join_preds psharp (l :: next) seen ((l, r, n) :: edgeset) renames
            | Loc l -> (
                match join_preds l psharp with
                | None -> build_graph ~cast ~join_preds psharp next seen edgeset renames
                | Some r -> (
                    let r = cast r in
                    if State.equal r n then
                      build_graph ~cast ~join_preds psharp next seen edgeset renames
                    else
                      (* Check if r is a parent of itself, if so break loop *)
                      match r.node with
                      | Join s when StateSet.mem n s ->
                          let s = join (StateSet.remove n s) in
                          let next, renames =
                            match s with
                            | None -> (next, renames)
                            | Some x -> (x :: next, StateMap.add x n renames)
                          in
                          build_graph ~cast ~join_preds psharp next seen
                            ((n, Guard (VarExpr.of_int 1), n) :: edgeset)
                            renames
                      | _ ->
                          build_graph ~cast ~join_preds psharp (r :: next) seen edgeset
                            (StateMap.add r n renames)))
            | Join set ->
                let next, renames =
                  StateSet.fold
                    (fun elt (next, renames) -> (elt :: next, StateMap.add elt n renames))
                    set (next, renames)
                in
                build_graph ~cast ~join_preds psharp next seen edgeset renames)

    let make ~cast ~join_preds psharp =
      let init = LocMap.fold (fun _ elt acc -> cast elt :: acc) psharp [] in
      let edgeset, renames =
        build_graph ~cast ~join_preds psharp init StateSet.empty [] StateMap.empty
      in
      let rec rename x =
        match StateMap.find_opt x renames with
        | Some x -> rename x
        | None -> x
      in
      let edgeset =
        List.fold_left
          (fun set (hd, rel, tl) -> EdgeSet.add (rename hd, rel, rename tl) set)
          EdgeSet.empty edgeset
      in
      EdgeSet.fold
        (fun (hd, rel, tl) graph -> StateMap.List.add tl (rel, hd) graph)
        edgeset StateMap.empty

    let pp fmt graph =
      Format.fprintf fmt "@[<v>%a@]"
        (StateMap.pp_map ~pp_sep:Format.pp_print_cut (fun fmt k v ->
             Format.fprintf fmt "%a:@   @[<v>%a@]" State.pp k
               (Format.pp_print_list ~pp_sep:Format.pp_print_cut (fun fmt (label, elem) ->
                    Format.fprintf fmt "@[<hv><--{@[<hv>%a@]}--@ %a@]" Relation.pp label State.pp
                      elem))
               v))
        graph

    let dotty_printer fmt autom =
      Format.fprintf fmt
        "digraph A {@\n\
        \  label = \"IMP Generated graph\\nGuard arrows in green\\nAssign arrows in dashed \
         blue\\n\";@\n\
        \  labelloc = \"t\";@\n\
        \  node [shape=\"rect\"];\n";
      Format.fprintf fmt "  @[\"%a\" [shape = \"diamond\"];@\n" StateNode.pp Entrypoint;
      StateMap.List.iter
        (fun dst (label, src) ->
          Format.fprintf fmt "@[<h>\"%a\" -> \"%a\" [label=\"%a\"%a];@]@\n" State.pp src State.pp
            dst Relation.pp label relation_arrow_style label)
        autom;
      (* Format.fprintf fmt "@]@\n}@." *)
      Format.fprintf fmt "}@."

    (** Flip edges of the given graph *)
    let co_graph (graph : t) : t =
      (* We use a seen set to ensure we add childless elements as well *)
      let co_graph, seen =
        StateMap.fold
          (fun child parents (co, seen) ->
            ( List.fold_left
                (fun co (label, parent) -> StateMap.List.add parent (label, child) co)
                co parents,
              StateSet.add child seen ))
          graph (StateMap.empty, StateSet.empty)
      in
      StateSet.fold
        (fun x co_graph ->
          match StateMap.find_opt x co_graph with
          | Some _ -> co_graph
          | None -> StateMap.add x [] co_graph)
        seen co_graph

    (** Count assertion in the input graph. An assertion is simply a node with no
    a single guarded child (or two children whose guards aren't "c" and "!c") *)
    let assertions_and_dead_ends ~cast graph psharp =
      let co_graph = co_graph graph in
      let is_end =
        match LocMap.find_opt pgrm_end psharp with
        | None -> fun _ -> false
        | Some x -> State.equal (cast x)
      in
      StateMap.fold
        (fun loc children ((assertion, dead_ends) as n) ->
          if is_end loc then n
          else
            match children with
            | [] -> (assertion, dead_ends + 1)
            | [ (Assign (_, _), _) ] -> n
            | [ (Guard e, _) ] -> (
                match e.node with
                | E_Const z when not (Z.equal z Z.zero) -> n
                | _ -> (assertion + 1, dead_ends))
            | [ (Guard e, _); (Guard e', _) ] ->
                if VarExpr.(equal e (unop LNot e') || equal e' (unop LNot e)) then n
                else (
                  Format.eprintf "Weird branching, shouldn't occur";
                  (assertion + 1, dead_ends))
            | _ -> n (*failwith "unsupported branching"*))
        co_graph (0, 0)
  end
end

open Hashconsing
open Instances

module SSA = struct
  let id = "FA"
  let name fmt () = Format.fprintf fmt "%s" "FreeAlgebra.SSA"
  let state_name = "SSA Algebraic location"
  let doc = "SSA domain implemented as free algebra of the SSA domain signature"

  module rec StateNode : sig
    type t =
      | Entrypoint
      | Loc of Loc.t
      | Assume of SSAExpr.t * SSA_State.t
      | Join of StateBindingsSet.t

    include TYPE with type t := t
  end = struct
    type t =
      | Entrypoint
      | Loc of Loc.t
      | Assume of SSAExpr.t * t Hcons.t
      | Join of StateBindingsSet.t

    let equal a b =
      match (a, b) with
      | Entrypoint, Entrypoint -> true
      | Loc a, Loc b -> Loc.equal a b
      | Assume (ar, al), Assume (br, bl) -> al.tag == bl.tag && SSAExpr.equal ar br
      | Join a, Join b -> StateBindingsSet.equal a b
      | _ -> false

    let compare a b =
      match (a, b) with
      | Entrypoint, Entrypoint -> 0
      | Entrypoint, _ -> -1
      | _, Entrypoint -> 1
      | Loc a, Loc b -> Loc.compare a b
      | Loc _, _ -> -1
      | _, Loc _ -> 1
      | Assume (ar, al), Assume (br, bl) ->
          let cmp = Int.compare al.tag bl.tag in
          if cmp <> 0 then cmp else SSAExpr.compare ar br
      | Assume _, _ -> -1
      | _, Assume _ -> 1
      | Join a, Join b -> StateBindingsSet.compare a b

    let order a b = Utils.order_of_compare (compare a b)

    let hash =
      let hash = Utils.hash_sum ~log_total:2 in
      function
      | Entrypoint -> hash ~nb:0 1
      | Loc l -> hash ~nb:1 (Loc.hash l)
      | Assume (guard, s) -> hash ~nb:2 (Utils.hash_combine (SSAExpr.hash guard) s.hkey)
      | Join s -> hash ~nb:3 (StateBindingsSet.hash s)

    let rec pp fmt = function
      | Entrypoint -> Format.fprintf fmt "@{<blue>Entrypoint@}"
      | Loc l -> Format.fprintf fmt "@{<blue>Loc@}(@[%a@])" Loc.pp l
      | Assume (c, l) -> pp_assume fmt [ c ] l (* joins nested Applys *)
      | Join l ->
          if StateBindingsSet.is_empty l then
            Format.fprintf fmt "@{<blue>%s@}" (Utils.Unicode.bottom ())
          else
            Format.fprintf fmt "@{<blue>Join@}{%a}"
              (StateBindingsSet.pp_set ~pp_sep:Utils.pp_semicolon)
              l

    (* Print "Apply(foo.bar, x)" instead of "Apply(bar, Apply(foo, x))" *)
    and pp_assume fmt seen x =
      match x.node with
      | Assume (c, l) -> pp_assume fmt (c :: seen) l
      | x ->
          Format.fprintf fmt "@{<blue>Assume@}(@[%a,@ %a@])"
            (Format.pp_print_list ~pp_sep:Utils.pp_dot SSAExpr.pp)
            seen pp x
  end

  and SSA_State : (TYPE_TO_INT with type t = StateNode.t Hcons.t) =
    Instances.InstanciateHcons (StateNode)

  and StateBindings : (TYPE with type t = SSA_State.t * Bindings.t) = MkBindings (SSA_State)

  and StateBindingsSet : (SET with type elt = StateBindings.t) = Utils.Set (StateBindings)

  and H : sig
    include Hashconsing.S with type key = StateNode.t
    val of_node : StateNode.t -> SSA_State.t
  end = struct
    include Hashconsing.Make (StateNode)
    let tbl = create 100

    (* Smart constructors *)
    let of_node n : SSA_State.t = hashcons tbl n
  end

  let loc l = H.of_node (Loc l)
  let assume guard state = Some (H.of_node (Assume (guard, state)))

  (** Don't construct [Join] with 0 or 1 elements *)
  let join set =
    match StateBindingsSet.is_singleton set with
    | None -> if StateBindingsSet.is_empty set then None else Some (H.of_node (Join set))
    | Some _ -> failwith "Shouldn't join singletons"

  (* let join_bin a b = join (StateBindingsSet.add b (StateBindingsSet.singleton a))
     let bottom = join StateBindingsSet.empty *)
  let entrypoint = H.of_node StateNode.Entrypoint

  let is_included loc l r =
    SSA_State.equal l r
    ||
    let open StateNode in
    match (l.node, r.node) with
    | Join l, Join r -> StateBindingsSet.subset l r
    (* | _, Join r -> StateSet.mem l r *)
    | _, Loc r when Loc.equal loc r -> true
    | _, _ -> false

  let join_and_widen l _ _ = Some (loc l)

  let entrypoint _ = entrypoint

  let query_is_constant = None
  let query_truth_value = None

  (* =================== Graph generation ===================== *)
  module Graph = struct
    module Relation = struct
      type t =
        | R_Bind of Bindings.t
        | R_Assume of SSAExpr.t
      [@@deriving ord]

      let pp fmt = function
        | R_Assume e -> SSAExpr.pp fmt e
        | R_Bind b -> Bindings.pp fmt b

      let dotty_pp fmt = function
        | R_Assume e -> SSAExpr.pp fmt e
        | R_Bind b ->
            if SSAVarMap.is_empty b then Format.fprintf fmt "%s" (Utils.Unicode.emptyset ())
            else
              Format.fprintf fmt "%a;\\l"
                (SSAVarMap.pp_map
                   ~pp_sep:(fun fmt () -> Format.fprintf fmt ";\\l")
                   (fun fmt var expr ->
                     Format.fprintf fmt "@[%a%s%a@]" SSAVar.pp var (Utils.Unicode.colon_eq ())
                       SSAExpr.pp expr))
                b
    end

    module StateMap = PatriciaTree.MakeMap (SSA_State)
    module StateSet = PatriciaTree.MakeSet (SSA_State)
    module EdgeSet = Set.Make (struct
      type t = SSA_State.t * Relation.t * SSA_State.t [@@deriving ord]
    end)
    type t = (Relation.t * SSA_State.t) list StateMap.t
    type state = SSA_State.t

    let size = StateMap.cardinal

    (** Returns the set of (unrenamed) edges, and the map of renames *)
    let rec build_graph ~cast ~join_preds psharp (next : SSA_State.t list) seen edgeset renames =
      match next with
      | [] -> (edgeset, renames)
      | n :: next -> (
          if StateSet.mem n seen then build_graph ~cast ~join_preds psharp next seen edgeset renames
          else
            let seen = StateSet.add n seen in
            match n.node with
            | Entrypoint -> build_graph ~cast ~join_preds psharp next seen edgeset renames
            | Assume (e, l) ->
                build_graph ~cast ~join_preds psharp (l :: next) seen
                  ((l, Relation.R_Assume e, n) :: edgeset)
                  renames
            | Loc l -> (
                match join_preds l psharp with
                | None -> build_graph ~cast ~join_preds psharp next seen edgeset renames
                | Some r ->
                    let r = cast r in
                    if SSA_State.equal r n then
                      build_graph ~cast ~join_preds psharp next seen edgeset renames
                    else
                      build_graph ~cast ~join_preds psharp (r :: next) seen edgeset
                        (StateMap.add r n renames))
            | Join set ->
                let next, edgeset =
                  StateBindingsSet.fold
                    (fun (elt, b) (next, edgeset) ->
                      (elt :: next, (elt, Relation.R_Bind b, n) :: edgeset))
                    set (next, edgeset)
                in
                build_graph ~cast ~join_preds psharp next seen edgeset renames)

    let make ~cast ~join_preds psharp =
      let init = LocMap.fold (fun _ elt acc -> cast elt :: acc) psharp [] in
      let edgeset, renames =
        build_graph ~cast ~join_preds psharp init StateSet.empty [] StateMap.empty
      in
      let rec rename x =
        match StateMap.find_opt x renames with
        | Some x -> rename x
        | None -> x
      in
      let edgeset =
        Stdlib.List.fold_left
          (fun set (hd, rel, tl) -> EdgeSet.add (rename hd, rel, rename tl) set)
          EdgeSet.empty edgeset
      in
      EdgeSet.fold
        (fun (hd, rel, tl) graph -> StateMap.List.add tl (rel, hd) graph)
        edgeset StateMap.empty

    let pp fmt graph =
      Format.fprintf fmt "@[<v>%a@]"
        (StateMap.pp_map ~pp_sep:Format.pp_print_cut (fun fmt k v ->
             Format.fprintf fmt "%a:@   @[<v>%a@]" SSA_State.pp k
               (Format.pp_print_list ~pp_sep:Format.pp_print_cut (fun fmt (label, elem) ->
                    Format.fprintf fmt "@[<hv><--{@[<hv>%a@]}--@ %a@]" Relation.pp label
                      SSA_State.pp elem))
               v))
        graph

    let arrow_color fmt = function
      | Relation.R_Assume _ -> guard_arrow_style fmt ()
      | Relation.R_Bind _ -> assign_arrow_style fmt ()

    let rec pp_node fmt (x : SSA_State.t) =
      match x.node with
      | Entrypoint -> Format.fprintf fmt "@{<blue>Entrypoint@}"
      | Loc l -> Format.fprintf fmt "@{<blue>Loc@}(@[%a@])" Loc.pp l
      | Assume (c, l) -> pp_assume fmt [ c ] l (* joins nested Applys *)
      | Join l ->
          if StateBindingsSet.is_empty l then
            Format.fprintf fmt "@{<blue>%s@}" (Utils.Unicode.bottom ())
          else Format.fprintf fmt "@{<blue>Join@}{...} # %d" x.tag
    (* (StateBindingsSet.pp_set ~pp_sep:Utils.pp_semicolon)
       l *)

    (* Print "Apply(foo.bar, x)" instead of "Apply(bar, Apply(foo, x))" *)
    and pp_assume fmt seen x =
      match x.node with
      | Assume (c, l) -> pp_assume fmt (c :: seen) l
      | _ ->
          Format.fprintf fmt "@{<blue>Assume:@}\\l    %a\\lFrom: %a\\l"
            (Format.pp_print_list ~pp_sep:(fun fmt () -> Format.fprintf fmt "\\l    ") SSAExpr.pp)
            seen pp_node x

    let dotty_printer fmt autom =
      Format.fprintf fmt
        "digraph A {@\n\
        \  label = \"SSA Generated graph\\nGuard arrows in green\\nAssign arrows in dashed \
         blue\\n\";@\n\
        \  labelloc = \"t\";@\n\
        \  node [shape=\"rect\"];\n";
      Format.fprintf fmt "  @[\"%a\" [shape = \"diamond\"];@]@\n" StateNode.pp Entrypoint;
      StateMap.List.iter
        (fun dst (label, src) ->
          Format.fprintf fmt "  @[<h>\"%a\" -> \"%a\" [label=\"%a\"%a];@]@\n" pp_node src pp_node
            dst Relation.dotty_pp label arrow_color label)
        autom;
      (* Format.fprintf fmt "@]@\n}@." *)
      Format.fprintf fmt "}@."

    (** Flip edges of the given graph *)
    let co_graph (graph : t) : t =
      (* We use a seen set to ensure we add childless elements as well *)
      let co_graph, seen =
        StateMap.fold
          (fun child parents (co, seen) ->
            ( Stdlib.List.fold_left
                (fun co (label, parent) -> StateMap.List.add parent (label, child) co)
                co parents,
              StateSet.add child seen ))
          graph (StateMap.empty, StateSet.empty)
      in
      StateSet.fold
        (fun x co_graph ->
          match StateMap.find_opt x co_graph with
          | Some _ -> co_graph
          | None -> StateMap.add x [] co_graph)
        seen co_graph

    open Relation

    (** Count assertion in the input graph. An assertion is simply a node with no
    a single guarded child (or two children whose guards aren't "c" and "!c") *)
    let assertions_and_dead_ends ~cast graph psharp =
      let co_graph = co_graph graph in
      let is_end =
        match LocMap.find_opt pgrm_end psharp with
        | None -> fun _ -> false
        | Some x -> SSA_State.equal (cast x)
      in
      StateMap.fold
        (fun loc children ((assertion, dead_ends) as n) ->
          if is_end loc then n
          else
            match children with
            | [] -> (assertion, dead_ends + 1)
            | [ (R_Bind _, _) ] -> n
            | [ (R_Assume e, _) ] -> (
                match e.node with
                | E_Const z when not (Z.equal z Z.zero) -> n
                | _ -> (assertion + 1, dead_ends))
            | [ (R_Assume e, _); (R_Assume e', _) ] ->
                if SSAExpr.(equal e (unop LNot e') || equal e' (unop LNot e)) then n
                else (
                  Format.eprintf "Weird branching, shouldn't occur";
                  (assertion + 1, dead_ends))
            | _ -> n (*failwith "unsupported branching"*))
        co_graph (0, 0)
  end
end
