(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Free algebra domains, both for IMP ({!Classical}) and {!SSA}. *)

open Std_extra.Types
open Sig
open Ast.Program

(** Represent an abstract state as a free algebra
    (each function is a constructor)
    We use hash-consing to speed up comparisons and set operations

    We can easily turn this into a functor that doesn't know anything about
    the relation type [Relation.t] being used. However, doing so prevents us from
    easily counting assertions in the resulting graph.

    Its command line id is [FA]. *)
module Classical (*(Relation : TYPE)*) : sig
  (** The type of states: a free algebra of the domain signature *)
  module rec StateNode : sig
    type t = private
      | Entrypoint
      | Loc of Loc.t
      | Apply of Relation.t * State.t
      | Join of StateSet.t
    include TYPE with type t := t
  end

  and State : (TYPE with type t = StateNode.t Hashconsing.hash_consed)
  (** States are just hash-consed nodes *)

  and StateSet : (SET with type elt = State.t)
  (** Set of states *)

  include
    Classical_Domain
      with type relation = Relation.t
       and module State := State
       and module StateSet := StateSet

  module Graph : Graph with type state = State.t
  (** Generate the graph from psharp, we use a [cast] function as argument
      to allow Psharp to have a different type. This is useful when this domain
      is part of a product in the analysis. *)
end

(** Free algebra of the SSA Domain signature (each function is a constructor)
    We use hash-consing to speed up comparisons and set operations.

    Its command line id is [FA] (under [Lift] or [LiftGVN]). *)
module SSA : sig
  (** The type of states: a free algebra of the domain signature *)
  module rec StateNode : sig
    type t = private
      | Entrypoint
      | Loc of Loc.t
      | Assume of SSAExpr.t * SSA_State.t
      | Join of StateBindingsSet.t
  end

  and SSA_State : (TYPE with type t = StateNode.t Hashconsing.hash_consed)
  (** SSA_State are just hash-consed nodes *)

  and StateBindingsSet : (SET with type elt = SSA_State.t * Bindings.t)
  (** Set of states paired with bindings, as used in the join *)

  include
    SSA_Domain with module SSA_State := SSA_State and module StateBindingsSet := StateBindingsSet

  module Graph : Graph with type state = SSA_State.t
  (** Generate the graph from psharp, we use a [cast] function as argument
      to allow Psharp to have a different type *)
end
