(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Sig
open Std_extra
open Ast.Program
open Ast

let ( let* ) = Stdlib.Option.bind

module HashTable = Hashtbl.Make (Instances.Instanciate (Instances.Array) (SSAExpr))

module SSA_Lift
    (SSA : SSA_Domain)
    (P : sig
      val do_gvn : bool
    end) =
struct
  let id = if P.do_gvn then "LiftGVN" else "Lift"
  let name fmt () = Format.fprintf fmt "%s(@[%a@])" "SSA_Lift" SSA.name ()
  let state_name = "SSA lifting state"
  let doc = "Lifting of an SSA Domain to a classical one"

  type nonrec relation = relation

  module State = struct
    (* Map of program variables -> SSA values, and an SSA State*)
    type t = {
      var_store : SSAExpr.t VarMap.t;
      ssa_state : SSA.SSA_State.t;
    }
    [@@deriving ord, eq]

    let order a b = Utils.order_of_compare (compare a b)

    let pp fmt { var_store; ssa_state } =
      Format.fprintf fmt "@[<v>Mapping to SSA:@,  @[<v>%a@]@,SSA State:@,  @[<v>%a@]@]"
        (VarMap.pp_map
           ~pp_sep:(fun fmt () -> Format.pp_print_break fmt 1 0)
           (fun fmt v e -> Format.fprintf fmt "@[%a -> %a@]" Var.pp v SSAExpr.pp e))
        var_store SSA.SSA_State.pp ssa_state

    let hash _ = assert false
    (* Utils.hash_combine (SSA.SSA_State.hash x.ssa_state) (VarMap.hash SSAExpr.hash x.var_store) *)
  end

  module StateSet = Utils.Set (State)
  open State

  let bind state expr =
    let bound_expr = Expr.bind SSAExpr.of_node (fun x -> VarMap.find x state.var_store) expr in
    (* SSAExpr.simplify_expr (SSA.query_is_constant state.ssa_state) *)
    bound_expr

  let apply rel state =
    match rel with
    | Assign (v, e) -> Some { state with var_store = VarMap.add v (bind state e) state.var_store }
    | Guard e ->
        let* ssa_state = SSA.assume (bind state e) state.ssa_state in
        Some { state with ssa_state }

  module ID = struct
    type t =
      | Of_Set of StateSet.t
      | Of_Loc of Loc.t
    [@@deriving ord]
  end

  module IDMap = Map.Make (ID)
  (** To keep things non-recursive, our {!SSAVar.t} (SSA variables) are
      not made of [Var.t * SSA.SSA_State.t] (as {!SSA.SSA_State.t} contains SSA variables
      when using the {!Free_algebra.SSA} domain)

      Instead our variables ar [Var.t * int] where the [int] a unique number
      associated with the state. We use an [IDMap.t ref] to remember which states
      have been given numbers. Its a bit complex because:
      - Some states must be numbered before the state can be build
        (in {!join}, from its arguments, using the {!Of_Set} constructor
        since the full {!join} term requires the variables).
        Note that once the resulting state has been built, it must also be added
        to the map with an {!Of_SSA_State} constructor and with the same id.
      - Some can only be obtained after the state has been built [Of_SSA_State]
        (in [rename], during a [widen])
      So we keep a reference to a map [ids: SSA.SSA_State.t -> int] storing the hashed, to
      avoid given two different numbers (via the two different methods) to the same state *)

  open ID
  let counter = ref 1
  let known_ids = ref IDMap.empty
  let find_id x =
    match IDMap.find_opt x !known_ids with
    | Some n -> n
    | None ->
        let n = !counter in
        incr counter;
        known_ids := IDMap.add x n !known_ids;
        n

  let find_phi_variables id state set =
    let placeholder_expr = SSAExpr.of_int 0 in
    let phi_vars = ref VarMap.empty in
    (* New var_store for the join, including placeholder for the phi_variables *)
    let new_var_store =
      StateSet.fold
        (fun st new_var_store ->
          if st == state then new_var_store
          else
            VarMap.idempotent_inter
              (fun v _ _ ->
                phi_vars := VarMap.add v () !phi_vars;
                placeholder_expr)
              st.var_store new_var_store)
        set state.var_store
    in
    (* Filter phi_vars to new var store, since some may have been dropped
       after being introduced (if they are absent in late elements of the set) *)
    let phi_vars = VarMap.nonidempotent_inter_no_share (fun _ () _ -> ()) !phi_vars new_var_store in
    if not P.do_gvn then
      (* No GVN, create one new variable per phi variable *)
      ( VarMap.fold
          (fun v () var_store ->
            let var = SSAExpr.var (ssavar_join v id) in
            VarMap.add v var var_store)
          phi_vars new_var_store,
        phi_vars )
    else
      (* Do GVN, hash old var values for quick comparison *)
      let hash = HashTable.create 20 in
      let n = StateSet.cardinal set in
      let new_var_store = ref new_var_store in
      let phi_vars =
        VarMap.filter
          (fun v () ->
            (* Store all previous values in an array *)
            let arr = Array.make n placeholder_expr in
            let _ =
              StateSet.fold
                (fun st i ->
                  arr.(i) <- VarMap.find v st.var_store;
                  i + 1)
                set 0
            in
            match HashTable.find hash arr with
            | v' ->
                new_var_store := VarMap.add v (SSAExpr.var (ssavar_join v' id)) !new_var_store;
                false
            | exception Not_found ->
                new_var_store := VarMap.add v (SSAExpr.var (ssavar_join v id)) !new_var_store;
                HashTable.replace hash arr v;
                true)
          phi_vars
      in
      (!new_var_store, phi_vars)

  (** Join multiples states together:
      - Fail on the empty set
      - returns the unique element for a singleton
      - for mor complex sets [{(store1,loc1); ... (storeN,locN)}]
        - computes [vars = [v1; ...; vM]], the set of non-equal variable (variable whose value depend on the incomming branch)
        - computes [loc = SSA.join {(v1->store1 v1 ... vM->store1 vM).loc1; ... (v1->storeN v1 ... vM->storeN vM).locN}]
        - returns [store1[v1->(v1,loc); ... vM->(vM,loc)], loc] *)
  let join ?loc set =
    match StateSet.is_singleton set with
    | Some _ as x -> x
    | None -> (
        match StateSet.choose_opt set with
        | None -> None
        | Some state ->
            let id =
              match loc with
              | None -> find_id (Of_Set set)
              | Some (l, _) -> find_id (Of_Loc l)
            in
            let var_store, phi_vars = find_phi_variables id state set in
            let ssa_state_set =
              StateSet.fold
                (fun { var_store; ssa_state } set ->
                  let bindings =
                    SSAVarMap.of_list
                      (VarMap.fold
                         (fun v _ l -> (ssavar_join v id, VarMap.find v var_store) :: l)
                         phi_vars [])
                  in
                  SSA.StateBindingsSet.add (ssa_state, bindings) set)
                set SSA.StateBindingsSet.empty
            in
            (* [SSA.join] is [None] only when the set is empty *)
            let* ssa_state =
              match loc with
              | None -> SSA.join ssa_state_set
              | Some (l, old) -> SSA.join_and_widen l old ssa_state_set
            in
            Some { var_store; ssa_state })

  let join_and_widen loc old set = join ~loc:(loc, old.ssa_state) set
  let join set = join set

  (* Initial vars are set to v_entrypoint, other undefined *)
  let entrypoint vars =
    let ssa_state = SSA.entrypoint vars in
    {
      var_store = VarMap.of_list (List.map (fun v -> (v, SSAExpr.var (ssavar_init v))) vars);
      ssa_state;
    }

  let store_inclusion _pos _ssa_pos lstore rstore =
    VarMap.reflexive_subset_domain_for_all2 (fun _v k k' -> SSAExpr.equal k k') lstore rstore
  (* ||
     match k'.node with
     | E_Var (v', loc') -> Var.equal v v' && SSA.is_included pos loc loc'
     | _ -> false)) *)

  let is_included pos l r =
    store_inclusion pos r.ssa_state l.var_store r.var_store
    && SSA.is_included pos l.ssa_state r.ssa_state

  let query_is_constant f state expr =
    let ssa_expr = bind state expr in
    f state.ssa_state ssa_expr

  let query_is_constant =
    match SSA.query_is_constant with
    | None -> None
    | Some f -> Some (query_is_constant f)

  let query_truth_value f state expr =
    let ssa_expr = bind state expr in
    f state.ssa_state ssa_expr

  let query_truth_value =
    match SSA.query_truth_value with
    | None -> None
    | Some f -> Some (query_truth_value f)
end
