(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** The {!SSA_Lift} functor, which lifts an {!SSA_Domain} into a {!Classical_Domain}
    and optionaly performs GVN. *)

open Sig
open Std_extra.Types
open Ast.Program

module HashTable : Hashtbl.S with type key = SSAExpr.t array
(** Hashtable module used for GVN *)

(** This functor generates a {!Classical_Domain} from an {!SSA_Domain} as
    described in the paper.
    The second parameter is used to toggle global value numbering (GVN) on or off.

    Its command line id is [Lift] or [LiftGVN]. *)
module SSA_Lift
    (SSA : SSA_Domain)
    (_ : sig
      val do_gvn : bool
    end) : sig
  (** The state contains:
      - {!var_store}: a mapping from IMP variables {!Var.t} to their current
        SSA expressions {!SSAExpr.t}
      - {!ssa_state}: the current SSA state *)
  module State : sig
    type t = {
      var_store : SSAExpr.t VarMap.t;
      ssa_state : SSA.SSA_State.t;
    }
    include TYPE with type t := t
  end

  include Classical_Domain with type relation = Ast.Program.relation and module State := State

  val bind : State.t -> VarExpr.t -> SSAExpr.t
  (** Convert a classical expression into an SSA one, by replacing all
      the variables by their values in the current state *)
end
