(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Ast.Program
open Std_extra
open Single_value_abstraction.Sig

let nb_backwards_iter = ref 3
let set_nb_backwards_iter i = nb_backwards_iter := i

module Classical (Zs : SingleValueAbstraction) = struct
  let id = "N"
  let name fmt () = Format.fprintf fmt "%s(@[%a@])" "Numeric.Classical" Zs.name ()
  let state_name = "Variable values"
  let doc = "Classical numeric non relational domain"

  type nonrec relation = relation

  module State = struct
    type t = Zs.t VarMap.t [@@deriving eq, ord]
    (** Absent variables are top *)

    let order a b = Utils.order_of_compare (compare a b)

    let pp fmt m =
      Format.fprintf fmt "@[<v>%a@]"
        (VarMap.pp_map
           ~pp_sep:(fun fmt () -> Format.pp_print_break fmt 1 0)
           (fun fmt v z ->
             Format.fprintf fmt "@[%a %s %a@]" Var.pp v (Utils.Unicode.elem_of ()) Zs.pp z))
        (VarMap.filter (fun _ z -> not @@ Zs.subseteq Zs.top z) m)

    let hash _ = assert false
  end
  module StateSet = Std_extra.Utils.Set (State)

  let entrypoint _vars = VarMap.empty
  (* VarMap.of_list (List.map (fun x -> (x, Zs.top)) vars) *)

  let inter f a b =
    VarMap.idempotent_inter_filter
      (fun _ za zb ->
        let z = f za zb in
        if Zs.equal z Zs.top then None else Some z)
      a b

  let join vset =
    if StateSet.is_empty vset then None
    else
      match StateSet.is_singleton vset with
      | Some _ as x -> x
      | None ->
          let start = StateSet.choose vset in
          Some (StateSet.fold (inter Zs.join) (StateSet.remove start vset) start)

  let join_and_widen _ old set = Option.map (inter Zs.widen old) (join set)

  let is_included _ a b =
    VarMap.reflexive_subset_domain_for_all2 (fun _ z z' -> Zs.subseteq z z') a b

  let find v store = try VarMap.find v store with Not_found -> Zs.top

  (** Compute the best approximation of [expr] using the approximations of the
      variables in [store] and the forward transfer functions of {!Zs.Forwards} *)
  let rec expr_forwards store (expr : VarExpr.t) =
    match expr.node with
    | E_Var v -> find v store
    | E_Const c -> Zs.Forwards.const c
    | E_Binop (l, b, r) ->
        let l = expr_forwards store l in
        let r = expr_forwards store r in
        Zs.Forwards.binop b l r
    | E_Unop (u, e) ->
        let e = expr_forwards store e in
        Zs.Forwards.unop u e
    | E_IfThenElse (i, t, e) ->
        let i = expr_forwards store i in
        let t = expr_forwards store t in
        let e = expr_forwards store e in
        Zs.Forwards.ternary i t e

  let rec expr_backwards store (expr : VarExpr.t) res =
    if Zs.equal res Zs.top then store
    else
      match expr.node with
      | E_Const c -> if Zs.contains res c then store else raise Utils.Bottom
      | E_Var v -> (
          match Zs.meet res (find v store) with
          | None -> raise Utils.Bottom
          | Some x -> VarMap.add v x store)
      | E_Unop (u, e) ->
          let e_forward = expr_forwards store e in
          back_propagate store e e_forward (Zs.Backwards.unop u e_forward ~res)
      | E_Binop (l, b, r) ->
          let l_forward = expr_forwards store l in
          let r_forward = expr_forwards store r in
          let l_back, r_back = Zs.Backwards.binop b l_forward r_forward ~res in
          let store = back_propagate store l l_forward l_back in
          back_propagate store r r_forward r_back
      | E_IfThenElse (i, t, e) ->
          let i_forward = expr_forwards store i in
          let t_forward = expr_forwards store t in
          let e_forward = expr_forwards store e in
          let i_back, t_back, e_back = Zs.Backwards.ternary i_forward t_forward e_forward ~res in
          let store = back_propagate store i i_forward i_back in
          let store = back_propagate store t t_forward t_back in
          back_propagate store e e_forward e_back

  (** Continue propagation if:
      - we learned something [v_back != None]
      - what we learned is more precise [meet < v_forward] *)
  and back_propagate store expr v_forward v_back =
    match v_back with
    | None -> store
    | Some v_back -> (
        match Zs.meet v_forward v_back with
        | None -> raise Utils.Bottom
        | Some meet ->
            (* Since meet <= v_forwards, it is either = or strictly included *)
            if Zs.equal meet v_forward then store else expr_backwards store expr meet)

  (* We can run backwards multiple times if we wish *)
  let rec expr_backwards_iter store e res = function
    | 0 -> store
    | n ->
        let store' = expr_backwards store e res in
        if VarMap.equal Zs.equal store store' then store'
        else expr_backwards_iter store' e res (Int.sub n 1)

  let apply r store =
    try
      match r with
      | Assign (v, e) ->
          let z = expr_forwards store e in
          if Zs.equal z Zs.top then Some (VarMap.remove v store) else Some (VarMap.add v z store)
      | Guard e ->
          Some
            (match e.node with
            | E_Binop (_, (Eq | Ne | Le | Lt | Ge | Gt), _) ->
                (* If a comparison operator doesn't return 0, it returns 1 *)
                expr_backwards_iter store e Zs.one !nb_backwards_iter
            | E_Unop (LNot, e) ->
                (* If lnot does not return 0, its argument is zero *)
                expr_backwards_iter store e Zs.zero !nb_backwards_iter
            | _ ->
                (* In theory [expr_backwards store e [Zs.top \ Zs.one],
                   but that is just [top] for the interval domain *)
                expr_backwards_iter store e Zs.non_zero !nb_backwards_iter)
    with Utils.Bottom | Division_by_zero -> None

  let query_truth_value' store expr =
    try
      let eval = expr_forwards store expr in
      Zs.to_ternary eval
    with Utils.Bottom | Division_by_zero -> False
  let query_truth_value = Some query_truth_value'

  let query_is_constant' store expr =
    try
      let eval = expr_forwards store expr in
      Zs.is_singleton eval
    with Utils.Bottom | Division_by_zero -> None
  let query_is_constant = Some query_is_constant'
end

module ClassicalQuery (Zs : SingleValueAbstraction) = struct
  include Classical (Zs)

  let id = "QN"

  let apply rel state =
    match rel with
    | Assign (x, e) -> (
        let diff = VarExpr.(binop MinusA e (var x)) in
        match query_is_constant' state diff with
        | Some z -> (
            if (* x - e is zero, so x := e is noop *)
               Z.equal z Z.zero then Some state
            else
              match query_is_constant' state e with
              | Some z ->
                  (* e is constant z, so replace x := e with x := z *)
                  apply (Assign (x, VarExpr.of_const z)) state
              | None ->
                  (* e - x is constant z, so replace x := e with x := z + x *)
                  apply (Assign (x, VarExpr.(binop PlusA (of_const z) (var x)))) state)
        | None -> (
            match query_is_constant' state e with
            | Some z ->
                (* e is constant z, so replace x := e with x := z *)
                apply (Assign (x, VarExpr.of_const z)) state
            | None -> apply rel state))
    | Guard e -> (
        match query_truth_value' state e with
        | False -> (* e is zero, guard is always false *) None
        | True -> Some state (* Non-zero constant *)
        | TrueFalse -> apply rel state)
end

module SSA (Zs : SingleValueAbstraction) = struct
  let id = "N"
  let name fmt () = Format.fprintf fmt "%s(@[%a@])" "Numeric.SSA" Zs.name ()
  let state_name = "SSA expression values"
  let doc = "Classical numeric non relational domain"

  module SSA_State = struct
    type t = Zs.t SSAExpr.Map.t [@@deriving eq, ord]

    let order a b = Utils.order_of_compare (compare a b)

    let pp fmt m =
      Format.fprintf fmt "@[<v>%a@]"
        (SSAExpr.Map.pp_map
           ~pp_sep:(fun fmt () -> Format.pp_print_break fmt 1 0)
           (fun fmt k v ->
             Format.fprintf fmt "@[%a %s %a@]" SSAExpr.pp k (Utils.Unicode.elem_of ()) Zs.pp v))
        (SSAExpr.Map.filter (fun _ z -> not @@ Zs.subseteq Zs.top z) m)

    let hash _ = assert false (* SSAExpr.Map.hash Zs.hash x *)
  end

  module StateBindings = MkBindings (SSA_State)

  module StateBindingsSet = Utils.Set (StateBindings)

  let add expr z state =
    if Ast.Expr.is_const expr || Zs.equal z Zs.top then state else SSAExpr.Map.add expr z state

  (** Evaluate an expression and add intermediate results to the state *)
  let rec expr_forwards_and_reduce state expr =
    match SSAExpr.Map.find_opt expr state with
    | Some z -> (z, state (* EvalReuse *))
    | None -> (
        match expr.node with
        | E_Const z -> (Zs.Forwards.const z, state) (* EvalCst *)
        | E_Var _ -> (Zs.top, state) (* EvalVar *)
        | E_Unop (op, expr') ->
            let z, state = expr_forwards_and_reduce state expr' in
            let z = Zs.Forwards.unop op z in
            (z, add expr z state)
        | E_Binop (l, op, r) ->
            let zl, state = expr_forwards_and_reduce state l in
            let zr, state = expr_forwards_and_reduce state r in
            let z = Zs.Forwards.binop op zl zr in
            (z, add expr z state)
        | E_IfThenElse _ -> assert false)

  (** Evaluate an expression and without adding intermediate results to the state *)
  let rec expr_forwards state expr =
    match SSAExpr.Map.find_opt expr state with
    | Some z -> z (* EvalReuse *)
    | None -> (
        match expr.node with
        | E_Const z -> Zs.Forwards.const z (* EvalCst *)
        | E_Var _ -> Zs.top (* EvalVar *)
        | E_Unop (op, expr') ->
            let z = expr_forwards state expr' in
            Zs.Forwards.unop op z
        | E_Binop (l, op, r) ->
            let zl = expr_forwards state l in
            let zr = expr_forwards state r in
            Zs.Forwards.binop op zl zr
        | E_IfThenElse _ -> assert false)

  let rec expr_backwards state (expr : SSAExpr.t) res =
    if Zs.equal res Zs.top then state
    else
      match expr.node with
      | E_Const c -> if Zs.contains res c then state else raise Utils.Bottom
      | E_Var _ -> state (* back_propagate has already added the new variable info to the state *)
      | E_Unop (u, e) ->
          let e_forward, state = expr_forwards_and_reduce state e in
          back_propagate state e e_forward (Zs.Backwards.unop u e_forward ~res)
      | E_Binop (l, b, r) ->
          let l_forward, state = expr_forwards_and_reduce state l in
          let r_forward, state = expr_forwards_and_reduce state r in
          let l_back, r_back = Zs.Backwards.binop b l_forward r_forward ~res in
          let state = back_propagate state l l_forward l_back in
          back_propagate state r r_forward r_back
      | E_IfThenElse _ -> assert false

  (** Continue propagation if:
      - we learned something [v_back != None]
      - what we learned is more precise [meet < v_forward] *)
  and back_propagate state expr v_forward v_back =
    match v_back with
    | None -> state
    | Some v_back -> (
        match Zs.meet v_forward v_back with
        | None -> raise Utils.Bottom
        | Some meet ->
            if Zs.equal meet v_forward then state
            else expr_backwards (add expr meet state) expr meet)

  (** We can run backwards multiple times if we wish *)
  let rec expr_backwards_iter state e res = function
    | 0 -> state
    | n ->
        let e_fwd, state = expr_forwards_and_reduce state e in
        let state' = back_propagate state e e_fwd (Some res) in
        if SSAExpr.Map.equal Zs.equal state state' then state'
        else expr_backwards_iter state' e res Stdlib.(n - 1)

  let assume (e : SSAExpr.t) state =
    try
      Some
        (match e.node with
        | E_Binop (_, (Eq | Ne | Le | Lt | Ge | Gt), _) ->
            (* If a comparison operator doesn't return 0, it returns 1 *)
            expr_backwards_iter state e Zs.one !nb_backwards_iter
        | E_Unop (LNot, e) ->
            (* If lnot does not return 0, its argument is zero *)
            expr_backwards_iter state e Zs.zero !nb_backwards_iter
        | _ -> expr_backwards_iter state e Zs.non_zero !nb_backwards_iter)
    with Utils.Bottom | Division_by_zero -> None

  (* let assume e state =
     match assume e state with
     | None ->
         Utils.printf "Assuming %a is bottom@." SSAExpr.pp e;
         None
     | Some st ->
         if not (SSA_State.equal state st) then
           Utils.printf "Assuming %a:@. From %a@. To %a@." SSAExpr.pp e SSA_State.pp state
             SSA_State.pp st;
         Some st *)

  let query_truth_value' store expr =
    try
      let eval = expr_forwards store expr in
      Zs.to_ternary eval
    with Division_by_zero -> False
  let query_truth_value = Some query_truth_value'

  let query_is_constant' store expr =
    try
      let eval = expr_forwards store expr in
      Zs.is_singleton eval
    with Division_by_zero -> None
  let query_is_constant = Some query_is_constant'

  let eval_and_add var expr state =
    let ze, state = expr_forwards_and_reduce state expr in
    SSAExpr.Map.add (SSAExpr.var var) ze state

  (** Add the new bindings to the current state *)
  let add_bindings (state, bindings) = SSAVarMap.fold eval_and_add bindings state

  (** Creates a new [SSA_State.t] by applying [f] on the intersection of [a] and [b] *)
  let inter f a b = SSAExpr.Map.idempotent_inter (fun _ ka kb -> f ka kb) a b

  let join joinset =
    try
      match StateBindingsSet.is_singleton joinset with
      | Some x -> Some (add_bindings x)
      | None -> (
          match StateBindingsSet.choose_opt joinset with
          | None -> None
          | Some init ->
              Some
                (StateBindingsSet.fold
                   (fun elt acc ->
                     let elt_with_bindings = add_bindings elt in
                     inter Zs.join acc elt_with_bindings)
                   (StateBindingsSet.remove init joinset)
                   (add_bindings init)))
    with Division_by_zero -> None

  let join_and_widen _ old set =
    match join set with
    | None -> None
    | Some j -> Some (SSAExpr.Map.idempotent_union (fun _ l r -> Zs.widen l r) old j)

  let entrypoint vars =
    SSAExpr.Map.of_list (List.map (fun v -> (SSAExpr.var (ssavar_init v), Zs.top)) vars)

  let is_included _ a b = SSAExpr.Map.for_all (fun expr z -> Zs.subseteq z (expr_forwards b expr)) a
end

module SSAQuery (Zs : SingleValueAbstraction) = struct
  include SSA (Zs)

  let id = "QN"

  let assume expr state =
    match query_truth_value' state expr with
    | False -> (* e is zero, guard is always false *) None
    | True -> Some state (* Non-zero constant *)
    | TrueFalse -> assume expr state
end
