(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Standard numeric domains, both for our classical and SSA programs.
    They all depend on a {!Single_value_abstraction.Sig.SingleValueAbstraction},
    used to represent values. *)

open Ast.Program
open Sig
open Single_value_abstraction.Sig

val set_nb_backwards_iter : int -> unit
(** Set the number of backwards iteration to perform when learning from an
    assumption. *)

(** Standard non-relational numeric domain in abstract interpration:
    each variable is mapped to an abstraction of their values.
    To keep the map small, variable with no value ({!Zs.top}) are removed from
    the map.

    Its command line id is [N] *)
module Classical (Zs : SingleValueAbstraction) : sig
  include Classical_Domain with type relation = relation and type State.t = Zs.t VarMap.t

  val expr_forwards : State.t -> VarExpr.t -> Zs.t
  (** [expr_forwards state expr] returns the best abstraction for the value of
      [expr] in the given [state]. *)
end

(** Same as {!Classical}, but performs a few additional simplifications
    (removing [if 0], or [x := e] when [expr_forwards state (x-e)] is zero)

    Its command line id is [QN] *)
module ClassicalQuery (Zs : SingleValueAbstraction) : sig
  include Classical_Domain with type relation = relation and type State.t = Zs.t VarMap.t

  val expr_forwards : State.t -> VarExpr.t -> Zs.t
  (** [expr_forwards state expr] returns the best abstraction for the value of
      [expr] in the given [state]. *)
end

(** Non relational numeric abstraction for SSA.
    Unlike {!Classical}, which only remembers abstractions for variables, this
    also remembers abstractions for sub-expressions, which is possible at a
    reasonable cost thanks to the static assignement properties of SSA.

    Its command line id is [N] (under [Lift] or [LiftGVN]). *)
module SSA (Zs : SingleValueAbstraction) : sig
  include SSA_Domain with type SSA_State.t = Zs.t SSAExpr.Map.t

  val expr_forwards : SSA_State.t -> SSAExpr.t -> Zs.t
  (** [expr_forwards state expr] returns the best abstraction for the value of
      [expr] in the given [state]. *)
end

(** Same as {!SSA}, but performs a few additional simplifications
    (removing "if 0" branches).

    Its command line id is [QN] (under [Lift] or [LiftGVN]). *)
module SSAQuery (Zs : SingleValueAbstraction) : sig
  include SSA_Domain with type SSA_State.t = Zs.t SSAExpr.Map.t

  val expr_forwards : SSA_State.t -> SSAExpr.t -> Zs.t
  (** [expr_forwards state expr] returns the best abstraction for the value of
      [expr] in the given [state]. *)
end
