(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Sig
open Std_extra
open Types
open Ast.Program

let ( let* ) = Option.bind

module Classical (L : Classical_Domain) (R : Classical_Domain with type relation = L.relation) =
struct
  let id = L.id ^ "*" ^ R.id
  let name fmt () = Format.fprintf fmt "%s(@[%a@])(@[%a@]}" "Prod.Classical" L.name () R.name ()
  let state_name = "Product state"
  let doc = "Product of two classical domains"

  module State = struct
    include Instances.InstanciateBoth (Instances.Pair) (L.State) (R.State)

    let equal (l1, r1) (l2, r2) = R.State.equal r1 r2 && L.State.equal l1 l2

    let pp fmt (l, r) =
      Format.fprintf fmt "@[<v>%s:@,  @[<v>%a@]@,%s:@,  @[<v>%a@]@]" L.state_name L.State.pp l
        R.state_name R.State.pp r
  end
  module StateSet = Utils.Set (State)
  type relation = L.relation

  let entrypoint vars = (L.entrypoint vars, R.entrypoint vars)
  let is_included loc (l1, r1) (l2, r2) = L.is_included loc l1 l2 && R.is_included loc r1 r2

  let apply rel (l, r) =
    let* l = L.apply rel l in
    let* r = R.apply rel r in
    Some (l, r)

  let split_sets lf rf set =
    let lset, rset =
      StateSet.fold
        (fun (l, r) (lset, rset) -> (L.StateSet.add l lset, R.StateSet.add r rset))
        set
        (L.StateSet.empty, R.StateSet.empty)
    in
    let* l = lf lset in
    let* r = rf rset in
    Some (l, r)

  let join set = split_sets L.join R.join set
  let join_and_widen loc (old_l, old_r) set =
    split_sets (L.join_and_widen loc old_l) (R.join_and_widen loc old_r) set

  let query_is_constant =
    match (L.query_is_constant, R.query_is_constant) with
    | None, None -> None
    | None, Some f -> Some (fun (_, r) expr -> f r expr)
    | Some f, None -> Some (fun (l, _) expr -> f l expr)
    | Some fl, Some fr ->
        Some
          (fun (l, r) expr ->
            match fl l expr with
            | Some _ as x -> x
            | None -> fr r expr)

  let query_truth_value =
    match (L.query_truth_value, R.query_truth_value) with
    | None, None -> None
    | None, Some f -> Some (fun (_, r) expr -> f r expr)
    | Some f, None -> Some (fun (l, _) expr -> f l expr)
    | Some fl, Some fr ->
        Some
          (fun (l, r) expr ->
            match fl l expr with
            | TrueFalse -> fr r expr
            | x -> x)
end

module SSA (L : SSA_Domain) (R : SSA_Domain) = struct
  let id = L.id ^ "*" ^ R.id
  let name fmt () = Format.fprintf fmt "%s(@[%a@])(@[%a@]}" "Prod.SSA" L.name () R.name ()
  let state_name = "SSA product state"
  let doc = "Product of two SSA domains"

  module SSA_State : TYPE with type t = L.SSA_State.t * R.SSA_State.t = struct
    include Instances.InstanciateBoth (Instances.Pair) (L.SSA_State) (R.SSA_State)

    let equal (l1, r1) (l2, r2) = R.SSA_State.equal r1 r2 && L.SSA_State.equal l1 l2

    let pp fmt (l, r) =
      Format.fprintf fmt "@[<v>%s:@,  @[<v>%a@]@,%s:@,  @[<v>%a@]@]" L.state_name L.SSA_State.pp l
        R.state_name R.SSA_State.pp r
  end

  module StateBindings = MkBindings (SSA_State)
  module StateBindingsSet = Utils.Set (StateBindings)

  let assume expr (l, r) =
    let* l = L.assume expr l in
    let* r = R.assume expr r in
    Some (l, r)

  let split_sets lf rf set =
    let lset, rset =
      StateBindingsSet.fold
        (fun ((l, r), b) (lset, rset) ->
          (L.StateBindingsSet.add (l, b) lset, R.StateBindingsSet.add (r, b) rset))
        set
        (L.StateBindingsSet.empty, R.StateBindingsSet.empty)
    in
    let* l = lf lset in
    let* r = rf rset in
    Some (l, r)

  let join set = split_sets L.join R.join set
  let join_and_widen loc (old_l, old_r) set =
    split_sets (L.join_and_widen loc old_l) (R.join_and_widen loc old_r) set

  let entrypoint vars = (L.entrypoint vars, R.entrypoint vars)
  let is_included loc (l1, r1) (l2, r2) = L.is_included loc l1 l2 && R.is_included loc r1 r2

  let query_is_constant =
    match (L.query_is_constant, R.query_is_constant) with
    | None, None -> None
    | None, Some f -> Some (fun (_, r) expr -> f r expr)
    | Some f, None -> Some (fun (l, _) expr -> f l expr)
    | Some fl, Some fr ->
        Some
          (fun (l, r) expr ->
            match fl l expr with
            | Some _ as x -> x
            | None -> fr r expr)

  let query_truth_value =
    match (L.query_truth_value, R.query_truth_value) with
    | None, None -> None
    | None, Some f -> Some (fun (_, r) expr -> f r expr)
    | Some f, None -> Some (fun (l, _) expr -> f l expr)
    | Some fl, Some fr ->
        Some
          (fun (l, r) expr ->
            match fl l expr with
            | TrueFalse -> fr r expr
            | x -> x)
end
