(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Common signatures of IMP {!Classical_Domain} and of {!SSA_Domain} *)

open Std_extra.Types
open Ast.Program

(** {1 Domain signatures}                                                     *)
(******************************************************************************)

(** Documentation information used for pretty-printing *)
module type Doc = sig
  val id : string
  (** Unique identifier for the domain, eg "N", "FA",... *)

  val name : Format.formatter -> unit -> unit
  (** Full human readable name *)

  val state_name : string
  (** Name of the domain's states *)

  val doc : string
  (** Long description of the domain *)
end

(** The type for a Classical (IMP) domain. *)
module type Classical_Domain = sig
  include Doc

  type relation
  (** program relations (edges in the CFG), typically {!Ast.Program.relation} *)

  module State : TYPE
  (** Abstract state *)

  module StateSet : SET with type elt = State.t
  (** Set of abstract states *)

  val apply : relation -> State.t -> State.t option
  (** None means state is unreacheble (eg: if false ...)
      @raises Division_by_zero *)

  val join : StateSet.t -> State.t option
  (** joins the states given as arguments
      - [None] on emptyset
      - single element on singleton *)

  val entrypoint : Var.t list -> State.t
  (** [entrypoint] take the list of variables bound at the start of the program
      as argument, and returns the abstraction for the program entry *)

  val join_and_widen : Loc.t -> State.t -> StateSet.t -> State.t option
  (** [join_and_widen l old set] is [widen loc old (join set)]
      Merged in a single operation as it avoids renamings in SSA.Lift *)

  val is_included : Loc.t -> State.t -> State.t -> bool
  (** State inclusion operation used to detect convergence.
      This was absent from the paper *)

  (** {2 Queries}                                                             *)

  val query_truth_value : (State.t -> VarExpr.t -> Single_value_abstraction.Ternary.t) option
  (** Possible truth value of the given expression.
      [None] when the domain provides no useful data *)

  val query_is_constant : (State.t -> VarExpr.t -> Z.t option) option
  (** [query_is_constant state expr] is [Some z] only if we can prove that
      [expr] is equal to constant [z] in the given state.
      [None] when the domain provides no useful data *)
end

(** This is almost the same as [Classical_Domain] except that:
    - the [join] operation takes a set of { list bindings; state }
      instead of just a set of states
    - one can only [apply] guards, thus the [apply] operation is renamed [assume]. *)
module type SSA_Domain = sig
  include Doc

  (** {2 Domain types}                                                        *)

  module SSA_State : TYPE

  module StateBindingsSet : SET with type elt = SSA_State.t * Bindings.t
  (** Set of states and bindings (phi variables). *)

  (** {2 Domain operations}                                                        *)

  val assume : SSAExpr.t -> SSA_State.t -> SSA_State.t option
  (** similar to [apply], [assume e s] returns the state obtained from [s] when
      assuming that [e] is non-zero. [None] means [e] is always zero
      @raises Division_by_zero *)

  val join : StateBindingsSet.t -> SSA_State.t option
  (** Join the given states.
      - [None] on empty set
      - the unique element on the singleton set (bindings are discarded) *)

  val entrypoint : Var.t list -> SSA_State.t
  (** Domain entrypoint, with a list of initial variables in scope *)

  val is_included : Loc.t -> SSA_State.t -> SSA_State.t -> bool
  (** [is_included l s1 s2] is [true] is [s1] included in [s2] according to the
      domain order at [l]. Used to know the fixpoint has been reached. *)

  val join_and_widen : Loc.t -> SSA_State.t -> StateBindingsSet.t -> SSA_State.t option
  (** Performs both join and widening simultaneously.
      This simplifies numbering the new states. *)

  (** {2 Queries}                                                             *)

  val query_truth_value : (SSA_State.t -> SSAExpr.t -> Single_value_abstraction.Ternary.t) option
  (** Possible truth value of the given expression
      [None] when the domain provides no useful data *)

  val query_is_constant : (SSA_State.t -> SSAExpr.t -> Z.t option) option
  (** [query_is_constant state expr] is [Some z] only if we can prove that
      [expr] is equal to constant [z] in the given state
      [None] when the domain provides no useful data *)
end

(** Signature for domains that can generate new program graphs *)
module type Graph = sig
  include PRINTABLE

  type state
  (** The free algebra state, it is often a part of the full domain state
      (eg left/right projection in a product).
      To account for this difference, graph operation take a [cast] argument,
      casting from the full state to the free algebra substate. *)

  val size : t -> int
  (** Number of nodes in the graph *)

  val dotty_printer : Format.formatter -> t -> unit
  (** Print the graph in dotty format *)

  val make :
    cast:('a -> state) -> join_preds:(Loc.t -> 'a LocMap.t -> 'a option) -> 'a LocMap.t -> t
  (** Generate the graph *)

  val assertions_and_dead_ends : cast:('a -> state) -> t -> 'a LocMap.t -> int * int
  (** Returns the number of assertions (nodes with a single guarded child)
      and dead ends (nodes with no children) in the graph *)
end

(** {1 Functor signatures}                                                    *)
(******************************************************************************)

(** Type of abitrary program transformation on top of an abstract domain.
    They typically override [S]'s [apply] function

    They can also change the type of relations used (for instance, to simplify
    the type of expressions by removing some operators)

    Note that this type implies the transformation doesn't know the type of
    [S.State.t]. It is opaque, meaning it:
    - can only be inspected through [query_truth_value] and [query_is_constant]
    - can only be built using domain operations [apply], [join] (and [entrypoint]) *)
module type Transformation = sig
  type relation_in
  type relation_out

  val id : string
  (** Command line identifier *)

  val doc : string
  (** Command line documentation *)

  (** The actual transformation functor *)
  module Domain (S : Classical_Domain with type relation = relation_out) :
    Classical_Domain
      with module State = S.State
       and module StateSet = S.StateSet
       and type relation = relation_in
end

module type Simple_Transformation =
  Transformation with type relation_in = relation and type relation_out = relation

(** Type of abitrary program transformation on top of an SSA abstract domain.
    They typically override [S]'s [assume] function

    Note that this type implies the transformation doesn't know the type of
    [S.SSA_State.t]. It is opaque, meaning it:
    - can only be inspected through [query_truth_value] and [query_is_constant]
    - can only be built using domain operations [apply], [join] (and [entrypoint]) *)
module type SSA_Transformation = sig
  val id : string
  (** Command line identifier *)

  val doc : string
  (** Command line documentation *)

  (** The actual transformation functor *)
  module Domain (S : SSA_Domain) :
    SSA_Domain with module SSA_State = S.SSA_State and module StateBindingsSet = S.StateBindingsSet
end
