(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Sig
open Ast.Program
open Ast.Expr
(* open Std_extra *)

let ( let* ) = Option.bind

module Ternary = struct
  type relation_in = relation
  type relation_out = relation

  let id = "T"
  let doc = "Replace ternary expressions with branching ifs"

  module Domain (S : Classical_Domain with type relation = relation) = struct
    include S

    let join_bin_opt a b =
      match (a, b) with
      | Some a, Some b -> S.join (S.StateSet.of_list [ a; b ])
      | None, (Some _ as x) | (Some _ as x), None -> x
      | None, None -> None

    let neg x = VarExpr.unop Neg x

    (** Returns three values:
          - the new state
          - the new replacement expression
          - a boolean true when a rewrite was performed *)
    let rec rewrite_ite_deep nb state (expr : VarExpr.t) =
      match expr.node with
      | E_Const _ | E_Var _ -> Some (state, expr, false)
      | E_Unop (u, e) ->
          let* state, e, r = rewrite_ite_deep nb state e in
          Some (state, VarExpr.unop u e, r)
      | E_Binop (l, b, r) ->
          let* state, l, r1 = rewrite_ite_deep nb state l in
          let* state, r, r2 = rewrite_ite_deep (if r1 then nb + 1 else nb) state r in
          Some (state, VarExpr.binop b l r, r1 || r2)
      | E_IfThenElse (if_, then_, else_) ->
          let* state, if_, _ = rewrite_ite_deep nb state if_ in
          let var = tmp_var nb in
          let left_branch =
            let* state = S.apply (Guard if_) state in
            let* state, then_, _ = rewrite_ite_deep nb state then_ in
            S.apply (Assign (var, then_)) state
          in
          let right_branch =
            let* state = S.apply (Guard (neg if_)) state in
            let* state, else_, _ = rewrite_ite_deep nb state else_ in
            S.apply (Assign (var, else_)) state
          in
          let* state = join_bin_opt left_branch right_branch in
          Some (state, VarExpr.var var, true)

    (** toplevel rewrite, doesn't need temporary variables *)
    let rec rewrite_ite_shallow rel state =
      match rel with
      | Assign (x, e) -> (
          match e.node with
          | E_IfThenElse (if_, then_, else_) ->
              let left_branch =
                let* state = rewrite_ite_shallow (Guard if_) state in
                rewrite_ite_shallow (Assign (x, then_)) state
              in
              let right_branch =
                let* state = rewrite_ite_shallow (Guard (neg if_)) state in
                rewrite_ite_shallow (Assign (x, else_)) state
              in
              join_bin_opt left_branch right_branch
          | _ ->
              let* state, e, _ = rewrite_ite_deep 0 state e in
              S.apply (Assign (x, e)) state)
      | Guard e -> (
          match e.node with
          | E_IfThenElse (if_, then_, else_) ->
              let left_branch =
                let* state = rewrite_ite_shallow (Guard if_) state in
                rewrite_ite_shallow (Guard then_) state
              in
              let right_branch =
                let* state = rewrite_ite_shallow (Guard (neg if_)) state in
                rewrite_ite_shallow (Guard else_) state
              in
              join_bin_opt left_branch right_branch
          | _ ->
              let* state, e, _ = rewrite_ite_deep 0 state e in
              S.apply (Guard e) state)

    let apply rel state = rewrite_ite_shallow rel state
  end
end

module DivGuard = struct
  type relation_in = relation
  type relation_out = relation

  let id = "D"
  let doc = "Adds an assertion d!=0 before every division"

  module Domain (S : Classical_Domain with type relation = relation) = struct
    (* We use a set [seen] to avoid duplicate and a list [found] to have terms
       in the correct order (i.e. subterms first, as they get evaluated first) *)

    include S

    let rec divisors seen found (e : VarExpr.t) =
      match e.node with
      | E_Var _ | E_Const _ -> (seen, found)
      | E_Unop (_, e) -> divisors seen found e
      | E_Binop (l, b, r) ->
          let add_r = (b = Div || b = Mod) && not (VarExpr.Set.mem r seen) in
          let seen, found =
            if add_r then
              divisors (VarExpr.Set.add r seen) (Guard (VarExpr.binop Ne r VarExpr.zero) :: found) l
            else divisors seen found l
          in
          divisors seen found r
      | E_IfThenElse (i, t, e) ->
          let seen, found = divisors seen found i in
          let seen, found = divisors seen found t in
          divisors seen found e

    let rec apply_n state = function
      | [] -> Some state
      | r :: rs ->
          let* state = S.apply r state in
          apply_n state rs

    let apply rel state =
      let expr = expr_of_rel rel in
      let _, guard = divisors VarExpr.Set.empty [ rel ] expr in
      apply_n state guard
  end
end

module ShiftGuard = struct
  type relation_in = relation
  type relation_out = relation

  let id = "S"
  let doc = "Adds an assertion s>=0 before right/left shift by s"

  module Domain (S : Classical_Domain with type relation = relation) = struct
    include S

    let rec shifters seen found (e : VarExpr.t) =
      match e.node with
      | E_Var _ | E_Const _ -> (seen, found)
      | E_Unop (_, e) -> shifters seen found e
      | E_Binop (l, b, r) ->
          let add_r = (b = Shiftlt || b = Shiftrt) && not (VarExpr.Set.mem r seen) in
          let seen, found =
            if add_r then
              shifters (VarExpr.Set.add r seen) (Guard (VarExpr.binop Ge r VarExpr.zero) :: found) l
            else shifters seen found l
          in
          shifters seen found r
      | E_IfThenElse (i, t, e) ->
          let seen, found = shifters seen found i in
          let seen, found = shifters seen found t in
          shifters seen found e

    let rec apply_n state = function
      | [] -> Some state
      | r :: rs ->
          let* state = S.apply r state in
          apply_n state rs

    let apply rel state =
      let expr = expr_of_rel rel in
      let _, guard = shifters VarExpr.Set.empty [ rel ] expr in
      apply_n state guard
  end
end

(* module LazyBooleans = struct
     type relation_in = relation
     type relation_out = relation

     let id = "B"
     let doc = "Replace lazy booleans with explicit control-flow jumps"

     module Domain (S : Classical_Domain with type relation = relation) = struct
       include S

       let add_opt x set =
         match x with
         | None -> set
         | Some x -> StateSet.add x set

       let rec rewrite_lazybool_shallow rel state =
         match rel with
         | Guard x -> (
             match x.node with
             | E_Binop (l, LOr, r) ->
                 let l' = rewrite_lazybool_shallow (Guard l) state in
                 let r' = rewrite_lazybool_shallow (Guard (VarExpr.unop LNot l)) state in
                 let r' = Option.bind r' (fun r' -> rewrite_lazybool_shallow (Guard r) r') in
                 let set = add_opt l' (add_opt r' StateSet.empty) in
                 join set
             | E_Binop (l, LAnd, r) ->
                 let l' = rewrite_lazybool_shallow (Guard l) state in
                 let r' = rewrite_lazybool_shallow (Guard (VarExpr.unop LNot l)) state in
                 let r' = Option.bind r' (fun r' -> rewrite_lazybool_shallow (Guard r) r') in
                 let set = add_opt l' (add_opt r' StateSet.empty) in
                 join set
             | _ -> apply rel state)
         | _ -> apply rel state
     end
   end *)

module Query = struct
  type relation_in = relation
  type relation_out = relation

  let id = "Q"
  let doc =
    "Perform simplification using simple queries. Also directly inlined "
    ^ "in Product.ClassicalQueryBoth and Numeric.ClassicalQuery"

  module Domain (S : Classical_Domain with type relation = relation) = struct
    include S

    let assign_with_query query_is_constant rel x e state =
      let diff = VarExpr.binop MinusA e (VarExpr.var x) in
      match query_is_constant state diff with
      | Some z -> (
          if (* x - e is zero, so x := e is noop *)
             Z.equal z Z.zero then Some state
          else
            match query_is_constant state e with
            | Some z ->
                (* e is constant z, so replace x := e with x := z *)
                apply (Assign (x, VarExpr.of_const z)) state
            | None ->
                (* e - x is constant z, so replace x := e with x := z + x *)
                apply (Assign (x, VarExpr.(binop PlusA (of_const z) (var x)))) state)
      | None -> (
          match query_is_constant state e with
          | Some z ->
              (* e is constant z, so replace x := e with x := z *)
              apply (Assign (x, VarExpr.of_const z)) state
          | None -> apply rel state)

    let guard_with_query query_truth_value rel e state =
      match query_truth_value state e with
      | Single_value_abstraction.Ternary.False -> (* e is zero, guard is always false *) None
      | True -> Some state (* Non-zero constant *)
      | TrueFalse -> apply rel state

    let apply =
      match (query_is_constant, query_truth_value) with
      | None, None -> apply
      | Some query, None -> (
          fun rel state ->
            match rel with
            | Assign (x, e) -> assign_with_query query rel x e state
            | Guard _ -> apply rel state)
      | None, Some query -> (
          fun rel state ->
            match rel with
            | Assign _ -> apply rel state
            | Guard e -> guard_with_query query rel e state)
      | Some q1, Some q2 -> (
          fun rel state ->
            match rel with
            | Assign (x, e) -> assign_with_query q1 rel x e state
            | Guard e -> guard_with_query q2 rel e state)
  end
end

module SimplifyExpression = struct
  type relation_in = relation
  type relation_out = relation

  let id = "E"
  let doc = "Simplify expressions using simple rewrites and queries"

  module Domain (S : Classical_Domain with type relation = relation) = struct
    include S

    let q_is_constant =
      match query_is_constant with
      | None -> fun _ e -> Ast.Expr.is_constant e
      | Some f -> f

    let apply rel state =
      match rel with
      | Guard e -> apply (Guard (VarExpr.simplify_expr (q_is_constant state) e)) state
      | Assign (x, e) -> apply (Assign (x, VarExpr.simplify_expr (q_is_constant state) e)) state
  end
end

module SSA_SimplifyExpression = struct
  let id = "E"
  let doc = "Simplify expressions using simple rewrites and queries"

  module Domain (S : SSA_Domain) = struct
    include S

    let q_is_constant =
      match query_is_constant with
      | None -> fun _ e -> Ast.Expr.is_constant e
      | Some f -> f

    let assume expr state = assume (SSAExpr.simplify_expr (q_is_constant state) expr) state
  end
end

module SSA_Query = struct
  let id = "Q"
  let doc =
    "Perform simplification using simple queries. Also directly inlined "
    ^ "in Product.SSAQueryBoth and Numeric.SSAQuery"

  module Domain (S : SSA_Domain) = struct
    include S

    let assume =
      match query_truth_value with
      | None -> assume
      | Some query_truth_value -> (
          fun expr state ->
            match query_truth_value state expr with
            | False -> (* e is zero, guard is always false *) None
            | True -> Some state (* Non-zero constant *)
            | TrueFalse -> assume expr state)
  end
end

(* module GVN (S : SSA_Domain) = struct
     let id = "GVN(" ^ S.id ^ ")"
     let name fmt () = Format.fprintf fmt "GVN(%a)" S.name ()
     let state_name = "SSA state with GVN"
     let doc = "Global value numbering functor for SSA"

     module SSA_State = struct
       type t = S.SSA_State.t * SSAVar.t SSAVarMap.t [@@deriving eq, ord]
       let order a b = Utils.order_of_compare (compare a b)
       let hash x = S.SSA_State.hash (fst x)
       let pp fmt (st, var) =
         Format.fprintf fmt "@[@[State: %a@]@ @[GVN: %a@]@]" S.SSA_State.pp st
           (SSAVarMap.pp_map ~pp_sep:Format.pp_print_space (fun fmt v v' ->
                Format.fprintf fmt "@[%a -> %a@]" SSAVar.pp v SSAVar.pp v'))
           var
     end

     module StateBindings = MkBindings (SSA_State)

     module StateBindingsSet = Utils.Set (StateBindings)

     let rec rename_vars renames v =
       match SSAVarMap.find v renames with
       | v -> rename_vars renames v
       | exception Not_found -> v

     let replace_variables expr (_, renames) = Ast.Expr.map SSAExpr.of_node (rename_vars renames) expr

     let entrypoint vars = (S.entrypoint vars, SSAVarMap.empty)
     let assume expr ((s, m) as state) = (S.assume (replace_variables expr state) s, m)

     let is_included loc (sl, ml) (sr, mr) =
       S.is_included loc sl sr
       && SSAVarMap.reflexive_same_domain_for_all2 (fun _ -> SSAVar.equal) ml mr

     let gvn state_bindings_set =
       let hash = Lift.HashTable.create 20 in
       let n = StateBindingsSet.cardinal state_bindings_set in
       let placeholder_expr = SSAExpr.of_int 0 in
       let new_var_store = ref new_var_store in
       let phi_vars =
         StateBindingsSet.filter
           (fun v ->
             (* Store all previous values in an array *)
             let arr = Array.make n placeholder_expr in
             let _ =
               StateBindingsSet.fold
                 (fun (state, bindings) i ->
                   arr.(i) <- SSAVarMap.find v bindings;
                   i + 1)
                 state_bindings_set 0
             in
             match Lift.HashTable.find hash arr with
             | v' ->
                 new_var_store := SSAVarMap.add v v' !new_var_store;
                 false
             | exception Not_found ->
                 new_var_store := VarMap.add v (SSAExpr.var (ssavar_join v id)) !new_var_store;
                 Lift.HashTable.replace hash arr v;
                 true)
           phi_vars

     let query_is_constant ((s, _) as state) expr =
       S.query_is_constant s (replace_variables expr state)

     let query_truth_value ((s, _) as state) expr =
       S.query_truth_value s (replace_variables expr state)
   end *)
