(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Std_extra
open Ast

let id = "I"
let name fmt () = Format.fprintf fmt "%s" "Interval"
let state_name = "Interval"
let doc = "Interval domain"

type t = {
  low : Z.t option; (* Use NONE for empty *)
  high : Z.t option;
}
[@@deriving eq, ord]

let order a b = Utils.order_of_compare (compare a b)

let interval low high =
  match (low, high) with
  | Some l, Some h when l > h -> None
  | _, _ -> Some { low; high }

let hash { low; high } =
  Utils.hash_combine (Instances.Option.hash Z.to_int low) (Instances.Option.hash Z.to_int high)

let pp_opt fmt = function
  | None -> Format.fprintf fmt "inf"
  | Some z -> Z.pp_print fmt z

let pp fmt { low; high } = Format.fprintf fmt "@{<cyan>@[[%a:%a]@]@}" pp_opt low pp_opt high

let option_and f a b =
  match (a, b) with
  | Some a, Some b -> Some (f a b)
  | _ -> None

(* let option_or f a b =
   match (a, b) with
   | Some a, Some b -> Some (f a b)
   | (Some _ as c), _ | _, (Some _ as c) -> c
   | _ -> None *)

let low_le a b =
  match (a, b) with
  | None, _ -> true
  | _, None -> false
  | Some a, Some b -> Z.leq a b

let low_lt a b =
  match (a, b) with
  | _, None -> false
  | None, _ -> true
  | Some a, Some b -> Z.lt a b

let high_le a b =
  match (a, b) with
  | _, None -> true
  | None, _ -> false
  | Some a, Some b -> Z.leq a b

let high_lt a b =
  match (a, b) with
  | None, _ -> false
  | _, None -> true
  | Some a, Some b -> Z.lt a b

let top = { low = None; high = None }

let join l r =
  {
    low = (if low_le l.low r.low then l.low else r.low);
    high = (if high_le l.high r.high then r.high else l.high);
  }
let meet l r =
  interval
    (if low_le l.low r.low then r.low else l.low)
    (if high_le l.high r.high then l.high else r.high)
let widen a b =
  {
    low = (if low_lt b.low a.low then None else a.low);
    high = (if high_lt a.high b.high then None else a.high);
  }

let subseteq l r = low_le r.low l.low && high_le l.high r.high
let contains i z =
  let z = Some z in
  low_le i.low z && high_le z i.high

open Expr

let is_singleton = function
  | { low = Some l as c; high = Some h } -> if Z.equal l h then c else None
  | _ -> None

(** Checks that both intervals are constant and equal *)
let equal_const l r =
  match is_singleton l with
  | None -> false
  | Some l -> (
      match is_singleton r with
      | None -> false
      | Some r -> Z.equal l r)

type interval_class =
  | Zero
  | Positive of Z.t * Z.t option (* May include zero *)
  | Negative of Z.t option * Z.t (* May include zero *)
  | MixedSign of Z.t option * Z.t option

let const z =
  let z = Some z in
  { low = z; high = z }

let zero = const Z.zero
let one = const Z.one
let zero_one = { low = Some Z.zero; high = Some Z.one }
let non_zero = top

let interval_class i =
  if equal i zero then Zero
  else if low_le (Some Z.zero) i.low then Positive (Option.get i.low, i.high)
  else if high_le i.high (Some Z.zero) then Negative (i.low, Option.get i.high)
  else MixedSign (i.low, i.high)

let interval_product l r =
  match (interval_class l, interval_class r) with
  | Zero, _ | _, Zero -> zero
  (* Positive*positive and negative*negative *)
  | Positive (a, None), Positive (c, _)
  | Positive (a, _), Positive (c, None)
  | Negative (None, a), Negative (_, c)
  | Negative (_, a), Negative (None, c) ->
      { low = Some Z.(a * c); high = None }
  | Positive (a, Some b), Positive (c, Some d) -> { low = Some Z.(a * c); high = Some Z.(b * d) }
  | Negative (Some a, b), Negative (Some c, d) -> { low = Some Z.(b * d); high = Some Z.(a * c) }
  (* Various products reach top *)
  | MixedSign (None, None), _
  | _, MixedSign (None, None)
  | MixedSign (None, _), MixedSign (_, _)
  | MixedSign (_, None), MixedSign (_, _)
  | MixedSign (_, _), MixedSign (None, _)
  | MixedSign (_, _), MixedSign (_, None)
  | Positive (_, None), MixedSign (_, _)
  | MixedSign (_, _), Positive (_, None)
  | Negative (None, _), MixedSign (_, _)
  | MixedSign (_, _), Negative (None, _) ->
      top
  (* Pos neg product *)
  | Positive (a, None), Negative (_, b)
  | Positive (a, _), Negative (None, b)
  | Negative (_, b), Positive (a, None)
  | Negative (None, b), Positive (a, _) ->
      { low = None; high = Some Z.(a * b) }
  | Positive (a, Some b), Negative (Some c, d) | Negative (Some c, d), Positive (a, Some b) ->
      { low = Some Z.(b * c); high = Some Z.(a * d) }
  (* Pos mixed product *)
  | Positive (_, Some b), MixedSign (None, Some d) | MixedSign (None, Some d), Positive (_, Some b)
    ->
      { low = None; high = Some Z.(b * d) }
  | Positive (_, Some b), MixedSign (Some c, None) | MixedSign (Some c, None), Positive (_, Some b)
    ->
      { low = Some Z.(b * c); high = None }
  | Positive (_, Some b), MixedSign (Some c, Some d)
  | MixedSign (Some c, Some d), Positive (_, Some b) ->
      { low = Some Z.(b * c); high = Some Z.(b * d) }
  | Negative (Some a, _), MixedSign (None, Some d) | MixedSign (None, Some d), Negative (Some a, _)
    ->
      { low = Some Z.(a * d); high = None }
  | Negative (Some a, _), MixedSign (Some c, None) | MixedSign (Some c, None), Negative (Some a, _)
    ->
      { low = None; high = Some Z.(a * c) }
  | Negative (Some a, _), MixedSign (Some c, Some d)
  | MixedSign (Some c, Some d), Negative (Some a, _) ->
      { low = Some Z.(a * d); high = Some Z.(a * c) }
  | MixedSign (Some a, Some b), MixedSign (Some c, Some d) ->
      { low = Some Z.(min (a * d) (b * c)); high = Some Z.(max (a * c) (b * d)) }

let of_ternary = function
  | Ternary.True -> one
  | Ternary.False -> zero
  | Ternary.TrueFalse -> zero_one

let to_ternary i =
  match is_singleton i with
  | Some z when Z.equal z Z.zero -> Ternary.False
  | Some _ -> Ternary.True
  | None -> if contains i Z.zero then Ternary.TrueFalse else Ternary.True

let may_be_null i = contains i Z.zero

module Forwards = struct
  let neg i = { low = Option.map Z.( ~- ) i.high; high = Option.map Z.( ~- ) i.low }
  let lnot i = if may_be_null i then if i = zero then one else zero_one else zero

  let const = const
  let unop u i =
    match u with
    | Neg -> neg i
    | LNot -> lnot i
    | BNot -> (
        match is_singleton i with
        | Some x -> const Z.(~!x)
        | None -> top)

  let eq il ir =
    if Option.is_none (meet il ir) then Ternary.False
    else if equal_const il ir then Ternary.True
    else Ternary.TrueFalse

  let leq il ir =
    match (il.low, il.high, ir.low, ir.high) with
    | _, Some h, Some l, _ when Z.leq h l -> Ternary.True
    | Some l, _, _, Some h when Z.lt h l -> Ternary.False
    | _ -> Ternary.TrueFalse

  let add il ir =
    { low = option_and Z.( + ) il.low ir.low; high = option_and Z.( + ) il.high ir.high }
  let sub il ir =
    { low = option_and Z.( - ) il.low ir.high; high = option_and Z.( - ) il.high ir.low }

  let div_range x ymn ymx =
    match (x.low, x.high) with
    | Some xmn, Some xmx ->
        let c1 = Z.div xmn ymn in
        let c2 = Z.div xmx ymn in
        let c3 = Z.div xmn ymx in
        let c4 = Z.div xmx ymx in
        let min = Z.min (Z.min c1 c2) (Z.min c3 c4) in
        let max = Z.max (Z.max c1 c2) (Z.max c3 c4) in
        { low = Some min; high = Some max }
    | _ -> top

  let div x y =
    match (y.low, y.high) with
    | None, _ | _, None -> top
    | Some min, Some max -> (
        let result_pos =
          if Z.gt max Z.zero then
            let lpos = if Z.gt min Z.zero then min else Z.one in
            Some (div_range x lpos max)
          else None
        in
        let result_neg =
          if Z.lt min Z.zero then
            let gneg = if Z.lt max Z.zero then max else Z.minus_one in
            Some (div_range x min gneg)
          else None
        in
        match (result_neg, result_pos) with
        | Some n, Some p -> join n p
        | Some x, _ | _, Some x -> x
        | _, _ -> raise Division_by_zero)

  let modulo t1 t2 =
    match (t2.low, t2.high) with
    | None, _ | _, None -> top
    | Some min2, Some max2 ->
        (* if Z.equal max2 Z.zero then raise Division_by_zero (* completely undefined. *)
           else *)
        (* Result is of the sign of x. Also, compute |x| to bound the result *)
        let neg, pos, max_x =
          ( low_le t1.low (Some Z.minus_one),
            high_le (Some Z.one) t1.high,
            match (t1.low, t1.high) with
            | Some mn, Some mx -> Some (Z.max (Z.abs mn) (Z.abs mx))
            | _ -> None )
        in
        (* Bound the result: no more than |x|, and no more than |y|-1 *)
        let pos_rem = Z.max (Z.abs min2) (Z.abs max2) in
        let bound = Z.pred pos_rem in
        let bound = Option.fold ~some:(Z.min bound) ~none:bound max_x in
        (* Compute result bounds using sign information *)
        let min = if neg then Some (Z.neg bound) else Some Z.zero in
        let max = if pos then Some bound else Some Z.zero in
        (* TODO: inject_range_or_bottom? *)
        { low = min; high = max }

  let two_power_of_int k = Z.shift_left Z.one k

  let two_power n =
    let k = Z.to_int n in
    if k > 1024 then raise Z.Overflow else two_power_of_int k

  let positive_integers = { low = Some Z.zero; high = None }
  let negative_integers = { low = None; high = Some Z.zero }

  let shift_aux op (x : t) (y : t) =
    match meet { low = Some Z.zero; high = None } y with
    | None -> top
    | Some y -> (
        try
          let min_factor = Option.map two_power y.low in
          let max_factor = Option.map two_power y.high in
          let factor = { low = min_factor; high = max_factor } in
          op x factor
        with Z.Overflow ->
          if subseteq x positive_integers then positive_integers
          else if subseteq x negative_integers then negative_integers
          else top)

  let shift_right il ir = shift_aux div il ir
  let shift_left il ir = shift_aux interval_product il ir

  let binop b il ir =
    match b with
    | PlusA | PlusPI -> add il ir
    | MinusA | MinusPI | MinusPP -> sub il ir
    | Mult -> interval_product il ir
    | Div -> div il ir
    | Mod -> modulo il ir
    | Eq -> of_ternary (eq il ir)
    | Ne -> of_ternary (Ternary.not (eq il ir))
    | Lt -> of_ternary (Ternary.not (leq ir il))
    | Gt -> of_ternary (Ternary.not (leq il ir))
    | Le -> of_ternary (leq il ir)
    | Ge -> of_ternary (leq ir il)
    | LOr -> of_ternary Ternary.(to_ternary il || to_ternary ir)
    | LAnd -> of_ternary Ternary.(to_ternary il && to_ternary ir)
    | Shiftlt -> shift_left il ir
    | Shiftrt -> shift_right il ir (* Shift right is division, brings closer to zero *)
    | BAnd | BOr | BXor -> top

  let ternary cond if_true if_false =
    match to_ternary cond with
    | False -> if_false
    | True -> if_true
    | TrueFalse -> join if_true if_false
end

let ( let+ ) o f = Option.map f o

let remove_const i z =
  let low = if i.low = Some z then Some Z.(z + one) else i.low in
  let high = if i.high = Some z then Some Z.(z - one) else i.high in
  interval low high

module Backwards = struct
  let unop u i ~res =
    match u with
    | Neg -> Some (Forwards.neg res)
    | LNot ->
        if not (may_be_null res) then Some zero
        else if res = zero then remove_const i Z.zero
        else None
    | BNot ->
        let+ x = is_singleton res in
        const Z.(~!x)

  let incr x = Option.map (fun x -> Z.(x + one)) x
  let decr x = Option.map (fun x -> Z.(x - one)) x

  let no_info = (None, None)
  let is_zero i = equal i zero

  let eq il ir = function
    | Ternary.True -> (Some ir, Some il)
    | Ternary.TrueFalse -> no_info
    | Ternary.False ->
        let il' =
          match is_singleton ir with
          | None -> None
          | Some z -> remove_const il z
        in
        let ir' =
          match is_singleton il with
          | None -> None
          | Some z -> remove_const ir z
        in
        (il', ir')

  let le il ir = function
    | Ternary.True -> (Some { low = None; high = ir.high }, Some { low = il.low; high = None })
    | Ternary.TrueFalse -> no_info
    | Ternary.False ->
        (Some { low = incr ir.low; high = None }, Some { low = None; high = decr il.high })

  let binop b il ir ~res =
    match b with
    | PlusA | PlusPI ->
        (* l + r = res <=> l = res - r <=> r = res - l  *)
        (Some (Forwards.sub res ir), Some (Forwards.sub res il))
    | MinusA | MinusPI | MinusPP ->
        (* l - r = res <=> l = res + r <=> r = l - res *)
        (Some (Forwards.add res ir), Some (Forwards.sub il res))
    | Mult ->
        if is_zero il || is_zero ir then no_info
        else
          let res_zero = is_zero res in
          if res_zero then
            if not (contains il Z.zero) then (None, Some zero) (* l != 0 and l*r = 0 => r = 0 *)
            else if not (contains ir Z.zero) then (Some zero, None)
            else no_info
          else (Some (Forwards.div res ir), Some (Forwards.div res il))
    | Div ->
        (* il = ir * res + rem where rem < ir *)
        (None, None)
    | Mod -> no_info
    | Eq -> eq il ir (to_ternary res)
    | Ne -> eq il ir (Ternary.not (to_ternary res))
    | Lt -> Utils.swap (le ir il (Ternary.not (to_ternary res)))
    | Gt -> le il ir (Ternary.not (to_ternary res)) (* l > r <=> not l <= r *)
    | Le -> le il ir (to_ternary res)
    | Ge -> Utils.swap (le ir il (to_ternary res))
    | Shiftrt -> (Some (Forwards.shift_left res ir), None)
    | _ -> no_info

  let ternary cond if_true if_false ~res =
    match (meet res if_true, meet res if_false) with
    | None, None -> raise Utils.Bottom
    | (Some _ as x), None -> (remove_const cond Z.zero, x, None)
    | None, (Some _ as x) -> (Some zero, None, x)
    | (Some _ as a), (Some _ as b) -> (
        match to_ternary cond with
        | False -> (None, None, b)
        | True -> (None, a, None)
        | _ -> (None, None, None))
end
