(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(* A significant portion of this file is taken from the Codex plugin of Frama-C *)
open Std_extra

let id = "K"
let state_name = "Known bits"
let name fmt () = Format.fprintf fmt "Known_bits"
let doc = "Known bits domain: bits are marked as 0, 1, or ?"

type bitvalue =
  | One
  | Zero
  | Unknown

(* First bitvector = 0 where the bits must be 0, 1 otherwise.
   Second bitvector = 1 where the bits must be 1, 0 otherwise. *)
type t = {
  zeros : Z.t;
  ones : Z.t;
}
[@@deriving eq, ord]

let order a b = Utils.order_of_compare (compare a b)

let hash x = Std_extra.Utils.hash_combine (Z.hash x.zeros) (Z.hash x.ones)

let is_bottom x =
  let open Z in
  not @@ equal zero (~!(x.zeros) land x.ones)

let testbit x index =
  if Z.testbit x.zeros index then if Z.testbit x.ones index then One else Unknown else Zero

let pp fmt x =
  if is_bottom x then Format.fprintf fmt "<bottom>"
  else
    let size = max (Z.numbits x.zeros) (Z.numbits x.ones) in
    for i = size - 1 downto 0 do
      match testbit x i with
      | Zero -> Format.fprintf fmt "0"
      | One -> Format.fprintf fmt "1"
      | Unknown -> Format.fprintf fmt "?"
    done

let pp fmt x = Format.fprintf fmt "@{<cyan>@[0b%a@]@}" pp x

let singleton x = { zeros = x; ones = x }

let one = singleton Z.one
let zero = singleton Z.zero
let top = { zeros = Z.minus_one; ones = Z.zero }
let non_zero = top

let join0 = Z.( lor )
let join1 = Z.( land )
let join x y = { zeros = join0 x.zeros y.zeros; ones = join1 x.ones y.ones }
let meet x y =
  let res = { zeros = Z.(x.zeros land y.zeros); ones = Z.(x.ones lor y.ones) } in
  if is_bottom res then None else Some res

let zero_one = join zero one

let subseteq x y = equal y (join x y)

let contains x z = subseteq (singleton z) x

let widen prev next =
  {
    zeros =
      (let t = join0 prev.zeros next.zeros in
       if Z.equal prev.zeros t then prev.zeros
         (* Set all high bits of Z to unknown, but preserve a (finite) number of low bits *)
       else if Z.lt Z.zero t then
         let nb = Z.log2 t + 1 in
         Z.((minus_one lsl nb) lor t)
       else t);
    ones =
      (let t = join1 prev.ones next.ones in
       if Z.equal prev.ones t then prev.ones
       else if Z.lt t Z.zero then
         let nb = Z.log2 (Z.neg t) + 1 in
         Z.(lognot (minus_one lsl nb) lor t)
       else t);
  }

let unknown x = Z.( lxor ) x.zeros x.ones

let is_singleton x = if Z.equal x.zeros x.ones then Some x.ones else None

open Ast.Expr

let is_power_of_two x = Z.equal Z.zero (Z.( land ) x (Z.pred x))
(* alternative: Z.popcount x == 1 *)

(** WARNING: not is_negative means "don't know", not "is positive" *)
let is_negative x = Z.lt x.ones Z.zero && Z.lt x.zeros Z.zero

(** WARNING: not is_positive means "don't know", not "is negative" *)
let is_positive x = Z.leq Z.zero x.zeros && Z.leq Z.zero x.ones

let is_zero t =
  match is_singleton t with
  | Some x -> Z.equal x Z.zero
  | None -> false

let of_ternary = function
  | Ternary.True -> one
  | Ternary.False -> zero
  | Ternary.TrueFalse -> zero_one

let to_ternary x =
  match is_singleton x with
  | Some z when Z.equal z Z.zero -> Ternary.False
  | _ -> if contains x Z.zero then Ternary.TrueFalse else Ternary.True

module Forwards = struct
  let const = singleton

  (* And: keep 0 when one is known, 1 when both are known. *)
  let band x y = { zeros = Z.( land ) x.zeros y.zeros; ones = Z.( land ) x.ones y.ones }

  (* Or: keep 1 when one is known, 0 when both are known. *)
  let bor x y = { zeros = Z.( lor ) x.zeros y.zeros; ones = Z.( lor ) x.ones y.ones }

  (* 6 operations. *)
  let bxor x y =
    let open Z in
    let xorx0y0 = x.zeros lxor y.zeros in
    let xorx1y1 = x.ones lxor y.ones in
    let unknown = unknown x in
    (* xorx0y0 lor xorx1y1 works for all but the double unknown case.
       So we add an "unkown" check by loring with (x0 xor x1, i.e. when xbit is unknown) *)
    { zeros = xorx0y0 lor xorx1y1 lor unknown; ones = xorx0y0 land xorx1y1 }

  (* Fast and relatively precise, without requiring a
     large number of iterations. *)
  let add x y =
    (* Ideal case: no carry at all, if for all indices at least one
       component is known to be 0. *)
    let carry = Z.( land ) x.zeros y.zeros in
    if Z.equal Z.zero carry then bxor x y
    else
      (* Fallback case: when initially there are two zeroes, the
         carries will not propagate across this position. So we know
         the value of the bit above. This is useful when both values
         are known to hold a lot of zeroes. *)
      let never_carry = Z.( lor ) x.zeros y.zeros in
      let res = bxor x y in
      let shifted_carry = Z.shift_left never_carry 1 in
      let nshifted_carry = Z.( ~! ) shifted_carry in
      { zeros = Z.( lor ) res.zeros shifted_carry; ones = Z.( land ) res.ones nshifted_carry }

  (* More precise. *)
  (* let _biadd ~size (x0, x1) (y0, y1) =
     (* let nocarry = Z.extract (Z.shift_left (Z.(land) x0 y0) 1) 0 size in *)
     (* Nocarry avoids infinitely propagating carries.
        I think it ensures termination. *)
     let nocarry = Z.( land ) x0 y0 in
     let rec loop a b =
       let half_sum = bxor ~size a b in
       let carry0, carry1 = band ~size a b in
       let carry0 = Z.( land ) carry0 nocarry in
       (* If a carry is possible. *)
       if Z.equal carry0 Z.zero then half_sum
       else
         let carry0 = Z.shift_left carry0 1 in
         let carry1 = Z.( land ) carry1 nocarry in
         let carry1 = Z.shift_left carry1 1 in
         loop half_sum (carry0, carry1)
     in
     let res0, res1 = loop (x0, x1) (y0, y1) in
     (Z.extract res0 0 size, Z.extract res1 0 size) *)

  (* This algorithm is symmetric to biadd's. *)
  let sub x y =
    (* Borrow happens when x may be 0 and y may be 1. *)
    let res = bxor x y in
    let no_borrow = Z.( land ) (Z.( ~! ) x.ones) y.zeros in
    if Z.equal Z.zero no_borrow then res
    else
      (* Fallback: no borrow when x_i is known to be 1 and y_i known to be 0.
         In this case, we use the half_sum computed by xor.
         Otherwise, we set the bit to unknown. *)
      let never_borrow = Z.( land ) x.ones (Z.( ~! ) y.zeros) in
      let shifted_borrow = Z.shift_left never_borrow 1 in
      let nshifted_borrow = Z.( ~! ) shifted_borrow in
      { zeros = Z.( lor ) res.zeros nshifted_borrow; ones = Z.( land ) res.ones shifted_borrow }

  let neg x = sub zero x

  let mul_singleton x k =
    if is_power_of_two k then { zeros = Z.( * ) x.zeros k; ones = Z.( * ) x.ones k } else top

  let mul x y =
    match (is_singleton x, is_singleton y) with
    | Some x, Some y -> singleton Z.(x * y)
    | Some x, None -> mul_singleton y x
    | None, Some y -> mul_singleton x y
    | None, None -> top

  (* If y is a multiple of 2^k, the last k bits of x and result are
     the same.
     Also, if y is smaller than some value, then the result must be
     smaller than this value. *)
  let unsigned_mod x y =
    if Z.equal Z.zero y.zeros then raise Division_by_zero
    else
      let k = Z.trailing_zeros y.zeros in
      (* Put to top all the bits above k. *)
      let mask0 = Z.sub (Z.shift_left Z.one k) Z.one in
      let mask1 = Z.( ~! ) mask0 in
      let k2 = Z.log2up y.zeros in
      (* Maximum value for y0. *)
      let nask = Z.sub (Z.shift_left Z.one k2) Z.one in
      {
        zeros = Z.( land ) nask @@ Z.( lor ) x.zeros mask1;
        ones = Z.( land ) nask @@ Z.( land ) x.ones mask0;
      }

  let signed_mod_alt x y =
    match is_singleton y with
    | None -> top
    | Some y ->
        if Z.equal Z.zero y then raise Division_by_zero
        else if is_power_of_two y then { zeros = Z.rem x.zeros y; ones = Z.rem x.ones y }
        else top

  let signed_mod x y =
    if is_positive x && is_positive y then unsigned_mod x y
      (* else if Z.equal Z.zero y.zeros then None: can't happen unless y is positive *)
    else
      let k = Z.trailing_zeros y.zeros in
      (* Put to top all the bits above k. *)
      let mask0 = Z.sub (Z.shift_left Z.one k) Z.one in
      let mask1 = Z.( ~! ) mask0 in
      { zeros = Z.( lor ) x.zeros mask1; ones = Z.( land ) x.ones mask0 }

  let signed_mod x y =
    let m = signed_mod x y in
    match meet m (signed_mod_alt x y) with
    | Some x -> x
    | None -> raise Utils.Bottom

  (** Assumes [y <> 0], returns None if no info [top] *)
  let div_by_singleton x y =
    if is_power_of_two y then
      if is_negative x then None
        (* Signed division of negative numbers is different from a
           shift. *)
      else Some { zeros = Z.div x.zeros y; ones = Z.div x.ones y }
    else None

  let div x y =
    match is_singleton y with
    | None -> top
    | Some y -> (
        if Z.equal Z.zero y then raise Division_by_zero
        else
          match div_by_singleton x y with
          | None -> top
          | Some x -> x)

  let shift_left_singleton x k =
    if k > 1024 then top else { zeros = Z.shift_left x.zeros k; ones = Z.shift_left x.ones k }
  let shift_right_singleton x k =
    if k > 1024 then top else { zeros = Z.shift_right x.zeros k; ones = Z.shift_right x.ones k }

  let eq x y =
    (* Bits for which both x0 and x1 are different. Outside of
       (bottom,top) that we do not consider, should happen only if
       one bit is known to be 0 and the other known to be 1. *)
    if not @@ Z.equal Z.zero Z.(x.zeros lxor y.zeros land (x.ones lxor y.ones)) then Ternary.False
    else if Z.equal x.zeros x.ones && Z.equal x.zeros y.zeros && Z.equal x.ones y.ones then
      Ternary.True
    else Ternary.TrueFalse

  (* must be true if no unknown bit and equal. Not the best domain for that. *)
  (* must be false if a single bit is known to be different.  *)

  let min_max x =
    if is_positive x then Some (x.zeros, x.ones)
    else if is_negative x then Some (x.ones, x.zeros)
    else None

  let leq x y =
    match min_max x with
    | None -> Ternary.TrueFalse
    | Some (minx, maxx) -> (
        match min_max y with
        | None -> Ternary.TrueFalse
        | Some (miny, maxy) ->
            if Z.leq maxx miny then Ternary.True
            else if Z.lt maxy minx then Ternary.False
            else Ternary.TrueFalse)

  (* Note that this probably duplicates what the interval will be doing. *)
  (* let comp min_max_extract ~size x y =
       let minx, maxx = min_max_extract ~size x in
       let miny, maxy = min_max_extract ~size y in
       if Z.leq maxx miny then one else if Z.lt maxy minx then zero else zero_one

       let min_max_signed x =
         if is_positive x then x.zeros, x.ones else
         if is_negative x then x.ones, x.zeros else
         match testbit (x0,x1) (size - 1) with
         | Zero -> (x1,x0)
         | One -> (Z.signed_extract x1 0 size, Z.signed_extract x0 0 size)                    (* If the number is negative, the more one the higher, like when it is positive. *)
         | Unknown ->
            (* Flip the bit before extraction, to ensure that min is negative and max positive. *)
            let flip = (Z.shift_left Z.one @@ size - 1) in
            (Z.signed_extract (Z.(lxor) x1 flip) 0 size,
             Z.extract (Z.(lxor) x0 flip) 0 size)

     let bisle = comp min_max_signed *)

  let unop u x =
    match u with
    | Neg -> neg x
    | BNot -> { zeros = Z.( ~! ) x.zeros; ones = Z.( ~! ) x.ones }
    | LNot ->
        if contains x Z.zero then
          match is_singleton x with
          | Some _ -> one
          | None -> zero_one
        else zero

  let binop b l r =
    match b with
    | PlusA | PlusPI -> add l r
    | MinusA | MinusPI | MinusPP -> sub l r
    | Mult -> mul l r
    | Div -> div l r
    | Mod -> signed_mod l r
    | Shiftlt -> (
        match is_singleton r with
        | Some x when Z.geq x Z.zero -> shift_left_singleton l (Z.to_int x)
        | _ -> top)
    | Shiftrt -> (
        match is_singleton r with
        | Some x when Z.geq x Z.zero -> shift_right_singleton l (Z.to_int x)
        | _ -> top)
    | Lt -> of_ternary (Ternary.not (leq r l))
    | Gt -> of_ternary (Ternary.not (leq l r))
    | Le -> of_ternary (leq l r)
    | Ge -> of_ternary (leq r l)
    | Eq -> of_ternary (eq l r)
    | Ne -> of_ternary (Ternary.not (eq l r))
    | BAnd -> band l r
    | BXor -> bxor l r
    | BOr -> bor l r
    | LOr -> of_ternary Ternary.(to_ternary l || to_ternary r)
    | LAnd -> of_ternary Ternary.(to_ternary l && to_ternary r)

  let ternary cond if_true if_false =
    match to_ternary cond with
    | False -> if_false
    | True -> if_true
    | TrueFalse -> join if_true if_false
end

let no_info = (None, None)

module Backwards = struct
  let unop u _ ~res =
    match u with
    | Neg -> Some (Forwards.neg res)
    | BNot -> Some { zeros = Z.( ~! ) res.zeros; ones = Z.( ~! ) res.ones }
    | LNot ->
        if contains res Z.zero then
          match is_singleton res with
          | Some _ -> Some non_zero
          | None -> None (* top *)
        else Some zero

  let eq x y res =
    match res with
    | Ternary.TrueFalse -> no_info
    | Ternary.True -> (Some y, Some x)
    | Ternary.False ->
        (* We can learn only one bit, if it is the only one unknown
           for x, that everything is known in y, and all the other
           bits are the same in x and y.
           Probably never triggers. *)
        let unknownx = unknown x in
        let unknowny = unknown y in
        (* Learn 1 bit of y from x. *)
        let learn x y =
          if Z.equal x.zeros y.zeros then { y with zeros = y.ones } else { y with ones = y.zeros }
        in
        if Z.equal unknownx Z.zero && is_power_of_two unknowny then (None, Some (learn x y))
        else if Z.equal unknowny Z.zero && is_power_of_two unknownx then (Some (learn y x), None)
        else no_info

  (** Since our numbers are arbitrary-precision, shifting left doesn't forget bits
      When going backwards, we just shift right again *)
  let shift_right_introduce_top x k = Forwards.shift_right_singleton x k

  let shift_left_introduce_top x k =
    if k == 0 then x
    else
      let x = Forwards.shift_left_singleton x k in
      let top_0 = Z.extract top.zeros 0 k in
      let top_1 = Z.extract top.ones 0 k in
      { zeros = Z.( lor ) x.zeros top_0; ones = Z.( lor ) x.ones top_1 }

  (* Only handle the cases where the division is a shift. *)
  let unsigned_div _ y r =
    match is_singleton y with
    | Some y when (not @@ Z.equal Z.zero y) && is_power_of_two y ->
        let k = Z.log2 y in
        (Some (shift_left_introduce_top r k), None)
    | _ -> no_info

  (* if r = x mod (2^k * z)
     then the k lastbits of r are the k last bits of x.
     TODO: This should be done on the forward operation too. *)
  let modulo _ y res =
    (* Find the number of trailing zeroes in y. *)
    if Z.equal Z.zero y.zeros (* y must be zero. *) then no_info
    else
      let k = Z.trailing_zeros y.zeros (* Number of known zeros. *) in
      (* Put to top all the bits above k. *)
      let mask0 = Z.sub (Z.shift_left Z.one k) Z.one in
      let mask1 = Z.( ~! ) mask0 in
      (Some { zeros = Z.( lor ) res.zeros mask1; ones = Z.( land ) res.ones mask0 }, None)

  (* Assume that (x0,x1) <= maxu. *)
  let assume_unsigned_max x maxu =
    (* Format.printf "assume unsigned max %s %a@." (Z.to_string maxu) pretty (x0,x1); *)
    (* Naive algorithm: iterate from the most significant bit.
       If max is 0: set to 0 and continue.
       If max is 1: if x is known as 1, continue; if unknown, or known 0, stop
       (we don't known about the next bits).

       Clever algorithm : find the limit of all the bits that will be copied
       from maxu, then do the copy. *)
    try
      let first_uncopied_bit = Z.log2 @@ Z.( land ) maxu @@ Z.( ~! ) @@ Z.( land ) x.zeros x.ones in
      let copy_bits_up_to = first_uncopied_bit + 1 in
      (* Format.printf "copy_bits_up_to %d@." copy_bits_up_to; *)
      let mask = Z.sub (Z.shift_left Z.one copy_bits_up_to) Z.one in
      let nmask = Z.( ~! ) mask in
      let maxu_copied_bits = Z.( land ) nmask maxu in
      {
        zeros = Z.( lor ) maxu_copied_bits (Z.( land ) mask x.zeros);
        ones = Z.( lor ) maxu_copied_bits (Z.( land ) mask x.ones);
      }
    with Invalid_argument _ -> if Z.equal Z.zero maxu then zero else x
  (* Do something here? *)

  (* Assume that min u <= (x0,x1). *)
  let assume_unsigned_min x minu =
    (* Naive algorithm: iterate from the most significant bit.
       If min is 1: set to 1 and continue.
       If min is 0: if x is known as 0, continue; if unknown, or known 1, stop
       (we don't known about the next bits).

       Clever algorithm : find the limit of all the bits that will be copied
       from maxu, then do the copy. *)
    let first_uncopied_bit =
      try Z.log2 @@ Z.( land ) (Z.( ~! ) minu) @@ Z.( lor ) x.zeros x.ones
      with Invalid_argument _ -> -1
    in
    let copy_bits_up_to = first_uncopied_bit + 1 in
    (* Format.printf "copy_bits_up_to %d@." copy_bits_up_to; *)
    let mask = Z.sub (Z.shift_left Z.one copy_bits_up_to) Z.one in
    let nmask = Z.( ~! ) mask in
    let maxu_copied_bits = Z.( land ) nmask minu in
    {
      zeros = Z.( lor ) maxu_copied_bits (Z.( land ) mask x.zeros);
      ones = Z.( lor ) maxu_copied_bits (Z.( land ) mask x.ones);
    }

  let unsigned_le x y = function
    | true -> (
        match Forwards.min_max y with
        | None -> no_info
        | Some (_, maxy) -> (
            match Forwards.min_max x with
            | None -> no_info
            | Some (minx, _) ->
                let newx = assume_unsigned_max x maxy in
                let newy = assume_unsigned_min y minx in
                (Some newx, Some newy)))
    | false -> (
        (* ~(x <= y) <=> x > y <=> x >= y + 1  *)
        match Forwards.min_max y with
        | None -> no_info
        | Some (miny, _) -> (
            match Forwards.min_max x with
            | None -> no_info
            | Some (_, maxx) ->
                let newx = assume_unsigned_min x (Z.add miny Z.one) in
                let newy = assume_unsigned_max y (Z.sub maxx Z.one) in
                (Some newx, Some newy)))

  let le x y = function
    | Ternary.TrueFalse -> no_info
    | Ternary.True ->
        if is_positive x && is_positive y then unsigned_le x y true
        else if is_negative x && is_negative y then
          match unsigned_le (Forwards.neg y) (Forwards.neg x) true with
          (* x <= y <=> -y <= -x *)
          | Some l, Some r -> (Some (Forwards.neg r), Some (Forwards.neg l))
          | _ -> no_info
        else no_info
    | Ternary.False ->
        if is_positive x && is_positive y then unsigned_le x y false
        else if is_negative x && is_negative y then
          match unsigned_le (Forwards.neg y) (Forwards.neg x) false with
          | Some l, Some r -> (Some (Forwards.neg r), Some (Forwards.neg l))
          | _ -> no_info
        else no_info

  let binop b l r ~res =
    match b with
    | PlusA | PlusPI ->
        (* l + r = res <=> l = res - r <=> r = res - l  *)
        (Some (Forwards.sub res r), Some (Forwards.sub res l))
    | MinusA | MinusPI | MinusPP ->
        (* l - r = res <=> l = res + r <=> r = l - res *)
        (Some (Forwards.add res r), Some (Forwards.sub l res))
    | Mult ->
        if is_zero l || is_zero r then no_info
        else
          let res_zero = is_zero res in
          if res_zero then
            if not (contains l Z.zero) then (None, Some zero) (* l != 0 and l*r = 0 => r = 0 *)
            else if not (contains r Z.zero) then (Some zero, None)
            else no_info
          else
            ( Option.bind (is_singleton l) (Forwards.div_by_singleton res),
              Option.bind (is_singleton r) (Forwards.div_by_singleton res) )
    | Div -> if is_positive l && is_positive r then unsigned_div l r res else no_info
    | Mod -> modulo l r res
    | Shiftlt -> (
        match is_singleton r with
        | Some x when Z.geq x Z.zero -> (Some (shift_right_introduce_top res (Z.to_int x)), None)
        | _ -> no_info)
    | Shiftrt -> (
        match is_singleton r with
        | Some x when Z.geq x Z.zero -> (Some (shift_left_introduce_top res (Z.to_int x)), None)
        | _ -> no_info)
    | Lt -> Utils.swap (le r l (Ternary.not (to_ternary res)))
    | Gt -> le l r (Ternary.not (to_ternary res)) (* l > r <=> not l <= r *)
    | Le -> le l r (to_ternary res)
    | Ge -> Utils.swap (le r l (to_ternary res))
    | Eq -> eq l r (to_ternary res)
    | Ne -> eq l r (Ternary.not (to_ternary res))
    | BAnd ->
        let open Z in
        (* x is 1 where res is 1 and x is 0 when res is 0 and y is 1 *)
        ( Some { zeros = res.zeros lor ~!(r.ones); ones = res.ones },
          Some { zeros = res.zeros lor ~!(l.ones); ones = res.ones } )
    | BXor -> (Some (Forwards.bxor l res), Some (Forwards.bxor r res))
    | BOr ->
        let open Z in
        (* x is 0 where res is 0 and x is 1 when res is 1 and y is 0 *)
        ( Some { zeros = res.zeros; ones = res.ones land ~!(r.zeros) },
          Some { zeros = res.zeros; ones = res.ones land ~!(l.zeros) } )
    | LAnd -> (
        match to_ternary res with
        | False ->
            if not (contains l Z.zero) then (None, Some zero)
            else if not (contains r Z.zero) then (Some zero, None)
            else no_info (* we can't really represent x <> 0 with this domain *)
        | True | TrueFalse -> no_info)
    | LOr -> (
        match to_ternary res with
        | False -> (Some zero, Some zero)
        | True | TrueFalse -> no_info)

  let ternary cond if_true if_false ~res =
    match (meet res if_true, meet res if_false) with
    | None, None -> raise Utils.Bottom
    | (Some _ as x), None -> (None (* non-zero is top, so None *), x, None)
    | None, (Some _ as x) -> (Some zero, None, x)
    | (Some _ as a), (Some _ as b) -> (
        match to_ternary cond with
        | False -> (None, None, b)
        | True -> (None, a, None)
        | _ -> (None, None, None))
end
