(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Sig
open Std_extra

module Make
    (L : SingleValueAbstraction)
    (R : SingleValueAbstraction)
    (P : sig
      val reduce : L.t * R.t -> L.t * R.t
    end) =
struct
  include Instances.InstanciateBoth (Instances.Pair) (L) (R)

  let pp fmt (l, r) =
    Format.fprintf fmt "@[%a %s %a@]" L.pp l (Utils.Unicode.intersection ()) R.pp r

  let id = L.id ^ R.id
  let state_name = L.state_name ^ " * " ^ R.state_name
  let name fmt () = Format.fprintf fmt "%s(@[%a@])(@[%a@]}" "SVA.Product" L.name () R.name ()
  let doc = state_name

  let contains (l, r) z = L.contains l z && R.contains r z

  let top = (L.top, R.top)
  let one = (L.one, R.one)
  let zero = (L.zero, R.zero)
  let non_zero = (L.non_zero, R.non_zero)

  let join (l1, r1) (l2, r2) = P.reduce (L.join l1 l2, R.join r1 r2)

  let meet (l1, r1) (l2, r2) =
    match L.meet l1 l2 with
    | None -> None
    | Some l -> (
        match R.meet r1 r2 with
        | None -> None
        | Some r -> Some (P.reduce (l, r)))

  let widen (l1, r1) (l2, r2) = P.reduce (L.widen l1 l2, R.widen r1 r2)

  let subseteq (l1, r1) (l2, r2) = L.subseteq l1 l2 || R.subseteq r1 r2

  let is_singleton (l, r) =
    match L.is_singleton l with
    | None -> R.is_singleton r
    | x -> x

  let of_ternary t = P.reduce (L.of_ternary t, R.of_ternary t)
  let to_ternary (l, r) =
    match L.to_ternary l with
    | TrueFalse -> R.to_ternary r
    | x -> x

  module Forwards = struct
    let const z = P.reduce (L.Forwards.const z, R.Forwards.const z)
    let unop u (l, r) = P.reduce (L.Forwards.unop u l, R.Forwards.unop u r)
    let binop b (l1, r1) (l2, r2) = P.reduce (L.Forwards.binop b l1 l2, R.Forwards.binop b r1 r2)
    let ternary (l1, r1) (l2, r2) (l3, r3) =
      P.reduce (L.Forwards.ternary l1 l2 l3, R.Forwards.ternary r1 r2 r3)
  end

  module Backwards = struct
    let combine l r =
      match (l, r) with
      | None, None -> None
      | Some l, Some r -> Some (P.reduce (l, r))
      | None, Some r -> Some (P.reduce (L.top, r))
      | Some l, None -> Some (P.reduce (l, R.top))

    (* let const : Z.t -> res:t -> t option *)
    (* We can skip const, we have that information in [contains res z] *)
    let unop u (l, r) ~res:(res_l, res_r) =
      combine (L.Backwards.unop u l ~res:res_l) (R.Backwards.unop u r ~res:res_r)

    let binop b (l1, r1) (l2, r2) ~res:(res_l, res_r) =
      let l1, l2 = L.Backwards.binop b l1 l2 ~res:res_l in
      let r1, r2 = R.Backwards.binop b r1 r2 ~res:res_r in
      (combine l1 r1, combine l2 r2)

    let ternary (l1, r1) (l2, r2) (l3, r3) ~res:(res_l, res_r) =
      let l1, l2, l3 = L.Backwards.ternary l1 l2 l3 ~res:res_l in
      let r1, r2, r3 = R.Backwards.ternary r1 r2 r3 ~res:res_r in
      (combine l1 r1, combine l2 r2, combine l3 r3)
  end
end

let reduce_IS (i, s) =
  match s with
  | Sentinel.Zero -> (Interval.zero, Sentinel.Zero)
  | Sentinel.NonZero -> (
      match Interval.remove_const i Z.zero with
      | Some i' -> (i', s)
      | None -> (i, s))
  | Sentinel.Top -> (i, Sentinel.of_ternary (Interval.to_ternary i))

let reduce_KS (k, s) =
  match s with
  | Sentinel.Zero -> (Known_bits.zero, Sentinel.Zero)
  | Sentinel.NonZero | Sentinel.Top -> (k, Sentinel.of_ternary (Known_bits.to_ternary k))

module Interval_Sentinel =
  Make (Interval) (Sentinel)
    (struct
      let reduce = reduce_IS
    end)

module KnownBits_Sentinel =
  Make (Known_bits) (Sentinel)
    (struct
      let reduce = reduce_KS
    end)

module Interval_KnownBits_Sentinel =
  Make (Interval) (KnownBits_Sentinel)
    (struct
      let reduce (i, (k, s)) =
        let i, s = reduce_IS (i, s) in
        (i, reduce_KS (k, s))
    end)
