(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(* A significant portion of this file is taken from the Codex plugin of Frama-C *)
(* Note: this was not heavily tested or reviewed. *)
open Std_extra

let id = "S"
let state_name = "Sentinel"
let name fmt () = Format.fprintf fmt "Sentinel"
let doc = "Sentinel domain: zero, non-zero, or top"

type t =
  | Zero
  | NonZero
  | Top
[@@deriving eq, ord]

let order a b = Utils.order_of_compare (compare a b)

let pp fmt x =
  Format.fprintf fmt "@{<cyan>@[%s@]@}"
    (match x with
    | Zero -> "{0}"
    | NonZero -> Utils.Unicode.z () ^ "\\{0}"
    | Top -> Utils.Unicode.z ())
let hash = function
  | Zero -> 2
  | NonZero -> 3
  | Top -> 4

let contains x i =
  match x with
  | Top -> true
  | Zero -> Z.(equal i zero)
  | NonZero -> not Z.(equal i zero)

let top = Top
let one = NonZero
let zero = Zero
let non_zero = NonZero

let of_ternary = function
  | Ternary.False -> Zero
  | Ternary.True -> NonZero
  | Ternary.TrueFalse -> Top

let to_ternary = function
  | Zero -> Ternary.False
  | NonZero -> Ternary.True
  | Top -> Ternary.TrueFalse

let join x y =
  match (x, y) with
  | Zero, Zero -> Zero
  | NonZero, NonZero -> NonZero
  | _ -> Top

let meet x y =
  match (x, y) with
  | Top, a | a, Top -> Some a
  | NonZero, NonZero -> Some NonZero
  | Zero, Zero -> Some Zero
  | _ -> None

let widen = join

let subseteq x y =
  match (x, y) with
  | Zero, NonZero | NonZero, Zero -> false
  | Zero, Zero | NonZero, NonZero -> true
  | _, Top -> true
  | Top, _ -> false

let is_singleton = function
  | Zero -> Some Z.zero
  | _ -> None

let lnot = function
  | Top -> Top
  | Zero -> NonZero
  | NonZero -> Zero

open Ast.Expr

module Forwards = struct
  let const i = if Z.equal i Z.zero then Zero else NonZero
  let unop u i =
    match u with
    | Neg -> i
    | LNot -> lnot i
    | BNot -> (
        match i with
        | Top -> Top
        | Zero -> NonZero
        | NonZero -> Top)

  let binop b il ir =
    match b with
    | PlusA | PlusPI | MinusA | MinusPI | MinusPP -> (
        match (il, ir) with
        | Zero, Zero -> Zero
        | NonZero, Zero | Zero, NonZero -> NonZero
        | _ -> Top)
    | Mult -> (
        match (il, ir) with
        | Zero, _ | _, Zero -> Zero
        | NonZero, NonZero -> NonZero
        | _ -> Top)
    | Div -> (
        match (il, ir) with
        | _, Zero -> raise Division_by_zero
        | Zero, _ -> Zero
        | NonZero, _ -> Top
        | Top, (NonZero | Top) -> Top)
    | Mod -> (
        match (il, ir) with
        | _, Zero -> raise Division_by_zero
        | Zero, _ -> Zero
        | _ -> Top)
    | Shiftlt | Shiftrt -> (
        match (il, ir) with
        | Zero, _ -> Zero
        | NonZero, Zero -> NonZero
        | _ -> Top)
    | Lt | Gt | Le | Ge -> Top
    | Eq -> (
        match (il, ir) with
        | Zero, Zero -> NonZero
        | Zero, NonZero | NonZero, Zero -> Zero
        | _ -> Top)
    | Ne -> (
        match (il, ir) with
        | Zero, Zero -> Zero
        | Zero, NonZero | NonZero, Zero -> NonZero
        | _ -> Top)
    | BAnd -> (
        match (il, ir) with
        | Zero, _ | _, Zero -> Zero
        | _ -> Top)
    | BXor -> (
        match (il, ir) with
        | Zero, Zero -> Zero
        | Zero, NonZero | NonZero, Zero -> NonZero
        | _ -> Top)
    | LOr | BOr -> (
        match (il, ir) with
        | Zero, Zero -> Zero
        | NonZero, _ | _, NonZero -> NonZero
        | _ -> Top)
    | LAnd -> (
        match (il, ir) with
        | NonZero, NonZero -> NonZero
        | Zero, _ | _, Zero -> Zero
        | _ -> Top)

  let ternary cond if_true if_false =
    match cond with
    | Zero -> if_false
    | NonZero -> if_true
    | Top -> join if_true if_false
end

module Backwards = struct
  let unop u _ ~res =
    (* With this particular domain, unary operators are all involutive *)
    Some (Forwards.unop u res)

  let no_info = (None, None)

  let add_sub x y = function
    | Zero -> (
        match (x, y) with
        | Zero, Top -> (None, Some Zero)
        | Top, Zero -> (Some Zero, None)
        | NonZero, Top -> (None, Some NonZero)
        | Top, NonZero -> (Some NonZero, None)
        | _ -> no_info)
    | NonZero -> (
        match (x, y) with
        | Zero, Top -> (None, Some NonZero)
        | Top, Zero -> (Some NonZero, None)
        | _ -> no_info)
    | _ -> no_info

  let eq x y = function
    | NonZero -> (
        match (x, y) with
        | _, Zero -> (Some Zero, None)
        | Zero, _ -> (None, Some Zero)
        | _, NonZero -> (Some NonZero, None)
        | NonZero, _ -> (None, Some NonZero)
        | Top, Top -> no_info)
    | Zero -> (
        match (x, y) with
        | _, Zero -> (Some NonZero, None)
        | Zero, _ -> (None, Some NonZero)
        | _ -> no_info)
    | _ -> no_info

  let le x y = function
    | Zero -> (
        match (x, y) with
        | _, Zero -> (Some NonZero, None) (* not (x <= 0) => x != 0 *)
        | Zero, _ -> (None, Some NonZero)
        | _ -> no_info)
    | _ -> no_info

  (** Right or left shift *)
  let shift x y = function
    | Zero -> (
        match (x, y) with
        | _, Zero -> (Some Zero, None) (* x << 0 = 0 => x = 0 *)
        | NonZero, _ -> (None, Some NonZero) (* x << y = 0 and x != 0 => y != 0 *)
        | _ -> no_info)
    | NonZero -> (Some NonZero, None)
    | _ -> no_info

  let swap (r, l) = (l, r)

  let binop b il ir ~res =
    match b with
    | PlusA | PlusPI | MinusA | MinusPI | MinusPP -> add_sub il ir res
    | Mult -> (
        match res with
        | Zero -> (
            match (il, ir) with
            | NonZero, _ -> (None, Some Zero)
            | _, NonZero -> (Some Zero, None)
            | _ -> no_info)
        | NonZero -> (Some NonZero, Some NonZero)
        | _ -> no_info)
    | Div -> (
        match res with
        | Zero -> (Some Zero, Some NonZero)
        | NonZero -> (Some NonZero, Some NonZero)
        | _ -> (None, Some NonZero))
    | Mod -> (
        match res with
        | NonZero -> (Some NonZero, Some NonZero)
        | _ -> (None, Some NonZero))
    | Shiftlt | Shiftrt -> shift il ir res
    | Lt -> swap (le ir il (lnot res)) (* l < r <=> not (r <= l) *)
    | Gt -> le il ir (lnot res) (* l > r <=> not (l <= r) *)
    | Le -> le il ir res
    | Ge -> swap (le ir il res)
    | Eq -> eq il ir res
    | Ne -> eq il ir (lnot res)
    | BAnd -> if res = NonZero then (Some NonZero, Some NonZero) else no_info
    | LOr | BOr -> (
        match res with
        | Zero -> (Some Zero, Some Zero)
        | NonZero -> (
            match (il, ir) with
            | _, Zero -> (Some NonZero, None)
            | Zero, _ -> (None, Some NonZero)
            | _ -> no_info)
        | _ -> no_info)
    | BXor -> (
        match res with
        | Zero -> (
            match (il, ir) with
            | _, Zero -> (Some Zero, None) (* a ^ 0 = 0 => a = 0 *)
            | Zero, _ -> (None, Some Zero)
            | _, NonZero -> (Some NonZero, None) (* a ^ x = 0 => a = x *)
            | NonZero, _ -> (None, Some NonZero)
            | _ -> no_info)
        | NonZero -> (
            match (il, ir) with
            | _, Zero -> (Some NonZero, None)
            | Zero, _ -> (None, Some NonZero)
            | _ -> no_info)
        | _ -> no_info)
    | LAnd -> (
        match res with
        | NonZero -> (Some NonZero, Some NonZero)
        | Zero -> (
            match (il, ir) with
            | NonZero, _ -> (None, Some Zero)
            | _, NonZero -> (Some Zero, None)
            | _ -> no_info)
        | _ -> no_info)

  let ternary cond if_true if_false ~res =
    match (meet res if_true, meet res if_false) with
    | None, None -> raise Utils.Bottom
    | (Some _ as x), None -> (Some non_zero, x, None)
    | None, (Some _ as x) -> (Some zero, None, x)
    | (Some _ as a), (Some _ as b) -> (
        match cond with
        | Zero -> (None, None, b)
        | NonZero -> (None, a, None)
        | _ -> (None, None, None))
end
