(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Common signature of all single value abstractions *)

open Std_extra.Types
open Ast.Expr

(** Documentation information used for pretty-printing *)
module type Doc = sig
  val id : string
  (** Unique identifier for the domain, eg "N", "FA",... *)

  val name : Format.formatter -> unit -> unit
  (** Full human readable name *)

  val state_name : string
  (** Name of the domain's states *)

  val doc : string
  (** Long description of the domain *)
end

module type SingleValueAbstraction = sig
  (** Domain that abstract a single value [Z.t] by [t], meaning [t]
      represents a (non necessarily finite) set of [Z.t],
      defined by the {!contains} function

      Also packs {!Forwards} and {!Backwards} transfer functions.contents

      Generally speaking, their is no value for bottom in our single value
      abstractions. Functions that could return bottom either return an option
      type, or raise {!Std_extra.Utils.Bottom}. *)

  include Doc

  include TYPE
  (** Type of abstract numbers, this represents a non-empty set of [Z.t] *)

  val contains : t -> Z.t -> bool
  (** [contains zs z] is [true] if and only if the concretization of [zs] contains [z].
      This essentially defines the concretization [t -> Z.t set]
      through its indicative function *)

  val top : t
  (** The top element, representing any number *)

  val one : t
  (** The best abstraction for the singleton \{1\} *)

  val zero : t
  (** The best abstraction for the singleton \{0\} *)

  val non_zero : t
  (** The best abstraction for Z \ \{0\} *)

  val join : t -> t -> t
  (** [join a b] is the union of both sets *)

  val meet : t -> t -> t option
  (** [meet a b] is the intersection of both sets, [None] when empty *)

  val widen : t -> t -> t
  (** [widen] is a union that should converge in finite time *)

  val subseteq : t -> t -> bool
  (** [subseteq l r] is true if and only if any number abstracted by [l] is also
      abstracted by [r], e.g. if forall x, [contains l x] implies [contains r x] *)

  val is_singleton : t -> Z.t option
  (** [is_singleton zs] is [Some z] if [zs] represents the singleton \{z\} *)

  val of_ternary : Ternary.t -> t
  (** [of_ternary t] turns [t] seen as \{0\}, \{1\}, or \{0,1\} into its best abstraction *)

  val to_ternary : t -> Ternary.t
  (** [to_ternary t] is used to convert to a boolean value for test.
      Any non-zero value is {!Ternary.True}, and {!zero} is {!Ternary.False}.
      A value containing both zero and non-zero values is {!Ternary.TrueFalse} *)

  (** Forward transfer functions for all expression constructs. Used to calculate
      the abstraction of an expression based on abstractions of its components

      For a given construct [C(z1, .., zn)], [Forwards.C z1' .. zn'] should verify:
      if for all i [contains zi' zi], then [contains (Forwards.C z1' .. zn') (C(z1,..,zn))] *)
  module Forwards : sig
    val const : Z.t -> t
    val unop : unop -> t -> t

    val binop : binop -> t -> t -> t
    (** @raises Division_by_zero *)

    val ternary : t -> t -> t -> t
  end

  (** Backward transfer functions for all expression constructs. Used to refine
      abstractions of a constructs arguments given a precise abstraction of its
      results.

      For a given construct [C(z1,..,zn)], [Backards.C z1' .. zn' ~res] should verify:
      if [contains res (C(z1,..,zn))] then for all i [contains (Backards.C z1' .. zn' ~res).i zi]

      These return [None] if no new information can be deduced for this argument.

      These do not perform the {!meet} with the arguments, they only return the new information.
      So if [Backwards.binop PlusA l r] is [Some l', None], the best abstraction
      for each argument is [meet l l'] and [r] respectively.
      If the {!meet} is [None], then this expression cannot possibly have the given value [~res].

      Also note that functions may return [Some x'] even if [x'] isn't more precise than
      the argument [x]. Sometimes deductions can be made without even inspecting that argument.
      Check that [not (subseteq x (meet x' x))] holds before doing recursive calls.

      @raises Bottom if some cases are unreachable. *)
  module Backwards : sig
    (* val const : Z.t -> res:t -> t option *)
    (* We can skip const, we have that information in [contains res z] *)
    val unop : unop -> t -> res:t -> t option
    val binop : binop -> t -> t -> res:t -> t option * t option
    val ternary : t -> t -> t -> res:t -> t option * t option * t option
  end
end
