(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Types
open Utils

module Bool = struct
  type t = bool
  let compare = Bool.compare
  let order a b = order_of_compare (compare a b)
  let equal = Bool.equal
  let pp = Format.pp_print_bool
  let hash = Bool.to_int
end

module Int = struct
  type t = int
  let compare = Int.compare
  let order a b = order_of_compare (compare a b)
  let equal = Int.equal
  let pp = Format.pp_print_int
  let hash (x : t) = x
end

module Char = struct
  type t = char
  let compare = Char.compare
  let order a b = order_of_compare (compare a b)
  let equal = Char.equal
  let pp = Format.pp_print_char
  let hash = int_of_char
end

module String = struct
  type t = string
  let compare = String.compare
  let order a b = order_of_compare (compare a b)
  let equal = String.equal
  let pp = Format.pp_print_string
  let hash (s : t) = Hashtbl.hash s
end

module Order = struct
  type t = order
  let compare a b =
    match (a, b) with
    | Lt, Lt | Eq, Eq | Gt, Gt -> 0
    | Lt, _ | Eq, Gt -> compare_inf
    | _ -> compare_sup

  let order a b =
    match (a, b) with
    | Lt, Lt | Eq, Eq | Gt, Gt -> Eq
    | Lt, _ | Eq, Gt -> Lt
    | _ -> Gt

  let equal a b =
    match (a, b) with
    | Lt, Lt | Eq, Eq | Gt, Gt -> true
    | _ -> false

  let pp fmt = function
    | Lt -> Format.fprintf fmt "Lt"
    | Eq -> Format.fprintf fmt "Eq"
    | Gt -> Format.fprintf fmt "Gt"

  let hash = function
    | Lt -> 1
    | Eq -> 2
    | Gt -> 3
end

module Instanciate (T : UNARY_TYPE) (A : TYPE) = struct
  type t = A.t T.t
  let compare = T.compare A.compare
  let order = T.order A.order
  let equal = T.equal A.equal
  let pp = T.pp A.pp
  let hash = T.hash A.hash
end

module Compose (A : UNARY_TYPE) (B : UNARY_TYPE) = struct
  type 'a t = 'a A.t B.t
  let compare cmp = B.compare (A.compare cmp)
  let order order = B.order (A.order order)
  let equal equal = B.equal (A.equal equal)
  let hash hash = B.hash (A.hash hash)
  let pp pp = B.pp (A.pp pp)
end

module ComposePreMonad (A : PREMONAD) (B : PREMONAD) = struct
  include Compose (A) (B)

  let map f = B.map (A.map f)
  let iter f = B.iter (A.iter f)
  let fold f = B.fold (A.fold f)
  let return x = B.return (A.return x)
end

module UnaryOfPoly (P : POLY_TYPE) : UNARY_TYPE with type 'a t = 'a P.t = struct
  include P

  let equal _ a b = equal_of_poly_equal (poly_equal a b)
  let order _ a b = order_of_poly_order (poly_order a b)
  let compare _ a b = compare_of_order (order_of_poly_order (poly_order a b))
  let hash _ a = hash a
end

module Option = struct
  type 'a t = 'a option

  let compare = Option.compare
  let order order a b = unary_order_of_compare compare order a b
  let equal = Option.equal
  let pp pp =
    Format.pp_print_option
      ~none:(fun fmt () -> Format.fprintf fmt "None")
      (fun fmt a -> Format.fprintf fmt "Some(@[%a@])" pp a)
  let hash hash = function
    | None -> 0
    | Some x -> 1 + hash x

  let map = Option.map
  let iter = Option.iter
  let fold f opt acc = Option.fold ~none:acc ~some:(fun o -> f o acc) opt
  let bind o f = Option.bind f o
  let return = Option.some
  let join = Option.join
end

type +!'a topbot =
  | Bottom
  | Elem of 'a
  | Top

module TopBot = struct
  type 'a t = 'a topbot

  let compare compare l r =
    match (l, r) with
    | Bottom, Bottom | Top, Top -> 0
    | Bottom, _ | _, Top -> compare_inf
    | Top, _ | _, Bottom -> compare_sup
    | Elem l, Elem r -> compare l r

  let order order a b = unary_order_of_compare compare order a b

  let equal equal l r =
    match (l, r) with
    | Bottom, Bottom | Top, Top -> true
    | Elem l, Elem r -> equal l r
    | _, _ -> false

  let hash hash = function
    | Bottom -> 0
    | Top -> 1
    | Elem x -> abs (2 + hash x)

  let pp pp fmt = function
    | Bottom -> Format.fprintf fmt "Bottom"
    | Top -> Format.fprintf fmt "Top"
    | Elem x -> Format.fprintf fmt "Elem(@[%a@])" pp x

  let map f = function
    | Elem x -> Elem (f x)
    | Top -> Top
    | Bottom -> Bottom
  let iter f = function
    | Elem x -> f x
    | _ -> ()
  let fold f x acc =
    match x with
    | Elem x -> f x acc
    | _ -> acc
  let bind f = function
    | Elem x -> f x
    | Top -> Top
    | Bottom -> Bottom
  let return x = Elem x
  let join = function
    | Elem x -> x
    | Top -> Top
    | Bottom -> Bottom
end

module List = struct
  type 'a t = 'a list
  let compare = List.compare
  let order order a b = unary_order_of_compare compare order a b
  let equal = List.equal
  let pp pp fmt l = Format.fprintf fmt "@[[%a]@]" (Format.pp_print_list ~pp_sep:pp_semicolon pp) l

  let hash hash l = List.fold_left (fun h x -> hash_combine h (hash x)) 0 l

  let map = List.map
  let iter = List.iter
  let fold f l acc = List.fold_left (fun acc e -> f e acc) acc l
  let bind = List.concat_map
  let join = List.concat
  let return x = [ x ]
end

module Seq = struct
  type 'a t = 'a Seq.t
  let compare = Seq.compare
  let order order a b = unary_order_of_compare compare order a b
  let equal = Seq.equal
  let pp pp fmt l = Format.fprintf fmt "@[[%a]@]" (Format.pp_print_seq ~pp_sep:pp_semicolon pp) l

  let hash hash l = Seq.fold_left (fun h x -> hash_combine h (hash x)) 0 l

  let map = Seq.map
  let iter = Seq.iter
  let fold f l acc = Seq.fold_left (fun acc e -> f e acc) acc l
  let bind = Seq.concat_map
  let join = Seq.concat
  let return x = Seq.cons x Seq.empty
end

module Array = struct
  type 'a t = 'a array

  let equal f a b =
    let len = Array.length a in
    if len <> Array.length b then false
    else
      let exception NotEqual in
      try
        for i = 0 to len - 1 do
          if not (f a.(i) b.(i)) then raise NotEqual
        done;
        true
      with NotEqual -> false

  let compare f a b =
    let len = Array.length a in
    let cmp = Int.compare len (Array.length b) in
    if cmp <> 0 then cmp
    else
      let exception NotEqual of int in
      try
        for i = 0 to len - 1 do
          let cmp = f a.(i) b.(i) in
          if cmp <> 0 then raise (NotEqual cmp)
        done;
        0
      with NotEqual x -> x

  let order f a b =
    let len = Array.length a in
    let cmp = Int.order len (Array.length b) in
    if cmp <> Eq then cmp
    else
      let exception NotEqual of order in
      try
        for i = 0 to len - 1 do
          let cmp = f a.(i) b.(i) in
          if cmp <> Eq then raise (NotEqual cmp)
        done;
        Eq
      with NotEqual x -> x

  let hash hash x =
    let len = min (Array.length x) 5 in
    let h = ref 0 in
    for i = 0 to len - 1 do
      h := hash_combine !h (hash x.(i))
    done;
    !h

  let pp pp fmt x = List.pp pp fmt (Array.to_list x)

  let map = Array.map
  let iter = Array.iter
  let fold f l acc = Array.fold_left (fun acc e -> f e acc) acc l
  let return x = Array.make 1 x
  let join x =
    match Array.to_list x with
    | [] -> Array.make 0 (Obj.magic 0)
    | t :: q -> List.fold Array.append q t
  let bind f x = join (map f x)
end

module Ref = struct
  type 'a t = 'a ref
  let compare compare l r = compare !l !r
  let order order a b = order !a !b
  let equal equal l r = equal !l !r
  let hash hash x = hash !x
  let pp pp fmt x = pp fmt !x

  let map f x = ref (f !x)
  let iter f x = f !x
  let fold f x acc = f !x acc
  let bind f x = f !x
  let join x = !x
  let return = ref
end

module Hcons = struct
  open Hashconsing
  type 'a t = 'a hash_consed
  let compare _ l r = Int.compare l.tag r.tag
  let order _ l r = Int.order l.tag r.tag
  let equal _ l r = l == r
  let pp pp fmt x = pp fmt x.node
  let hash _ x = x.hkey
end

module InstanciateHcons (T : PRINTABLE) = struct
  open Hashconsing
  type t = T.t hash_consed
  let compare l r = Int.compare l.tag r.tag
  let order l r = Int.order l.tag r.tag
  let equal l r = l == r
  let pp fmt x = T.pp fmt x.node
  let hash x = x.hkey
  let to_int x = x.tag
end

module Pair = struct
  type ('a, 'b) t = 'a * 'b

  let compare cmp_a cmp_b (l1, l2) (r1, r2) = compare_pair cmp_a cmp_b l1 r1 l2 r2
  let order ord_a ord_b (l1, l2) (r1, r2) =
    match ord_a l1 r1 with
    | Eq -> ord_b l2 r2
    | x -> x
  let equal eq_a eq_b (l1, l2) (r1, r2) = eq_a l1 r1 && eq_b l2 r2
  let hash hash_a hash_b (a, b) = hash_combine (hash_a a) (hash_b b)
  let pp pp_a pp_b fmt (a, b) = Format.fprintf fmt "(@[%a,@,@ %a@])" pp_a a pp_b b
end

module Either = struct
  type ('a, 'b) t = ('a, 'b) Either.t
  let compare left right = Either.compare ~left ~right
  let order left right a b =
    match (a, b) with
    | Either.Left a, Either.Left b -> left a b
    | Either.Right a, Either.Right b -> right a b
    | Either.Left _, _ -> Lt
    | _ -> Gt
  let equal left right = Either.equal ~left ~right
  let hash hash_l hash_r = function
    | Either.Left l -> hash_sum ~log_total:1 ~nb:0 (hash_l l)
    | Either.Right r -> hash_sum ~log_total:1 ~nb:1 (hash_r r)

  let pp pp_l pp_r fmt = function
    | Either.Left l -> Format.fprintf fmt "Left(@[%a@])" pp_l l
    | Either.Right r -> Format.fprintf fmt "Right(@[%a@])" pp_r r
end

module Result = struct
  type ('a, 'b) t = ('a, 'b) result
  let compare ok error = Result.compare ~ok ~error
  let equal ok error = Result.equal ~ok ~error
  let order ok error a b =
    match (a, b) with
    | Ok a, Ok b -> ok a b
    | Error a, Error b -> error a b
    | Ok _, _ -> Lt
    | _ -> Gt
  let hash hash_l hash_r = function
    | Ok l -> hash_sum ~log_total:1 ~nb:0 (hash_l l)
    | Error r -> hash_sum ~log_total:1 ~nb:0 (hash_r r)

  let pp pp_l pp_r fmt = function
    | Ok l -> Format.fprintf fmt "Ok(@[%a@])" pp_l l
    | Error r -> Format.fprintf fmt "Error(@[%a@])" pp_r r
end

module InstanciateLeft (T : BINARY_TYPE) (L : TYPE) = struct
  type 'a t = (L.t, 'a) T.t
  let compare compare = T.compare L.compare compare
  let order order = T.order L.order order
  let equal equal = T.equal L.equal equal
  let hash hash = T.hash L.hash hash
  let pp pp = T.pp L.pp pp
end
module InstanciateRight (T : BINARY_TYPE) (R : TYPE) = struct
  type 'a t = ('a, R.t) T.t
  let compare compare = T.compare compare R.compare
  let order order = T.order order R.order
  let equal equal = T.equal equal R.equal
  let hash hash = T.hash hash R.hash
  let pp pp = T.pp pp R.pp
end
module InstanciateBoth (T : BINARY_TYPE) (L : TYPE) (R : TYPE) = struct
  type t = (L.t, R.t) T.t
  let compare = T.compare L.compare R.compare
  let order = T.order L.order R.order
  let equal = T.equal L.equal R.equal
  let hash = T.hash L.hash R.hash
  let pp = T.pp L.pp R.pp
end

module ComposeLeft (T : BINARY_TYPE) (L : UNARY_TYPE) = struct
  type ('a, 'b) t = ('a L.t, 'b) T.t
  let compare compare_a compare_b = T.compare (L.compare compare_a) compare_b
  let order ord_a ord_b = T.order (L.order ord_a) ord_b
  let equal equal_a equal_b = T.equal (L.equal equal_a) equal_b
  let hash hash_a hash_b = T.hash (L.hash hash_a) hash_b
  let pp pp_a pp_b = T.pp (L.pp pp_a) pp_b
end
module ComposeRight (T : BINARY_TYPE) (R : UNARY_TYPE) = struct
  type ('a, 'b) t = ('a, 'b R.t) T.t
  let compare compare_a compare_b = T.compare compare_a (R.compare compare_b)
  let order ord_a ord_b = T.order ord_a (R.order ord_b)
  let equal equal_a equal_b = T.equal equal_a (R.equal equal_b)
  let hash hash_a hash_b = T.hash hash_a (R.hash hash_b)
  let pp pp_a pp_b = T.pp pp_a (R.pp pp_b)
end

module ComposeBinary (T : UNARY_TYPE) (B : BINARY_TYPE) = struct
  type ('a, 'b) t = ('a, 'b) B.t T.t
  let compare compare_a compare_b = T.compare (B.compare compare_a compare_b)
  let order ord_a ord_b = T.order (B.order ord_a ord_b)
  let equal equal_a equal_b = T.equal (B.equal equal_a equal_b)
  let hash hash_a hash_b = T.hash (B.hash hash_a hash_b)
  let pp pp_a pp_b = T.pp (B.pp pp_a pp_b)
end

(* Liftings *)

module Lift
    (T : TYPE)
    (I : sig
      type t
      val inject : t -> T.t
    end) =
struct
  type t = I.t
  let compare l r = T.compare (I.inject l) (I.inject r)
  let order l r = T.order (I.inject l) (I.inject r)
  let equal l r = T.equal (I.inject l) (I.inject r)
  let hash x = T.hash (I.inject x)
  let pp fmt x = T.pp fmt (I.inject x)
end
module LiftUnary
    (T : UNARY_TYPE)
    (I : sig
      type 'a t
      val inject : 'a t -> 'a T.t
    end) =
struct
  type 'a t = 'a I.t
  let compare compare l r = T.compare compare (I.inject l) (I.inject r)
  let order order l r = T.order order (I.inject l) (I.inject r)
  let equal equal l r = T.equal equal (I.inject l) (I.inject r)
  let hash hash x = T.hash hash (I.inject x)
  let pp pp fmt x = T.pp pp fmt (I.inject x)
end
module LiftBinary
    (T : BINARY_TYPE)
    (I : sig
      type ('a, 'b) t
      val inject : ('a, 'b) t -> ('a, 'b) T.t
    end) =
struct
  type ('a, 'b) t = ('a, 'b) I.t
  let compare compare_a compare_b l r = T.compare compare_a compare_b (I.inject l) (I.inject r)
  let order order_a order_b l r = T.order order_a order_b (I.inject l) (I.inject r)
  let equal equal_a equal_b l r = T.equal equal_a equal_b (I.inject l) (I.inject r)
  let hash hash_a hash_b x = T.hash hash_a hash_b (I.inject x)
  let pp pp_a pp_b fmt x = T.pp pp_a pp_b fmt (I.inject x)
end
