(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Provides implementation for the common interface from {!Types} for most of
    OCaml's standard librairy types, as well as functors to combine them
    (e.g. create a [module FooBar : TYPE with type t = Foo.t Bar.t] from a
    [module Foo : TYPE] and a [module Bar ; UNARY_TYPE]) *)

open Types

(** {1 Types} *)

type +!'a topbot =
  | Bottom
  | Elem of 'a
  | Top

module Order : TYPE with type t = order
module Int : TYPE with type t = int
module Bool : TYPE with type t = bool
module Char : TYPE with type t = char
module String : TYPE with type t = string

(** {1 Unary builders} *)

module Instanciate (T : UNARY_TYPE) (A : TYPE) : TYPE with type t = A.t T.t
module Compose (A : UNARY_TYPE) (B : UNARY_TYPE) : UNARY_TYPE with type 'a t = 'a A.t B.t
module ComposePreMonad (A : PREMONAD) (B : PREMONAD) : PREMONAD with type 'a t = 'a A.t B.t
module UnaryOfPoly (P : POLY_TYPE) : UNARY_TYPE with type 'a t = 'a P.t

(** {1 Unary types} *)

module TopBot : MONAD with type 'a t = 'a topbot
module Option : MONAD with type 'a t = 'a option
module List : MONAD with type 'a t = 'a list
module Seq : MONAD with type 'a t = 'a Seq.t
module Ref : MONAD with type 'a t = 'a ref
module Array : MONAD with type 'a t = 'a array
module Hcons : UNARY_TYPE with type 'a t = 'a Hashconsing.hash_consed

(** We can instanciate {!Hcons} with fewer paramters, since [compare], [equal] and [hash]
    are built-in and don't depend on [T.compare], [T.equal] and [T.hash] *)
module InstanciateHcons (T : PRINTABLE) : TYPE_TO_INT with type t = T.t Hashconsing.hash_consed

(** {1 Binary builders} *)

module InstanciateLeft (T : BINARY_TYPE) (L : TYPE) : UNARY_TYPE with type 'a t = (L.t, 'a) T.t
module InstanciateRight (T : BINARY_TYPE) (R : TYPE) : UNARY_TYPE with type 'a t = ('a, R.t) T.t
module InstanciateBoth (T : BINARY_TYPE) (L : TYPE) (R : TYPE) : TYPE with type t = (L.t, R.t) T.t
module ComposeLeft (T : BINARY_TYPE) (L : UNARY_TYPE) :
  BINARY_TYPE with type ('a, 'b) t = ('a L.t, 'b) T.t
module ComposeRight (T : BINARY_TYPE) (R : UNARY_TYPE) :
  BINARY_TYPE with type ('a, 'b) t = ('a, 'b R.t) T.t
module ComposeBinary (T : UNARY_TYPE) (B : BINARY_TYPE) :
  BINARY_TYPE with type ('a, 'b) t = ('a, 'b) B.t T.t

(** {1 Binary types} *)

module Pair : BINARY_TYPE with type ('a, 'b) t = 'a * 'b
module Either : BINARY_TYPE with type ('a, 'b) t = ('a, 'b) Either.t
module Result : BINARY_TYPE with type ('a, 'b) t = ('a, 'b) result

(** {1 Liftings} *)

module Lift
    (T : TYPE)
    (I : sig
      type t
      val inject : t -> T.t
    end) : TYPE with type t = I.t
module LiftUnary
    (T : UNARY_TYPE)
    (I : sig
      type 'a t
      val inject : 'a t -> 'a T.t
    end) : UNARY_TYPE with type 'a t = 'a I.t
module LiftBinary
    (T : BINARY_TYPE)
    (I : sig
      type ('a, 'b) t
      val inject : ('a, 'b) t -> ('a, 'b) T.t
    end) : BINARY_TYPE with type ('a, 'b) t = ('a, 'b) I.t
