(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Common types and modules types used throughout the rest of the code, provides
    a common interface for useful functions on all types (pretty-printer, equality,
    comparison) through the {!TYPE}, {!UNARY_TYPE} and {!BINARY_TYPE} module types.

    In the rest of the code, whenever reading [module Foo : TYPE], it is best to
    interpret it as a simple type declaration. *)

type 'a pp = Format.formatter -> 'a -> unit
(** Type alias for pretty printer *)

(** Global ordering types *)
type order =
  | Lt
  | Eq
  | Gt

(** Polymorphic equality, to use with GADTs *)
type (_, _) poly_equal =
  | Diff : ('a, 'b) poly_equal
  | Equal : ('a, 'a) poly_equal

(** Polymorphic order, to use with GADTs *)
type (_, _) poly_order =
  | P_Lt : ('a, 'b) poly_order
  | P_Eq : ('a, 'a) poly_order
  | P_Gt : ('a, 'b) poly_order

(** Type with a pretty-printer *)
module type PRINTABLE = sig
  type t
  val pp : t pp
end

(** Standard useful functions for types *)
module type TYPE = sig
  include PRINTABLE

  val compare : t -> t -> int
  val order : t -> t -> order
  val equal : t -> t -> bool
  val hash : t -> int
end

(** Types that can be injected into [int] *)
module type TYPE_TO_INT = sig
  include TYPE
  val to_int : t -> int
end

(** Standard useful functions for unary types *)
module type UNARY_TYPE = sig
  type 'a t

  val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
  val order : ('a -> 'a -> order) -> 'a t -> 'a t -> order
  val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
  val pp : 'a pp -> 'a t pp
  val hash : ('a -> int) -> 'a t -> int
end

(** Standard useful functions for binary types *)
module type BINARY_TYPE = sig
  type ('a, 'b) t

  val compare : ('a -> 'a -> int) -> ('b -> 'b -> int) -> ('a, 'b) t -> ('a, 'b) t -> int
  val order : ('a -> 'a -> order) -> ('b -> 'b -> order) -> ('a, 'b) t -> ('a, 'b) t -> order
  val equal : ('a -> 'a -> bool) -> ('b -> 'b -> bool) -> ('a, 'b) t -> ('a, 'b) t -> bool
  val pp : 'a pp -> 'b pp -> ('a, 'b) t pp
  val hash : ('a -> int) -> ('b -> int) -> ('a, 'b) t -> int
end

(** {!UNARY_TYPE} with some monadic operations *)
module type PREMONAD = sig
  include UNARY_TYPE

  val map : ('a -> 'b) -> 'a t -> 'b t
  val iter : ('a -> unit) -> 'a t -> unit
  val fold : ('a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  val return : 'a -> 'a t
end

(** {!UNARY_TYPE} with all monadic operations *)
module type MONAD = sig
  include PREMONAD

  val bind : ('a -> 'b t) -> 'a t -> 'b t
  val join : 'a t t -> 'a t
end

(** Polymorphic type (GADT) *)
module type POLY_TYPE = sig
  type 'a t

  val poly_equal : 'a t -> 'b t -> ('a, 'b) poly_equal
  val poly_order : 'a t -> 'b t -> ('a, 'b) poly_order
  val pp : 'a pp -> 'a t pp
  val hash : 'a t -> int
end

(** Extends [Stdlib.Set.S] with some useful functions *)
module type SET = sig
  include Set.S

  val order : t -> t -> order

  val bind : (elt -> t) -> t -> t
  val is_singleton : t -> elt option
  val intersect : t -> t -> bool

  val pp_set : ?pp_sep:unit pp -> t pp
  (** [pp_sep] default to [Utils.pp_semicolon] (print semicolon and a cut) *)

  val pp : t pp
  (** Prints semi-colon separated set, no external delimiters *)

  val find_any : (elt -> bool) -> t -> elt
  val find_any_opt : (elt -> bool) -> t -> elt option

  val hash : t -> int
end

(** Extends [Stdlib.Map.S] with some useful functions *)
module type MAP = sig
  include Map.S

  val order : ('a -> 'a -> order) -> 'a t -> 'a t -> order

  val find_any : (key -> 'a -> bool) -> 'a t -> key * 'a
  val find_any_opt : (key -> 'a -> bool) -> 'a t -> (key * 'a) option
  val is_singleton : 'a t -> (key * 'a) option

  val of_list : (key * 'a) list -> 'a t

  val pp_map : ?pp_sep:unit pp -> (Format.formatter -> key -> 'a -> unit) -> 'a t pp
  (** Custom pretty printer *)

  val hash : ('a -> int) -> 'a t -> int

  (** Map whose images are lists of elements *)
  module List : sig
    val add : key -> 'a -> 'a list t -> 'a list t
    (** Add element to the list at key, create list if non-existent *)

    val mem : key -> 'a -> 'a list t -> bool

    val find : key -> 'a list t -> 'a list
    (** empty list if not present *)

    (* Iterate on all elements for each key, so key is seen multiple times
       keys are seen in order, then elements in order of the list as map.key *)
    val fold : (key -> 'a -> 'b -> 'b) -> 'a list t -> 'b -> 'b
    val iter : (key -> 'a -> unit) -> 'a list t -> unit
    val mapi : (key -> 'a -> 'b) -> 'a list t -> 'b list t
    val map : ('a -> 'b) -> 'a list t -> 'b list t

    (* These maintain the invariant that no element points to [[]] *)
    val filter : (key -> 'a -> bool) -> 'a list t -> 'a list t
    val mapi_bind : (key -> 'a -> 'b list) -> 'a list t -> 'b list t
    val map_bind : ('a -> 'b list) -> 'a list t -> 'b list t
    val mapi_filter : (key -> 'a -> 'b option) -> 'a list t -> 'b list t
    val map_filter : ('a -> 'b option) -> 'a list t -> 'b list t
  end

  (** Maps whoses images are sets of elements *)
  module Set (S : SET) : sig
    val add : key -> S.elt -> S.t t -> S.t t
    val mem : key -> S.elt -> S.t t -> bool

    val find : key -> S.t t -> S.t
    (** [S.empty] when not present *)

    (* Iterate on all elements for each key, so key is seen multiple times
       keys are seen in order, then elements in order of the set as map.key *)
    val fold : (key -> S.elt -> 'b -> 'b) -> S.t t -> 'b -> 'b
    val iter : (key -> S.elt -> unit) -> S.t t -> unit
    val mapi : (key -> S.elt -> S.elt) -> S.t t -> S.t t
    val map : (S.elt -> S.elt) -> S.t t -> S.t t

    (* These maintain the invariant that no element points to [S.empty] *)
    val filter : (key -> S.elt -> bool) -> S.t t -> S.t t
    val mapi_bind : (key -> S.elt -> S.t) -> S.t t -> S.t t
    val map_bind : (S.elt -> S.t) -> S.t t -> S.t t
    val mapi_filter : (key -> S.elt -> S.elt option) -> S.t t -> S.t t
    val map_filter : (S.elt -> S.elt option) -> S.t t -> S.t t
  end

  (** Maps whoses images are maps of elements *)
  module Map (M : Map.S) : sig
    val add : key -> M.key -> 'a -> 'a M.t t -> 'a M.t t
    val mem : key -> M.key -> 'a M.t t -> bool

    val find : key -> M.key -> 'a M.t t -> 'a
    (** @raises [Not_found] if not present *)

    val find_opt : key -> M.key -> 'a M.t t -> 'a option

    (* Iterate on all elements for each key, so key is seen multiple times
       keys are seen in order, then elements in order of the set as map.key *)
    val fold : (key -> M.key -> 'a -> 'b -> 'b) -> 'a M.t t -> 'b -> 'b
    val iter : (key -> M.key -> 'a -> unit) -> 'a M.t t -> unit
    val mapi : (key -> M.key -> 'a -> 'b) -> 'a M.t t -> 'b M.t t
    val map : ('a -> 'b) -> 'a M.t t -> 'b M.t t

    (* These maintain the invariant that no element points to [M.empty] *)
    val filter : (key -> M.key -> 'a -> bool) -> 'a M.t t -> 'a M.t t
    val mapi_filter : (key -> M.key -> 'a -> 'b option) -> 'a M.t t -> 'b M.t t
    val map_filter : ('a -> 'b option) -> 'a M.t t -> 'b M.t t
  end
end
