(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

open Types

let pp_dot fmt () = Format.fprintf fmt ".@,"
let pp_semicolon fmt () = Format.fprintf fmt ";@ "

let use_color = ref (Unix.isatty Unix.stdout)
let use_unicode = ref true

module Unicode = struct
  (** Function taken from frama-c utf8_logic*)
  let select_char unicode ascii () = if !use_unicode then unicode else ascii

  let colon_eq = select_char "≔" ":="
  let leq = select_char "≤" "<="
  let geq = select_char "≥" ">="
  let eq = select_char "=" "=="
  let neq = select_char "≠" "!="
  let emptyset = select_char "∅" "empty"
  let logical_not = select_char "¬" "!"
  let logical_and = select_char "∧" "&&"
  let logical_or = select_char "∨" "||"
  let elem_of = select_char "∈" "in"
  let intersection = select_char "∩" "inter"
  let microseconds = select_char "µs" "microseconds"
  let z = select_char "ℤ" "Z"
  let bottom = select_char "⊥" "bottom"
end

let printf = if !use_color then Ocolor_format.printf else Format.printf

(** boost::hash_combine: smart-ish way to combine hashes, better than xor
    (not symetric, less 0 collisions...)
    https://stackoverflow.com/questions/5889238/why-is-xor-the-default-way-to-combine-hashes *)
let hash_combine lhs rhs = lhs lxor (rhs + 0x9e3779b9 + (lhs lsl 6) + (lhs lsr 2))

(** Hash for a sum type *)
let hash_sum ~log_total ~nb hash = abs ((hash lsl log_total) + nb)

(** Compare a pair in lexicographic order *)
let compare_pair cmp_1 cmp_2 l1 r1 l2 r2 =
  let cmp = cmp_1 l1 r1 in
  if cmp = 0 then cmp_2 l2 r2 else cmp

let swap (a, b) = (b, a)

let compare_inf = Int.compare 0 1 (* The return value of [compare x y] when x<y *)
let compare_sup = -compare_inf

let order_of_compare = function
  | 0 -> Eq
  | x when x < 0 -> Lt
  | _ -> Gt

let equal_of_compare = function
  | 0 -> true
  | _ -> false

let compare_of_order = function
  | Lt -> compare_inf
  | Eq -> 0
  | Gt -> compare_sup

let equal_of_order = function
  | Eq -> true
  | _ -> false

let unary_order_of_compare compare order a b =
  order_of_compare (compare (fun a b -> compare_of_order (order a b)) a b)

let order_of_poly_order : type a b. (a, b) poly_order -> order = function
  | P_Lt -> Lt
  | P_Eq -> Eq
  | P_Gt -> Gt

let equal_of_poly_equal : type a b. (a, b) poly_equal -> bool = function
  | Equal -> true
  | Diff -> false

exception Not_Singleton

exception Bottom
(** Exception raised when reaching an impossible state *)

module Set (A : Types.TYPE) = struct
  include Set.Make (A)

  let pp_set ?(pp_sep = pp_semicolon) fmt s =
    Format.fprintf fmt "@[%a@]" (Format.pp_print_seq ~pp_sep A.pp) (to_seq s)
  let pp fmt s = pp_set fmt s

  let bind f s = fold (fun elt s -> union (f elt) s) s empty

  let order a b = order_of_compare (compare a b)

  let find_any = find_first
  let find_any_opt = find_first_opt

  let intersect l r = not (disjoint l r)

  let is_singleton s =
    try fold (fun a b -> if b = None then Some a else raise Not_Singleton) s None
    with Not_Singleton -> None

  (* Sets with same elements but different tree structure will have same hash with this... *)
  let hash x = fold (fun elt hash -> hash_combine (A.hash elt) hash) x 0
end

module MapExtra (M : Map.S) = struct
  include M

  let find_any f t = find_first (fun k -> f k (find k t)) t
  let find_any_opt f t = find_first_opt (fun k -> f k (find k t)) t

  let order order a b = unary_order_of_compare compare order a b

  let is_singleton s =
    try fold (fun k v b -> if b = None then Some (k, v) else raise Not_Singleton) s None
    with Not_Singleton -> None

  let of_list l = of_seq (List.to_seq l)

  let pp_map ?pp_sep pp_k_v fmt t =
    Format.fprintf fmt "%a"
      (Format.pp_print_seq ?pp_sep (fun fmt (k, v) -> pp_k_v fmt k v))
      (to_seq t)

  let hash hash_key hash_val map =
    fold
      (fun key value hash -> hash_combine (hash_combine (hash_key key) (hash_val value)) hash)
      map 0

  module List = struct
    let add k e m =
      match find_opt k m with
      | Some l -> add k (e :: l) m
      | None -> add k [ e ] m

    let mem k e m = try List.mem e (find k m) with Not_found -> false

    let find k m = try find k m with Not_found -> []

    let fold f m acc = fold (fun k l acc -> List.fold_left (fun acc e -> f k e acc) acc l) m acc

    let iter f m = iter (fun k l -> List.iter (f k) l) m

    let mapi f m = mapi (fun k v -> List.map (f k) v) m
    let map f m = map (List.map f) m

    let list2opt = function
      | [] -> None
      | l -> Some l
    let mapi_bind f m = filter_map (fun k v -> list2opt (List.concat_map (f k) v)) m
    let map_bind f m = filter_map (fun _ v -> list2opt (List.concat_map f v)) m

    let mapi_filter f m = filter_map (fun k v -> list2opt (List.filter_map (f k) v)) m
    let map_filter f m = filter_map (fun _ v -> list2opt (List.filter_map f v)) m
    let filter f m = filter_map (fun k v -> list2opt (List.filter (f k) v)) m
  end

  module Set (S : Types.SET) = struct
    let add k e m =
      match find_opt k m with
      | Some s -> add k (S.add e s) m
      | None -> add k (S.singleton e) m

    let mem k e m = try S.mem e (find k m) with Not_found -> false

    let find k m = try find k m with Not_found -> S.empty

    let fold f m acc = fold (fun k l acc -> S.fold (f k) l acc) m acc

    let iter f m = iter (fun k l -> S.iter (f k) l) m

    let mapi f m = mapi (fun k v -> S.map (f k) v) m
    let map f m = map (S.map f) m

    let set2opt s = if S.is_empty s then None else Some s
    let mapi_bind f m = filter_map (fun k v -> set2opt (S.bind (f k) v)) m
    let map_bind f m = filter_map (fun _ v -> set2opt (S.bind f v)) m

    let mapi_filter f m = filter_map (fun k v -> set2opt (S.filter_map (f k) v)) m
    let map_filter f m = filter_map (fun _ v -> set2opt (S.filter_map f v)) m
    let filter f m = filter_map (fun k v -> set2opt (S.filter (f k) v)) m
  end

  module Map (M : Map.S) = struct
    let add k k' e m =
      match find_opt k m with
      | None -> add k (M.singleton k' e) m
      | Some m' -> add k (M.add k' e m') m
    let mem k k' m = try M.mem k' (find k m) with Not_found -> false
    let find k k' m = M.find k' (find k m)
    let find_opt k k' m = try Some (find k k' m) with Not_found -> None

    let fold f m acc = fold (fun k m' acc -> M.fold (f k) m' acc) m acc

    let iter f m = iter (fun k m' -> M.iter (f k) m') m

    let mapi f m = mapi (fun k m' -> M.mapi (f k) m') m
    let map f m = map (M.map f) m

    let map2opt m = if M.is_empty m then None else Some m

    let mapi_filter f m = filter_map (fun k m' -> map2opt (M.filter_map (f k) m')) m
    let map_filter f m = filter_map (fun _ m' -> map2opt (M.filter_map (fun _ -> f) m')) m
    let filter f m = filter_map (fun k m' -> map2opt (M.filter (f k) m')) m
  end
end

module Map (A : Types.TYPE) = struct
  module M = Map.Make (A)
  include MapExtra (M)

  let hash hash_val m = hash A.hash hash_val m
end

let debug = ref false

let set_debug x = debug := x

let debug_pause () =
  if !debug then
    let _ = read_line () in
    ()

let debug_printf x = if !debug then printf x else Format.ifprintf Format.std_formatter x
