(**************************************************************************)
(*  This file is part of the TAI abstract interpreter.                    *)
(*                                                                        *)
(*  Copyright (C) 2024                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  You can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(**************************************************************************)

(** Various utility functions and modules. Includes pretty-printer,
    extended Stdlib Sets and Maps, unicode characters and more. *)

open Types

val pp_dot : unit pp
(** print a "." separator with break *)

val pp_semicolon : unit pp
(** print a "; " separator with breaking space *)

val use_color : bool ref
(** Toggle on/off ansi codes in [printf]. Default is [Unix.isatty] *)

val use_unicode : bool ref
(** Toggle on/off unicode displays *)

(** Unicode symbols (or full text equivalent if unicode is off) *)
module Unicode : sig
  val colon_eq : unit -> string
  val leq : unit -> string
  val geq : unit -> string
  val eq : unit -> string
  val neq : unit -> string
  val emptyset : unit -> string
  val logical_and : unit -> string
  val logical_or : unit -> string
  val logical_not : unit -> string
  val elem_of : unit -> string
  val intersection : unit -> string
  val microseconds : unit -> string
  val z : unit -> string
  val bottom : unit -> string
end

val printf : ('a, Format.formatter, unit) format -> 'a
(** Set to [Ocolor_format.printf] or [Format.printf] to enable/disable coloring *)

val debug_printf : ('a, Format.formatter, unit) format -> 'a
(** Print debug information in step by step mode *)

val debug_pause : unit -> unit
(** Pause (using read_line) in step by step mode *)

val set_debug : bool -> unit
(** Set/unset step by step mode *)

val hash_combine : int -> int -> int
(** Function used to combine a pair of hashes *)

val hash_sum : log_total:int -> nb:int -> int -> int
(** [hash_sum ~nb ~log_total h] returns a hash for contructor [~nb]
    (between [0] and [2^(~log_total)-1]) whose arguments hash top [h] *)

val swap : 'a * 'b -> 'b * 'a

val compare_pair : ('a -> 'a -> int) -> ('b -> 'b -> int) -> 'a -> 'a -> 'b -> 'b -> int
(** Compare a pair in lexicographic order *)

val compare_inf : int
val compare_sup : int

val order_of_compare : int -> order
val equal_of_compare : int -> bool

val compare_of_order : order -> int
val equal_of_order : order -> bool

val order_of_poly_order : ('a, 'b) poly_order -> order
val equal_of_poly_equal : ('a, 'b) poly_equal -> bool

val unary_order_of_compare :
  (('a -> 'a -> int) -> 'at -> 'at -> int) -> ('a -> 'a -> order) -> 'at -> 'at -> order

(** Set with extra stuff (pretty-printer) *)
module Set (A : Types.TYPE) : Types.SET with type elt = A.t

exception Bottom
(** Exception raised when reaching an impossible state *)

module MapExtra (M : Map.S) : sig
  include Types.MAP with type 'a t = 'a M.t and type key = M.key
  val hash : (key -> int) -> ('a -> int) -> 'a t -> int
end

(** Map with extra stuff (pretty-printer) *)
module Map (A : Types.TYPE) : Types.MAP with type key = A.t
